(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/core.js                                                      //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
var Crypto = Npm.require("crypto");                                    // 17
var Capnp = Npm.require("capnp");                                      // 18
                                                                       //
var PersistentHandle = Capnp.importSystem("sandstorm/supervisor.capnp").PersistentHandle;
var SandstormCore = Capnp.importSystem("sandstorm/supervisor.capnp").SandstormCore;
var SandstormCoreFactory = Capnp.importSystem("sandstorm/backend.capnp").SandstormCoreFactory;
var PersistentOngoingNotification = Capnp.importSystem("sandstorm/supervisor.capnp").PersistentOngoingNotification;
var PersistentUiView = Capnp.importSystem("sandstorm/persistentuiview.capnp").PersistentUiView;
                                                                       //
var SandstormCoreImpl = (function () {                                 //
  function SandstormCoreImpl(grainId) {                                // 27
    babelHelpers.classCallCheck(this, SandstormCoreImpl);              //
                                                                       //
    this.grainId = grainId;                                            // 28
  }                                                                    //
                                                                       //
  SandstormCoreImpl.prototype.restore = (function () {                 // 26
    function restore(sturdyRef, requiredPermissions) {                 // 31
      var _this = this;                                                // 32
      return inMeteor(function () {                                    // 33
        var hashedSturdyRef = hashSturdyRef(sturdyRef);                // 34
        var token = ApiTokens.findOne(hashedSturdyRef);                // 35
        var requirements = [{                                          // 36
          tokenValid: hashedSturdyRef                                  // 38
        }];                                                            //
                                                                       //
        if (requiredPermissions && token.owner.grain.introducerIdentity) {
          requirements.push({                                          // 43
            permissionsHeld: {                                         // 44
              permissions: requiredPermissions,                        // 45
              identityId: token.owner.grain.introducerIdentity,        // 46
              grainId: _this.grainId                                   // 47
            }                                                          //
          });                                                          //
        }                                                              //
                                                                       //
        return restoreInternal(hashedSturdyRef, { grain: Match.ObjectIncluding({ grainId: _this.grainId }) }, requirements, sturdyRef);
      });                                                              //
    }                                                                  //
                                                                       //
    return restore;                                                    //
  })();                                                                //
                                                                       //
  SandstormCoreImpl.prototype.drop = (function () {                    // 26
    function drop(sturdyRef) {                                         // 58
      var _this = this;                                                // 59
      return inMeteor(function () {                                    // 60
        return dropInternal(sturdyRef, { grain: Match.ObjectIncluding({ grainId: _this.grainId }) });
      });                                                              //
    }                                                                  //
                                                                       //
    return drop;                                                       //
  })();                                                                //
                                                                       //
  SandstormCoreImpl.prototype.makeToken = (function () {               // 26
    function makeToken(ref, owner, requirements) {                     // 65
      var _this = this;                                                // 66
      return inMeteor(function () {                                    // 67
        var sturdyRef = new Buffer(generateSturdyRef());               // 68
        var hashedSturdyRef = hashSturdyRef(sturdyRef);                // 69
        ApiTokens.insert({                                             // 70
          _id: hashedSturdyRef,                                        // 71
          grainId: _this.grainId,                                      // 72
          objectId: ref,                                               // 73
          owner: owner,                                                // 74
          created: new Date(),                                         // 75
          requirements: requirements                                   // 76
        });                                                            //
                                                                       //
        return {                                                       // 79
          token: sturdyRef                                             // 80
        };                                                             //
      });                                                              //
    }                                                                  //
                                                                       //
    return makeToken;                                                  //
  })();                                                                //
                                                                       //
  SandstormCoreImpl.prototype.makeChildToken = (function () {          // 26
    function makeChildToken(parent, owner, requirements) {             // 85
      var _this = this;                                                // 86
      return inMeteor(function () {                                    // 87
        return makeChildTokenInternal(hashSturdyRef(parent), owner, requirements, _this.grainId);
      });                                                              //
    }                                                                  //
                                                                       //
    return makeChildToken;                                             //
  })();                                                                //
                                                                       //
  SandstormCoreImpl.prototype.getOwnerNotificationTarget = (function () {
    function getOwnerNotificationTarget() {                            // 92
      var grainId = this.grainId;                                      // 93
      return {                                                         // 94
        owner: {                                                       // 95
          addOngoing: function (displayInfo, notification) {           // 96
            return inMeteor(function () {                              // 97
              var grain = Grains.findOne({ _id: grainId });            // 98
              if (!grain) {                                            // 99
                throw new Error("Grain not found.");                   // 100
              }                                                        //
                                                                       //
              var castedNotification = notification.castAs(PersistentOngoingNotification);
              var wakelockToken = waitPromise(castedNotification.save()).sturdyRef;
                                                                       //
              // We have to close both the casted cap and the original. Perhaps this should be fixed in
              // node-capnp?                                           //
              castedNotification.close();                              // 108
              notification.close();                                    // 109
              var notificationId = Notifications.insert({              // 110
                ongoing: wakelockToken,                                // 111
                grainId: grainId,                                      // 112
                userId: grain.userId,                                  // 113
                text: displayInfo.caption,                             // 114
                timestamp: new Date(),                                 // 115
                isUnread: true                                         // 116
              });                                                      //
                                                                       //
              return { handle: makeNotificationHandle(notificationId, false) };
            });                                                        //
          }                                                            //
        }                                                              //
      };                                                               //
    }                                                                  //
                                                                       //
    return getOwnerNotificationTarget;                                 //
  })();                                                                //
                                                                       //
  return SandstormCoreImpl;                                            //
})();                                                                  //
                                                                       //
var makeSandstormCore = function (grainId) {                           // 127
  return new Capnp.Capability(new SandstormCoreImpl(grainId), SandstormCore);
};                                                                     //
                                                                       //
var NotificationHandle = (function () {                                //
  function NotificationHandle(notificationId, saved) {                 // 132
    babelHelpers.classCallCheck(this, NotificationHandle);             //
                                                                       //
    this.notificationId = notificationId;                              // 133
    this.saved = saved;                                                // 134
  }                                                                    //
                                                                       //
  NotificationHandle.prototype.close = (function () {                  // 131
    function close() {                                                 // 137
      var _this = this;                                                // 138
      return inMeteor(function () {                                    // 139
        if (!_this.saved) {                                            // 140
          dismissNotification(_this.notificationId);                   // 141
        }                                                              //
      });                                                              //
    }                                                                  //
                                                                       //
    return close;                                                      //
  })();                                                                //
                                                                       //
  NotificationHandle.prototype.save = (function () {                   // 131
    function save(params) {                                            // 146
      return saveFrontendRef({ notificationHandle: this.notificationId }, params.sealFor);
    }                                                                  //
                                                                       //
    return save;                                                       //
  })();                                                                //
                                                                       //
  return NotificationHandle;                                           //
})();                                                                  //
                                                                       //
var PersistentUiViewImpl = (function () {                              //
  function PersistentUiViewImpl(token, sturdyRef, membraneRequirements) {
    babelHelpers.classCallCheck(this, PersistentUiViewImpl);           //
                                                                       //
    this._token = token;                                               // 153
    this._sturdyRef = sturdyRef;                                       // 154
    // We need this token so we can later pass it to SandstormPermissions.createNewApiToken, which
    // requires the sturdyRef of the token it's producing a child for, not just the hashed token id.
                                                                       //
    this._membraneRequirements = membraneRequirements || [];           // 158
    // TODO(someday): implement SystemPersistent.addRequirements() in a way that collects
    // requirements, and then saves them appropriately in save().      //
  }                                                                    //
                                                                       //
  PersistentUiViewImpl.prototype.save = (function () {                 // 151
    function save(params) {                                            // 163
      var _this2 = this;                                               //
                                                                       //
      // Create a new child apiToken for the user whose session attempted to save this capability.
      // The token should specify this.token as the parent token, and needs to gain an additional membrane
      // requirement (TODO):                                           //
      // * if params.sealFor is a user, then the user must still have access to the grain that offered
      //   this cap.  Today, this is handled by offer() in SessionContextImpl.
      var res = inMeteor(function () {                                 // 169
        var owner = _.extend({}, params.sealFor);                      // 170
        var grainId = _this2._token.grainId;                           // 171
                                                                       //
        if (owner.user) {                                              // 173
          // Copy the owner's save label for this capability, if available, to be the new title.
          if (_this2._token && _this2._token.owner && _this2._token.owner.grain && _this2._token.owner.grain.saveLabel && _this2._token.owner.grain.saveLabel.defaultText) {
            owner.user.title = _this2._token.owner.grain.saveLabel.defaultText;
          }                                                            //
                                                                       //
          // createNewApiToken() expects only these fields if owner is a user, and will add the rest itself.
          owner = { user: {                                            // 181
              identityId: owner.user.identityId,                       // 182
              title: owner.user.title                                  // 183
            } };                                                       //
        }                                                              //
                                                                       //
        var ret = SandstormPermissions.createNewApiToken(globalDb, { rawParentToken: _this2._sturdyRef.toString() }, _this2._token.grainId, _this2._token.petname, _this2._token.roleAssignment, owner);
                                                                       //
        return { sturdyRef: new Buffer(ret.token) };                   // 195
      });                                                              //
      return res;                                                      // 197
    }                                                                  //
                                                                       //
    return save;                                                       //
  })();                                                                //
                                                                       //
  // All other UiView methods are currently unimplemented, which, while not strictly correct,
  // results in the same overall behavior, since users can't call restore() on a PersistentUiView,
  // and grains can't call methods on UiViews because they lack the "is human" pseudopermission.
  return PersistentUiViewImpl;                                         //
})();                                                                  //
                                                                       //
var makePersistentUiView = function (token, sturdyRef) {               // 205
  return new Capnp.Capability(new PersistentUiViewImpl(token, sturdyRef, []), PersistentUiView);
};                                                                     //
                                                                       //
function makeNotificationHandle(notificationId, saved) {               // 209
  return new Capnp.Capability(new NotificationHandle(notificationId, saved), PersistentHandle);
}                                                                      //
                                                                       //
function dropWakelock(grainId, wakeLockNotificationId) {               // 213
  waitPromise(globalBackend.useGrain(grainId, function (supervisor) {  // 214
    return supervisor.drop({ wakeLockNotification: wakeLockNotificationId });
  }));                                                                 //
}                                                                      //
                                                                       //
function dismissNotification(notificationId, callCancel) {             // 219
  var notification = Notifications.findOne({ _id: notificationId });   // 220
  if (notification) {                                                  // 221
    Notifications.remove({ _id: notificationId });                     // 222
    if (notification.ongoing) {                                        // 223
      // For some reason, Mongo returns an object that looks buffer-like, but isn't a buffer.
      // Only way to fix seems to be to copy it.                       //
      var id = new Buffer(notification.ongoing);                       // 226
                                                                       //
      if (!callCancel) {                                               // 228
        dropInternal(id, { frontend: null });                          // 229
      } else {                                                         //
        var notificationCap = restoreInternal(hashSturdyRef(id), { frontend: null }).cap;
        var castedNotification = notificationCap.castAs(PersistentOngoingNotification);
        dropInternal(id, { frontend: null });                          // 233
        try {                                                          // 234
          waitPromise(castedNotification.cancel());                    // 235
          castedNotification.close();                                  // 236
          notificationCap.close();                                     // 237
        } catch (err) {                                                //
          if (err.kjType !== "disconnected") {                         // 239
            // ignore disconnected errors, since cancel may shutdown the grain before the supervisor
            // responds.                                               //
            throw err;                                                 // 242
          }                                                            //
        }                                                              //
      }                                                                //
    } else if (notification.appUpdates) {                              //
      _.forEach(notification.appUpdates, function (app, appId) {       // 247
        deletePackage(app.packageId);                                  // 248
      });                                                              //
    }                                                                  //
  }                                                                    //
}                                                                      //
                                                                       //
hashSturdyRef = function (sturdyRef) {                                 // 254
  return Crypto.createHash("sha256").update(sturdyRef).digest("base64");
};                                                                     //
                                                                       //
function generateSturdyRef() {                                         // 258
  return Random.id(22);                                                // 259
}                                                                      //
                                                                       //
Meteor.methods({                                                       // 262
  dismissNotification: function (notificationId) {                     // 263
    // This will remove notifications from the database and from view of the user.
    // For ongoing notifications, it will begin the process of cancelling and dropping them from
    // the app.                                                        //
                                                                       //
    check(notificationId, String);                                     // 268
                                                                       //
    var notification = Notifications.findOne({ _id: notificationId });
    if (!notification) {                                               // 271
      throw new Meteor.Error(404, "Notification id not found.");       // 272
    } else if (notification.userId !== Meteor.userId()) {              //
      throw new Meteor.Error(403, "Notification does not belong to current user.");
    } else {                                                           //
      dismissNotification(notificationId, true);                       // 276
    }                                                                  //
  },                                                                   //
                                                                       //
  readAllNotifications: function () {                                  // 280
    // Marks all notifications as read for the current user.           //
    if (!Meteor.userId()) {                                            // 282
      throw new Meteor.Error(403, "User not logged in.");              // 283
    }                                                                  //
                                                                       //
    Notifications.update({ userId: Meteor.userId() }, { $set: { isUnread: false } }, { multi: true });
  }                                                                    //
});                                                                    //
                                                                       //
saveFrontendRef = function (frontendRef, owner, requirements) {        // 290
  return inMeteor(function () {                                        // 291
    var sturdyRef = new Buffer(generateSturdyRef());                   // 292
    var hashedSturdyRef = hashSturdyRef(sturdyRef);                    // 293
    ApiTokens.insert({                                                 // 294
      _id: hashedSturdyRef,                                            // 295
      frontendRef: frontendRef,                                        // 296
      owner: owner,                                                    // 297
      created: new Date(),                                             // 298
      requirements: requirements                                       // 299
    });                                                                //
    return { sturdyRef: sturdyRef };                                   // 301
  });                                                                  //
};                                                                     //
                                                                       //
checkRequirements = function (requirements) {                          // 305
  if (!requirements) {                                                 // 306
    return true;                                                       // 307
  }                                                                    //
                                                                       //
  for (var i in babelHelpers.sanitizeForInObject(requirements)) {      // 310
    var requirement = requirements[i];                                 // 311
    if (requirement.tokenValid) {                                      // 312
      var token = ApiTokens.findOne({ _id: requirement.tokenValid }, { fields: { requirements: 1 } });
      if (!checkRequirements(token.requirements)) {                    // 314
        return false;                                                  // 315
      }                                                                //
    } else if (requirement.permissionsHeld) {                          //
      var p = requirement.permissionsHeld;                             // 318
      var viewInfo = Grains.findOne(p.grainId, { fields: { cachedViewInfo: 1 } }).cachedViewInfo;
      var currentPermissions = SandstormPermissions.grainPermissions(globalDb, { grain: { _id: p.grainId, identityId: p.identityId } }, viewInfo || {});
      if (!currentPermissions) {                                       // 322
        return false;                                                  // 323
      }                                                                //
                                                                       //
      for (var ii = 0; ii < p.permissions.length; ++ii) {              // 326
        if (p.permissions[ii] && !currentPermissions[ii]) {            // 327
          return false;                                                // 328
        }                                                              //
      }                                                                //
    } else if (requirement.userIsAdmin) {                              //
      if (!isAdminById(requirement.userIsAdmin)) {                     // 332
        return false;                                                  // 333
      }                                                                //
    } else {                                                           //
      throw new Meteor.Error(403, "Unknown requirement");              // 336
    }                                                                  //
  }                                                                    //
                                                                       //
  return true;                                                         // 340
};                                                                     //
                                                                       //
restoreInternal = function (tokenId, ownerPattern, requirements, parentToken) {
  // Restores `sturdyRef`, checking first that its owner matches `ownerPattern`.
  // parentToken and requirements are optional params that are only used in the case of an objectId
  // token                                                             //
  var token = ApiTokens.findOne(tokenId);                              // 347
  if (!token) {                                                        // 348
    throw new Meteor.Error(403, "No token found to restore");          // 349
  }                                                                    //
                                                                       //
  if (token.revoked) {                                                 // 352
    throw new Meteor.Error(403, "Token has been revoked");             // 353
  }                                                                    //
  // The ownerPattern should specify the appropriate user or grain involved, if appropriate.
  check(token.owner, ownerPattern);                                    // 356
  if (!checkRequirements(token.requirements)) {                        // 357
    throw new Meteor.Error(403, "Requirements not satisfied.");        // 358
  }                                                                    //
                                                                       //
  if (token.expires && token.expires.getTime() <= Date.now()) {        // 361
    throw new Meteor.Error(403, "Authorization token expired");        // 362
  }                                                                    //
                                                                       //
  if (token.expiresIfUnused) {                                         // 365
    if (token.expiresIfUnused.getTime() <= Date.now()) {               // 366
      throw new Meteor.Error(403, "Authorization token expired");      // 367
    } else {                                                           //
      // It's getting used now, so clear the expiresIfUnused field.    //
      ApiTokens.update(token._id, { $set: { expiresIfUnused: null } });
    }                                                                  //
  }                                                                    //
                                                                       //
  if (token.frontendRef) {                                             // 374
    // A token which represents a capability implemented by a pseudo-driver.
    if (token.frontendRef.notificationHandle) {                        // 376
      var notificationId = token.frontendRef.notificationHandle;       // 377
      return { cap: makeNotificationHandle(notificationId, true) };    // 378
    } else if (token.frontendRef.ipNetwork) {                          //
      return { cap: makeIpNetwork(tokenId) };                          // 380
    } else if (token.frontendRef.ipInterface) {                        //
      return { cap: makeIpInterface(tokenId) };                        // 382
    } else {                                                           //
      throw new Meteor.Error(500, "Unknown frontend token type.");     // 384
    }                                                                  //
  } else if (token.objectId) {                                         //
    // A token which represents a specific capability exported by a grain.
    if (!checkRequirements(requirements)) {                            // 388
      throw new Meteor.Error(403, "Requirements not satisfied.");      // 389
    }                                                                  //
                                                                       //
    if (token.objectId.appRef) {                                       // 392
      token.objectId.appRef = new Buffer(token.objectId.appRef);       // 393
    }                                                                  //
                                                                       //
    // Ensure the grain is running, then restore the capability.       //
    return waitPromise(globalBackend.useGrain(token.grainId, function (supervisor) {
      return supervisor.restore(token.objectId, requirements, parentToken);
    }));                                                               //
  } else if (token.parentToken) {                                      //
    // A token which chains to some parent token.  Restore the parent token (possibly recursively),
    // checking requirements on the way up.                            //
    return restoreInternal(token.parentToken, Match.Any, requirements, parentToken);
  }                                                                    //
                                                                       //
  // Note that hereafter, `token` should be, by process of elimination, a UiView token.
  if (!token.grainId) {                                                // 407
    throw new Meteor.Error(500, "Expected token " + token._id + " to have a grainId");
  }                                                                    //
                                                                       //
  // If a grain is attempting to restore a UiView, it gets a UiView which filters out all
  // the method calls.  In the future, we may allow grains to restore UiViews that pass along the
  // "is human" pseudopermission (say, to allow embedding grains inside other grains), which will
  // return a different capability.                                    //
  return { cap: makePersistentUiView(token, parentToken) };            // 415
};                                                                     //
                                                                       //
function dropInternal(sturdyRef, ownerPattern) {                       // 418
  // Drops `sturdyRef`, checking first that its owner matches `ownerPattern`.
                                                                       //
  var hashedSturdyRef = hashSturdyRef(sturdyRef);                      // 421
  var token = ApiTokens.findOne({ _id: hashedSturdyRef });             // 422
  if (!token) {                                                        // 423
    return;                                                            // 424
  }                                                                    //
                                                                       //
  check(token.owner, ownerPattern);                                    // 427
                                                                       //
  if (token.frontendRef) {                                             // 429
    if (token.frontendRef.notificationHandle) {                        // 430
      var notificationId = token.frontendRef.notificationHandle;       // 431
      ApiTokens.remove({ _id: hashedSturdyRef });                      // 432
      var anyToken = ApiTokens.findOne({ "frontendRef.notificationHandle": notificationId });
      if (!anyToken) {                                                 // 434
        // No other tokens referencing this notification exist, so dismiss the notification
        dismissNotification(notificationId);                           // 436
      }                                                                //
    } else {                                                           //
      throw new Error("Unknown frontend token type.");                 // 439
    }                                                                  //
  } else if (token.objectId) {                                         //
    if (token.objectId.wakeLockNotification) {                         // 442
      dropWakelock(token.grainId, token.objectId.wakeLockNotification);
    } else {                                                           //
      throw new Error("Unknown objectId token type.");                 // 445
    }                                                                  //
  } else {                                                             //
    throw new Error("Unknown token type.");                            // 448
  }                                                                    //
}                                                                      //
                                                                       //
makeChildTokenInternal = function (hashedParent, owner, requirements, grainId) {
  var sturdyRef = new Buffer(generateSturdyRef());                     // 453
  var hashedSturdyRef = hashSturdyRef(sturdyRef);                      // 454
                                                                       //
  requirements = requirements.filter(function (requirement) {          // 456
    return requirement.tokenValid !== hashedParent;                    // 457
  });                                                                  //
                                                                       //
  var tokenInfo = {                                                    // 460
    _id: hashedSturdyRef,                                              // 461
    parentToken: hashedParent,                                         // 462
    owner: owner,                                                      // 463
    created: new Date(),                                               // 464
    requirements: requirements                                         // 465
  };                                                                   //
  if (grainId) {                                                       // 467
    tokenInfo.grainId = grainId;                                       // 468
  }                                                                    //
                                                                       //
  ApiTokens.insert(tokenInfo);                                         // 471
                                                                       //
  return {                                                             // 473
    token: sturdyRef                                                   // 474
  };                                                                   //
};                                                                     //
                                                                       //
function SandstormCoreFactoryImpl() {}                                 // 478
                                                                       //
SandstormCoreFactoryImpl.prototype.getSandstormCore = function (grainId) {
  return { core: makeSandstormCore(grainId) };                         // 482
};                                                                     //
                                                                       //
makeSandstormCoreFactory = function () {                               // 485
  return new Capnp.Capability(new SandstormCoreFactoryImpl(), SandstormCoreFactory);
};                                                                     //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=core.js.map
