(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/pre-meteor.js                                                //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
// This file implements logic that we place in front of our main Meteor application,
// including routing of requests to proxies and handling of static web publishing.
                                                                       //
var Url = Npm.require("url");                                          // 20
var Fs = Npm.require("fs");                                            // 21
var Dns = Npm.require("dns");                                          // 22
var Promise = Npm.require("es6-promise").Promise;                      // 23
var Future = Npm.require("fibers/future");                             // 24
var Http = Npm.require("http");                                        // 25
                                                                       //
var HOSTNAME = Url.parse(process.env.ROOT_URL).hostname;               // 27
var DDP_HOSTNAME = process.env.DDP_DEFAULT_CONNECTION_URL && Url.parse(process.env.DDP_DEFAULT_CONNECTION_URL).hostname;
var CACHE_TTL_SECONDS = 30; // 30 seconds.  Cache-Control expects units of seconds, not millis.
var DNS_CACHE_TTL = CACHE_TTL_SECONDS * 1000; // DNS cache is in millis.
                                                                       //
var staticHandlers = {};                                               // 33
// Maps grain public IDs to Connect handlers.                          //
// TODO(perf): Garbage-collect this map?                               //
                                                                       //
var dnsCache = {};                                                     // 37
// Unfortunately, node's DNS library doesn't cache results, so we do our own caching.
// Unfortunately, node's DNS library also dosen't give us TTLs. So, we'll cache for
// DNS_CACHE_TTL (a relatively small value) and rely on the upstream DNS server to implement
// better caching.                                                     //
                                                                       //
function isSandstormShell(hostname) {                                  // 43
  // Is this hostname mapped to the Sandstorm shell?                   //
                                                                       //
  return hostname === HOSTNAME || DDP_HOSTNAME && hostname === DDP_HOSTNAME;
}                                                                      //
                                                                       //
var mime = Connect["static"].mime;                                     // 49
                                                                       //
function wwwHandlerForGrain(grainId) {                                 // 51
  return function (request, response, cb) {                            // 52
    var path = request.url;                                            // 53
                                                                       //
    // If a directory, open 'index.html'.                              //
    if (path.slice(-1) === "/") {                                      // 56
      path = path + "index.html";                                      // 57
    }                                                                  //
                                                                       //
    // Strip leading '/'.                                              //
    if (path[0] === "/") path = path.slice(1);                         // 61
                                                                       //
    // Strip query.                                                    //
    path = path.split("?")[0];                                         // 64
                                                                       //
    var type = mime.lookup(path);                                      // 66
    var charset = mime.charsets.lookup(type);                          // 67
    if (charset) {                                                     // 68
      type = type + "; charset=" + charset;                            // 69
    } else if (type === "application/json") {                          //
      // HACK: Apparently the MIME module does not assume UTF-8 for JSON. :(
      type = type + "; charset=utf-8";                                 // 72
    }                                                                  //
                                                                       //
    var started = false;                                               // 75
    var sawEnd = false;                                                // 76
                                                                       //
    // TODO(perf): Automatically gzip text content? Use Express's 'compress' middleware for this?
    //   Note that nginx will also auto-compress things...             //
                                                                       //
    var headers = {                                                    // 81
      "Content-Type": type,                                            // 82
      "Cache-Control": "public, max-age=" + CACHE_TTL_SECONDS          // 83
    };                                                                 //
                                                                       //
    if (path === "apps/index.json" || path === "apps/index-experimental.json" || path.match(/apps\/[a-z0-9]{52}[.]json/)) {
      // TODO(cleanup): Extra special terrible hack: The app index needs to serve these JSON files
      //   cross-origin. We could almost just make all web sites allow cross-origin since generally
      //   web publishing is meant to publish public content. There is one case where this is
      //   problematic, though: sites behind a firewall. Those sites could potentially be read
      //   by outside sites if CORS is enabled on them. Some day we should make it so apps can
      //   explicitly opt-in to allowing cross-origin queries but that day is not today.
      headers["Access-Control-Allow-Origin"] = "*";                    // 95
    }                                                                  //
                                                                       //
    var stream = {                                                     // 98
      expectSize: function (size) {                                    // 99
        if (!started) {                                                // 100
          started = true;                                              // 101
          response.writeHead(200, _.extend(headers, { "Content-Length": size }));
        }                                                              //
      },                                                               //
                                                                       //
      write: function (data) {                                         // 106
        if (!started) {                                                // 107
          started = true;                                              // 108
          response.writeHead(200, headers);                            // 109
        }                                                              //
                                                                       //
        response.write(data);                                          // 112
      },                                                               //
                                                                       //
      done: function (data) {                                          // 115
        if (!started) {                                                // 116
          started = true;                                              // 117
          response.writeHead(200, _.extend(headers, { "Content-Length": 0 }));
        }                                                              //
                                                                       //
        sawEnd = true;                                                 // 121
        response.end();                                                // 122
      }                                                                //
    };                                                                 //
                                                                       //
    globalBackend.useGrain(grainId, function (supervisor) {            // 126
      return supervisor.getWwwFileHack(path, stream).then(function (result) {
        // jscs:disable disallowQuotedKeysInObjects                    //
        var status = result.status;                                    // 129
        if (status === "file") {                                       // 130
          if (!sawEnd) {                                               // 131
            console.error("getWwwFileHack didn't write file to stream");
            if (!started) {                                            // 133
              response.writeHead(500, {                                // 134
                "Content-Type": "text/plain"                           // 135
              });                                                      //
              response.end("Internal server error");                   // 137
            }                                                          //
                                                                       //
            response.end();                                            // 140
          }                                                            //
        } else if (status === "directory") {                           //
          if (started) {                                               // 143
            console.error("getWwwFileHack wrote to stream for directory");
            if (!sawEnd) {                                             // 145
              response.end();                                          // 146
            }                                                          //
          } else {                                                     //
            response.writeHead(303, {                                  // 149
              "Content-Type": "text/plain",                            // 150
              "Location": "/" + path + "/",                            // 151
              "Cache-Control": "public, max-age=" + CACHE_TTL_SECONDS  // 152
            });                                                        //
            response.end("redirect: /" + path + "/");                  // 154
          }                                                            //
        } else if (status === "notFound") {                            //
          if (started) {                                               // 157
            console.error("getWwwFileHack wrote to stream for notFound");
            if (!sawEnd) {                                             // 159
              response.end();                                          // 160
            }                                                          //
          } else {                                                     //
            response.writeHead(404, {                                  // 163
              "Content-Type": "text/plain"                             // 164
            });                                                        //
            response.end("404 not found: /" + path);                   // 166
          }                                                            //
        } else {                                                       //
          console.error("didn't understand result of getWwwFileHack:", status);
          if (!started) {                                              // 170
            response.writeHead(500, {                                  // 171
              "Content-Type": "text/plain"                             // 172
            });                                                        //
            response.end("Internal server error");                     // 174
          }                                                            //
        }                                                              //
      });                                                              //
    })["catch"](function (err) {                                       //
      console.error(err.stack);                                        // 179
    });                                                                //
  };                                                                   //
}                                                                      //
                                                                       //
function writeErrorResponse(res, err) {                                // 184
  var status = 500;                                                    // 185
  if (err instanceof Meteor.Error && typeof err.error === "number" && err.error >= 400 && err.error < 600) {
    status = err.error;                                                // 188
  } else if (err.httpErrorCode) {                                      //
    status = err.httpErrorCode;                                        // 190
  }                                                                    //
                                                                       //
  // Log errors that are our fault, but not errors that are the client's fault.
  if (status >= 500) console.error(err.stack);                         // 194
                                                                       //
  res.writeHead(status, { "Content-Type": err.htmlMessage ? "text/html" : "text/plain" });
  res.end(err.htmlMessage || err.message);                             // 197
}                                                                      //
                                                                       //
var PNG_MAGIC = new Buffer([0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A]);
var JPEG_MAGIC = new Buffer([0xFF, 0xD8, 0xFF]);                       // 201
                                                                       //
function checkMagic(buf, magic) {                                      // 203
  if (buf.length < magic.length) return false;                         // 204
                                                                       //
  for (var i = 0; i < magic.length; i++) {                             // 206
    if (buf[i] != magic[i]) return false;                              // 207
  }                                                                    //
                                                                       //
  return true;                                                         // 210
}                                                                      //
                                                                       //
function serveSelfTest(req, res) {                                     // 213
  try {                                                                // 214
    if (req.method === "GET" && req.url === "/") {                     // 215
      var content = new Buffer("Self-test OK.");                       // 217
      res.writeHead(200, {                                             // 218
        "Content-Type": "text/plain",                                  // 219
        "Content-Length": content.length,                              // 220
        "Access-Control-Allow-Origin": process.env.ROOT_URL            // 221
      });                                                              //
      res.end(content);                                                // 223
    } else {                                                           //
      res.writeHead(400, { "Content-Type": "text/plain" });            // 225
      res.end("Bad request to self-test subdomain.");                  // 226
      return;                                                          // 227
    }                                                                  //
  } catch (err) {                                                      //
    console.log("WARNING: An error occurred while serving self-test; proceeding anyway:", err);
  }                                                                    //
}                                                                      //
                                                                       //
function serveStaticAsset(req, res) {                                  // 234
  inMeteor(function () {                                               // 235
    if (req.method === "GET") {                                        // 236
      // jscs:disable validateQuoteMarks                               //
      var assetCspHeader = "default-src 'none'; style-src 'unsafe-inline'; sandbox";
      if (req.headers['if-none-match'] === 'permanent') {              // 239
        // Cache never invalidates since we use a new URL for every resource.
        res.writeHead(304, {                                           // 241
          'Cache-Control': 'public, max-age=31536000',                 // 242
          'ETag': 'permanent',                                         // 243
                                                                       //
          // To be safe, send these again, although it shouldn't be necessary.
          'Content-Security-Policy': assetCspHeader,                   // 246
          'Access-Control-Allow-Origin': '*',                          // 247
          'X-Content-Type-Options': 'nosniff'                          // 248
        });                                                            //
        res.end();                                                     // 250
        return;                                                        // 251
      }                                                                //
                                                                       //
      var url = Url.parse(req.url);                                    // 254
      var asset = globalDb.getStaticAsset(url.pathname.slice(1));      // 255
                                                                       //
      if (asset) {                                                     // 257
        var headers = {                                                // 258
          'Content-Type': asset.mimeType,                              // 259
          'Content-Length': asset.content.length,                      // 260
                                                                       //
          // Assets can be cached forever because each one has a unique ID.
          'Cache-Control': 'public, max-age=31536000',                 // 263
                                                                       //
          // Since different resources get different URLs, we can use a static etag.
          'ETag': 'permanent',                                         // 266
                                                                       //
          // Set strict Content-Security-Policy to prevent static assets from executing any script
          // or doing basically anything when browsed to directly. The static assets host is not
          // intended to serve HTML. Mostly, it serves images and javascript -- note that setting
          // the CSP header on Javascript files does not prevent other hosts from voluntarily
          // specifying these scripts in <script> tags.                //
          'Content-Security-Policy': assetCspHeader,                   // 273
                                                                       //
          // Allow any host to fetch these assets. This is safe since requests to this host are
          // totally side-effect-free and the asset ID acts as a capability to prevent loading
          // assets you're not supposed to know about.                 //
          'Access-Control-Allow-Origin': '*',                          // 278
                                                                       //
          // Extra protection against content type trickery.           //
          'X-Content-Type-Options': 'nosniff'                          // 281
        };                                                             //
                                                                       //
        if (asset.encoding) {                                          // 284
          headers['Content-Encoding'] = asset.encoding;                // 285
        }                                                              //
                                                                       //
        res.writeHead(200, headers);                                   // 288
        res.end(asset.content);                                        // 289
      } else {                                                         //
        res.writeHead(404, { 'Content-Type': 'text/plain' });          // 291
        res.end('not found');                                          // 292
      }                                                                //
    } else if (req.method === 'POST') {                                //
      var _ret = (function () {                                        //
        res.setHeader('Access-Control-Allow-Origin', '*');             // 295
                                                                       //
        var url = Url.parse(req.url);                                  // 297
        var purpose = globalDb.fulfillAssetUpload(url.pathname.slice(1));
        if (!purpose) {                                                // 299
          res.writeHead(404, { 'Content-Type': 'text/plain' });        // 300
          res.end('Upload token not found or expired.');               // 301
          return {                                                     // 302
            v: undefined                                               //
          };                                                           //
        }                                                              //
                                                                       //
        var userId = purpose.profilePicture.userId;                    // 305
        var identityId = purpose.profilePicture.identityId;            // 306
        check(userId, String);                                         // 307
        check(identityId, String);                                     // 308
                                                                       //
        var buffers = [];                                              // 310
        var totalSize = 0;                                             // 311
        var done = new Future();                                       // 312
        req.on('data', function (buf) {                                // 313
          totalSize += buf.length;                                     // 314
          if (totalSize <= 64 * 1024) {                                // 315
            buffers.push(buf);                                         // 316
          }                                                            //
        });                                                            //
        req.on('end', done["return"].bind(done));                      // 319
        req.on('error', done["throw"].bind(done));                     // 320
        done.wait();                                                   // 321
                                                                       //
        if (totalSize > 64 * 1024) {                                   // 323
          // TODO(soon): Resize the image ourselves.                   //
          res.writeHead(400, { 'Content-Type': 'text/plain' });        // 325
          res.end('Picture too large; please use an image under 64 KiB.');
          return {                                                     // 327
            v: undefined                                               //
          };                                                           //
        }                                                              //
                                                                       //
        var content = Buffer.concat(buffers);                          // 330
        var type = undefined;                                          // 331
        if (checkMagic(content, PNG_MAGIC)) {                          // 332
          type = 'image/png';                                          // 333
        } else if (checkMagic(content, JPEG_MAGIC)) {                  //
          type = 'image/jpeg';                                         // 335
        } else {                                                       //
          res.writeHead(400, { 'Content-Type': 'text/plain' });        // 337
          res.end('Image must be PNG or JPEG.');                       // 338
          return {                                                     // 339
            v: undefined                                               //
          };                                                           //
        }                                                              //
                                                                       //
        var assetId = globalDb.addStaticAsset({ mimeType: type }, content);
        var old = Meteor.users.findAndModify({                         // 343
          query: { '_id': identityId },                                // 344
          update: { $set: { 'profile.picture': assetId } },            // 345
          fields: { 'profile.picture': 1 }                             // 346
        });                                                            //
                                                                       //
        if (old && old.profile && old.profile.picture) {               // 349
          globalDb.unrefStaticAsset(old.profile.picture);              // 350
        }                                                              //
                                                                       //
        res.writeHead(204, {});                                        // 353
        res.end();                                                     // 354
      })();                                                            //
                                                                       //
      if (typeof _ret === "object") return _ret.v;                     //
    } else if (req.method === 'OPTIONS') {                             //
      var requestedHeaders = req.headers['access-control-request-headers'];
      if (requestedHeaders) {                                          // 357
        res.setHeader('Access-Control-Allow-Headers', requestedHeaders);
      }                                                                //
                                                                       //
      res.writeHead(204, {                                             // 361
        'Access-Control-Allow-Origin': '*',                            // 362
        'Access-Control-Allow-Methods': 'GET, HEAD, POST, OPTIONS',    // 363
        'Access-Control-Max-Age': '3600'                               // 364
      });                                                              //
      res.end();                                                       // 366
    } else {                                                           //
      res.writeHead(405, 'Method Not Allowed', {                       // 368
        'Allow': 'GET, POST, OPTIONS',                                 // 369
        'Content-Type': 'text/plain'                                   // 370
      });                                                              //
      res.end('405 Method Not Allowed: ' + req.method);                // 372
    }                                                                  //
  })["catch"](function (err) {                                         //
    writeErrorResponse(res, err);                                      // 375
  });                                                                  //
}                                                                      //
                                                                       //
Meteor.startup(function () {                                           // 379
                                                                       //
  var meteorUpgradeListeners = WebApp.httpServer.listeners('upgrade');
  WebApp.httpServer.removeAllListeners('upgrade');                     // 382
                                                                       //
  WebApp.httpServer.on('upgrade', function (req, socket, head) {       // 384
    Promise.resolve(undefined).then(function () {                      // 385
      if (!req.headers.host) {                                         // 386
        throw new Meteor.Error(400, 'Missing Host header');            // 387
      } else if (isSandstormShell(req.headers.host.split(':')[0])) {   //
        // Go on to Meteor.                                            //
        for (var ii = 0; ii < meteorUpgradeListeners.length; ++ii) {   // 390
          meteorUpgradeListeners[ii](req, socket, head);               // 391
        }                                                              //
                                                                       //
        return true;                                                   // 394
      } else {                                                         //
        var id = matchWildcardHost(req.headers.host);                  // 396
        if (id) {                                                      // 397
          return tryProxyUpgrade(id, req, socket, head);               // 398
        } else {                                                       //
          return false;                                                // 400
        }                                                              //
      }                                                                //
    }).then(function (handled) {                                       //
      if (!handled) socket.destroy();                                  // 404
    })["catch"](function (err) {                                       //
      console.error('WebSocket event handler failed:', err.stack);     // 406
      socket.destroy();                                                // 407
    });                                                                //
  });                                                                  //
                                                                       //
  // BlackrockPayments is only defined in the Blackrock build of Sandstorm.
  if (global.BlackrockPayments) {                                      // 412
    // Have to check with global, because it could be undefined.       //
    WebApp.rawConnectHandlers.use(BlackrockPayments.makeConnectHandler(globalDb));
  }                                                                    //
                                                                       //
  // This function serves responses on Sandstorm's main HTTP/HTTPS     //
  // port.                                                             //
  var serveMeteorOrStaticPublishing = function (req, res, next) {      // 418
    return dispatchToMeteorOrStaticPublishing(req, res, next, false);  // 419
  };                                                                   //
                                                                       //
  // "Alternate ports" are ports other than the main HTTP or HTTPS     //
  // port. For requests to the shell & grains, we redirect to the main
  // port. For static publishing, we serve it.                         //
  //                                                                   //
  // They are bound to FD #4 and higher.                               //
                                                                       //
  function getNumberOfAlternatePorts() {                               // 428
    var numPorts = process.env.PORT.split(',').length;                 // 429
    var numAlternatePorts = numPorts - 1;                              // 430
    return numAlternatePorts;                                          // 431
  };                                                                   //
                                                                       //
  var canonicalizeShellOrWildcardUrl = function (hostname, url) {      // 434
    // Start with ROOT_URL, apply host & path from inbound URL, then   //
    // redirect.                                                       //
    var targetUrl = Url.parse(process.env.ROOT_URL);                   // 437
                                                                       //
    // Retain the protocol & port from ROOT_URL but use the inbound    //
    // hostname.                                                       //
    targetUrl.host = hostname + ':' + targetUrl.port;                  // 441
                                                                       //
    // The following allows to avoid decoding + re-encoding query      //
    // string parameters, if provided.                                 //
    targetUrl = Url.resolve(Url.format(targetUrl), url);               // 445
    return targetUrl;                                                  // 446
  };                                                                   //
                                                                       //
  var redirectToMeteorOrServeStaticPublishing = function (req, res, next) {
    return dispatchToMeteorOrStaticPublishing(req, res, next, true);   // 450
  };                                                                   //
                                                                       //
  for (var i = 0; i < getNumberOfAlternatePorts(); i++) {              // 453
    // Call createServerForSandstorm() to skip our monkey patching.    //
    var alternatePortServer = Http.createServerForSandstorm(redirectToMeteorOrServeStaticPublishing);
    alternatePortServer.listen({ fd: i + 4 });                         // 456
  }                                                                    //
                                                                       //
  var dispatchToMeteorOrStaticPublishing = function (req, res, next, redirectRatherThanServeShell) {
    if (!req.headers.host) {                                           // 460
      res.writeHead(400, { 'Content-Type': 'text/plain' });            // 461
      res.end('Missing Host header');                                  // 462
      return;                                                          // 463
    }                                                                  //
                                                                       //
    var hostname = req.headers.host.split(':')[0];                     // 466
    if (isSandstormShell(hostname)) {                                  // 467
      // Go on to Meteor, or serve a redirect.                         //
      if (redirectRatherThanServeShell) {                              // 469
        res.writeHead(302, { 'Location': canonicalizeShellOrWildcardUrl(hostname, req.url) });
        res.end();                                                     // 471
        return;                                                        // 472
      } else {                                                         //
        return next();                                                 // 474
      }                                                                //
    }                                                                  //
                                                                       //
    // This is not our main host. See if it's a member of the wildcard.
    var publicIdPromise = undefined;                                   // 479
                                                                       //
    var id = matchWildcardHost(req.headers.host);                      // 481
    if (id) {                                                          // 482
      // Match!                                                        //
      if (redirectRatherThanServeShell) {                              // 484
        res.writeHead(302, { 'Location': canonicalizeShellOrWildcardUrl(hostname, req.url) });
        res.end();                                                     // 486
        return;                                                        // 487
      }                                                                //
                                                                       //
      if (id === 'static') {                                           // 490
        // Static assets domain.                                       //
        serveStaticAsset(req, res);                                    // 492
        return;                                                        // 493
      }                                                                //
                                                                       //
      if (id.match(/^selftest-/)) {                                    // 496
        // Self test domain pattern. Starts w/ hyphen to avoid ambiguity with grain session/static
        // publishing wildcard hosts.                                  //
        serveSelfTest(req, res);                                       // 499
        return;                                                        // 500
      }                                                                //
                                                                       //
      // Try to route the request to a session.                        //
      publicIdPromise = tryProxyRequest(id, req, res).then(function (handled) {
        if (handled) {                                                 // 505
          return null;                                                 // 506
        } else {                                                       //
          return id;                                                   // 508
        }                                                              //
      });                                                              //
    } else {                                                           //
      // Not a wildcard host. Perhaps it is a custom host.             //
      publicIdPromise = lookupPublicIdFromDns(hostname);               // 513
    }                                                                  //
                                                                       //
    publicIdPromise.then(function (publicId) {                         // 516
      if (publicId) {                                                  // 517
        return Promise.resolve(undefined).then(function () {           // 518
          var handler = staticHandlers[publicId];                      // 519
          if (handler) {                                               // 520
            return handler;                                            // 521
          } else {                                                     //
            // We don't have a handler for this publicId, so look it up in the grain DB.
            return inMeteor(function () {                              // 524
              var grain = Grains.findOne({ publicId: publicId }, { fields: { _id: 1 } });
              if (!grain) {                                            // 526
                throw new Meteor.Error(404, 'No such grain for public ID: ' + publicId);
              }                                                        //
                                                                       //
              var grainId = grain._id;                                 // 530
              return staticHandlers[publicId] = wwwHandlerForGrain(grainId);
            });                                                        //
          }                                                            //
        }).then(function (handler) {                                   //
          handler(req, res, function (err) {                           // 535
            if (err) {                                                 // 536
              next(err);                                               // 537
            } else {                                                   //
              res.writeHead(404, { 'Content-Type': 'text/plain' });    // 539
              res.end('404 not found: ' + req.url);                    // 540
            }                                                          //
          });                                                          //
        });                                                            //
      } else {                                                         //
        return Promise.resolve(undefined);                             // 545
      }                                                                //
    })["catch"](function (err) {                                       //
      writeErrorResponse(res, err);                                    // 548
    });                                                                //
  };                                                                   //
                                                                       //
  WebApp.rawConnectHandlers.use(serveMeteorOrStaticPublishing);        // 552
});                                                                    //
                                                                       //
var errorTxtMapping = {};                                              // 555
errorTxtMapping[Dns.NOTFOUND] = '<br>\n' + 'If you were trying to connect this address to a Sandstorm app hosted at this server,<br>\n' + 'you either have not set your DNS TXT records correctly or the DNS cache has not<br>\n' + 'updated yet (may take a while).<br>\n';
errorTxtMapping[Dns.NODATA] = errorTxtMapping[Dns.NOTFOUND];           // 560
errorTxtMapping[Dns.TIMEOUT] = '<br>\n' + 'The DNS query has timed out, which may be a sign of poorly configured DNS on the server.<br>\n';
errorTxtMapping[Dns.CONNREFUSED] = '<br>\n' + 'The DNS server refused the connection, which means either your DNS server is down/unreachable,<br>\n' + 'or the server has misconfigured their DNS.<br>\n';
                                                                       //
function lookupPublicIdFromDns(hostname) {                             // 567
  // Given a hostname, determine its public ID.                        //
  // We look for a TXT record indicating the public ID. Unfortunately, according to spec, a single
  // hostname cannot have both a CNAME and a TXT record, because a TXT lookup on a CNAME'd host
  // should actually be redirected to the CNAME, just like an A lookup would be. In practice lots
  // of DNS software actually allows TXT records on CNAMEs, and it seems to work, but some software
  // does not allow it and it's explicitly disallowed by the spec. Therefore, we instead look for
  // the TXT record on a subdomain.                                    //
  //                                                                   //
  // I also considered having the CNAME itself point to <publicId>.<hostname>, where
  // *.<hostname> is in turn a CNAME for the Sandstorm server. This approach seemed elegant at
  // first, but has a number of problems, the biggest being that it breaks the ability to place a
  // CDN like CloudFlare in front of the site.                         //
                                                                       //
  var cache = dnsCache[hostname];                                      // 581
  if (cache && Date.now() < cache.expiration) {                        // 582
    return Promise.resolve(cache.value);                               // 583
  }                                                                    //
                                                                       //
  return new Promise(function (resolve, reject) {                      // 586
    Dns.resolveTxt('sandstorm-www.' + hostname, function (err, records) {
      if (err) {                                                       // 588
        var errorMsg = errorTxtMapping[err.code] || '';                // 589
        var error = new Error('Error looking up DNS TXT records for host "' + hostname + '": ' + err.message);
        error.htmlMessage = '<p>Error looking up DNS TXT records for host "' + hostname + '": ' + err.message + '<br>\n' + '<br>\n' + 'This Sandstorm server\'s main interface is at: <a href=\'' + process.env.ROOT_URL + '\'>' + process.env.ROOT_URL + '</a><br>\n' + errorMsg + '<br>\n' + 'If you are the server admin and want to use this address as the main interface,<br>\n' + 'edit /opt/sandstorm/sandstorm.conf, modify the BASE_URL setting, and restart.<br>\n' + '<br>\n' + 'If you got here after trying to log in via OAuth (e.g. through Github or Google),<br>\n' + 'the problem is probably that the OAuth callback URL was set wrong. You need to<br>\n' + 'update it through the respective login provider\'s management console.</p>';
        error.httpErrorCode = _.contains(['ENOTFOUND', 'ENODATA'], err.code) ? 404 : 500;
        reject(error);                                                 // 606
      } else if (records.length !== 1) {                               //
        reject(new Error('Host "sandstorm-www.' + hostname + '" must have exactly one TXT record.'));
      } else {                                                         //
        var result = records[0];                                       // 610
        dnsCache[hostname] = { value: result, expiration: Date.now() + DNS_CACHE_TTL };
        resolve(result);                                               // 612
      }                                                                //
    });                                                                //
  });                                                                  //
}                                                                      //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=pre-meteor.js.map
