(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// shared/grain.js                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
// This file implements /grain, i.e. the main view into an app.        //
                                                                       //
var DEFAULT_TITLE = "Sandstorm";                                       // 19
                                                                       //
if (Meteor.isServer) {                                                 // 21
  (function () {                                                       //
    var Crypto = Npm.require("crypto");                                // 22
                                                                       //
    Meteor.publish("grainTopBar", function (grainId) {                 // 24
      check(grainId, String);                                          // 25
                                                                       //
      var result = [Grains.find({                                      // 27
        _id: grainId,                                                  // 29
        $or: [{ userId: this.userId }, { "private": { $ne: true } }]   // 30
      }, {                                                             //
        fields: {                                                      // 35
          title: 1,                                                    // 36
          userId: 1,                                                   // 37
          identityId: 1,                                               // 38
          "private": 1                                                 // 39
        }                                                              //
      })];                                                             //
      if (this.userId) {                                               // 43
        var myIdentityIds = SandstormDb.getUserIdentityIds(globalDb.getUser(this.userId));
        result.push(ApiTokens.find({                                   // 45
          grainId: grainId,                                            // 46
          $or: [{ "owner.user.identityId": { $in: myIdentityIds } }, { identityId: { $in: myIdentityIds } }]
        }));                                                           //
      }                                                                //
                                                                       //
      return result;                                                   // 54
    });                                                                //
                                                                       //
    // We allow users to learn package information about a grain they own.
    // This is used for obtaining icon and app title information for grains
    // you own, which is used in the sidebar. It is not a security/privacy
    // risk since it only exposes this information for grains the user owns.
    Meteor.publish("packageByGrainId", function (grainId) {            // 61
      check(grainId, String);                                          // 62
      var publishThis = [];                                            // 63
      // We need to publish the packageId so that client-side code can
      // find the right package.                                       //
      var thisGrainCursor = Grains.find({                              // 66
        _id: grainId,                                                  // 67
        userId: this.userId                                            // 68
      }, {                                                             //
        fields: { packageId: 1 }                                       // 70
      });                                                              //
      publishThis.push(thisGrainCursor);                               // 72
                                                                       //
      if (thisGrainCursor.count()) {                                   // 74
        var thisGrain = thisGrainCursor.fetch()[0];                    // 75
        var thisPackageCursor = Packages.find({ _id: thisGrain.packageId });
        publishThis.push(thisPackageCursor);                           // 77
      }                                                                //
                                                                       //
      return publishThis;                                              // 80
    });                                                                //
                                                                       //
    Meteor.publish("tokenInfo", function (token) {                     // 83
      // Allows the client side to map a raw token to its entry in ApiTokens, and the additional
      // metadata that it will need to display the app icon and title.  We do not care about making
      // the metadata reactive.                                        //
      check(token, String);                                            // 87
                                                                       //
      var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");
      var apiToken = ApiTokens.findOne({                               // 90
        _id: hashedToken                                               // 91
      }, {                                                             //
        fields: {                                                      // 93
          grainId: 1,                                                  // 94
          owner: 1                                                     // 95
        }                                                              //
      });                                                              //
      if (!apiToken) {                                                 // 98
        this.added("tokenInfo", token, { invalidToken: true });        // 99
      } else {                                                         //
        var grainId = apiToken.grainId;                                // 101
        var grain = Grains.findOne({                                   // 102
          _id: grainId                                                 // 103
        }, {                                                           //
          fields: {                                                    // 105
            packageId: 1,                                              // 106
            appId: 1                                                   // 107
          }                                                            //
        });                                                            //
        if (!grain) {                                                  // 110
          this.added("tokenInfo", token, { invalidToken: true });      // 111
        } else {                                                       //
          if (apiToken.owner && apiToken.owner.user) {                 // 113
            var identity = globalDb.getIdentity(apiToken.owner.user.identityId);
            var metadata = apiToken.owner.user.denormalizedGrainMetadata;
            if (identity && metadata) {                                // 116
              this.added("tokenInfo", token, { identityOwner: _.pick(identity, "_id", "profile"),
                grainId: grainId,                                      // 119
                grainMetadata: metadata });                            // 120
            } else {                                                   //
              this.added("tokenInfo", token, { invalidToken: true });  // 122
            }                                                          //
          } else if (!apiToken.owner || "webkey" in apiToken.owner) {  //
            var pkg = Packages.findOne({ _id: grain.packageId }, { fields: { manifest: 1 } });
            var appTitle = pkg && pkg.manifest && pkg.manifest.appTitle || { defaultText: "" };
            var appIcon = undefined;                                   // 127
            if (pkg && pkg.manifest && pkg.manifest.metadata && pkg.manifest.metadata.icons) {
              var icons = pkg.manifest.metadata.icons;                 // 129
              appIcon = icons.grain || icons.appGrid;                  // 130
            }                                                          //
                                                                       //
            var denormalizedGrainMetadata = {                          // 133
              appTitle: appTitle,                                      // 134
              icon: appIcon,                                           // 135
              appId: appIcon ? undefined : grain.appId                 // 136
            };                                                         //
            this.added("tokenInfo", token, { webkey: true,             // 138
              grainId: grainId,                                        // 140
              grainMetadata: denormalizedGrainMetadata });             // 141
          } else {                                                     //
            this.added("tokenInfo", token, { invalidToken: true });    // 143
          }                                                            //
        }                                                              //
      }                                                                //
                                                                       //
      this.ready();                                                    // 148
      return;                                                          // 149
    });                                                                //
                                                                       //
    Meteor.publish("grainSize", function (grainId) {                   // 152
      // Publish pseudo-collection containing the size of the grain opened in the given session.
      check(grainId, String);                                          // 154
                                                                       //
      var grain = Grains.findOne(grainId);                             // 156
      if (!grain || grain.userId !== this.userId) {                    // 157
        return [];                                                     // 158
      }                                                                //
                                                                       //
      var supervisor = globalBackend.cap().getGrain(this.userId, grainId).supervisor;
                                                                       //
      var _this = this;                                                // 163
      var stopped = false;                                             // 164
      var promise = getGrainSize(supervisor);                          // 165
                                                                       //
      function getNext(oldSize) {                                      // 167
        promise = getGrainSize(supervisor, oldSize);                   // 168
        promise.then(function (size) {                                 // 169
          if (!stopped) {                                              // 170
            if (size !== oldSize) {                                    // 171
              // sometimes there are false alarms                      //
              _this.changed("grainSizes", grainId, { size: size });    // 172
            }                                                          //
                                                                       //
            getNext(size);                                             // 175
          }                                                            //
        }, function (err) {                                            //
                                                                       //
          if (!stopped) {                                              // 179
            if (err.kjType === "disconnected") {                       // 180
              _this.stop();                                            // 181
            } else {                                                   //
              _this.error(err);                                        // 183
            }                                                          //
          }                                                            //
        });                                                            //
      }                                                                //
                                                                       //
      promise.then(function (size) {                                   // 189
        if (!stopped) {                                                // 190
          _this.added("grainSizes", grainId, { size: size });          // 191
          _this.ready();                                               // 192
          getNext(size);                                               // 193
        }                                                              //
      }, function (err) {                                              //
                                                                       //
        if (!stopped) {                                                // 197
          if (err.kjType === "disconnected") {                         // 198
            _this.stop();                                              // 199
          } else {                                                     //
            _this.error(err);                                          // 201
          }                                                            //
        }                                                              //
      });                                                              //
                                                                       //
      _this.onStop(function () {                                       // 206
        stopped = true;                                                // 207
        promise.cancel();                                              // 208
      });                                                              //
    });                                                                //
  })();                                                                //
}                                                                      //
                                                                       //
// GrainSizes is used by grainview.js                                  //
GrainSizes = new Mongo.Collection("grainSizes");                       // 214
// TokenInfo is used by grainview.js                                   //
TokenInfo = new Mongo.Collection("tokenInfo");                         // 216
// Pseudo-collections published above.                                 //
                                                                       //
Meteor.methods({                                                       // 219
  deleteGrain: function (grainId) {                                    // 220
    check(grainId, String);                                            // 221
                                                                       //
    if (this.userId) {                                                 // 223
      var grain = Grains.findOne({ _id: grainId, userId: this.userId });
      if (grain) {                                                     // 225
        Grains.remove(grainId);                                        // 226
        ApiTokens.remove({                                             // 227
          grainId: grainId,                                            // 228
          $or: [{ owner: { $exists: false } }, { owner: { webkey: null } }]
        });                                                            //
        if (grain.lastUsed) {                                          // 234
          DeleteStats.insert({ type: isDemoUser() ? "demoGrain" : "grain",
            lastActive: grain.lastUsed, appId: grain.appId });         // 236
        }                                                              //
                                                                       //
        if (!this.isSimulation) {                                      // 239
          waitPromise(globalBackend.deleteGrain(grainId, this.userId));
          Meteor.call("deleteUnusedPackages", grain.appId);            // 241
        }                                                              //
      }                                                                //
    }                                                                  //
  },                                                                   //
                                                                       //
  forgetGrain: function (grainId, identityId) {                        // 247
    // TODO(cleanup): For now we are ignoring `identityId`, but maybe we should expose finer-grained
    //  forgetting.                                                    //
                                                                       //
    check(grainId, String);                                            // 251
    check(identityId, String);                                         // 252
                                                                       //
    if (!this.userId) {                                                // 254
      throw new Meteor.Error(403, "Must be logged in to forget a grain.");
    }                                                                  //
                                                                       //
    SandstormDb.getUserIdentityIds(Meteor.user()).forEach(function (identityId) {
      ApiTokens.remove({ grainId: grainId, "owner.user.identityId": identityId });
    });                                                                //
  },                                                                   //
                                                                       //
  updateGrainPreferredIdentity: function (grainId, identityId) {       // 263
    check(grainId, String);                                            // 264
    check(identityId, String);                                         // 265
    if (!this.userId) {                                                // 266
      throw new Meteor.Error(403, "Must be logged in.");               // 267
    }                                                                  //
                                                                       //
    var grain = globalDb.getGrain(grainId) || {};                      // 270
    if (!grain.userId === this.userId) {                               // 271
      throw new Meteor.Error(403, "Grain not owned by current user.");
    }                                                                  //
                                                                       //
    Grains.update({ _id: grainId }, { $set: { identityId: identityId } });
  },                                                                   //
                                                                       //
  updateGrainTitle: function (grainId, newTitle, identityId) {         // 278
    check(grainId, String);                                            // 279
    check(newTitle, String);                                           // 280
    check(identityId, String);                                         // 281
    if (this.userId) {                                                 // 282
      var grain = Grains.findOne(grainId);                             // 283
      if (grain) {                                                     // 284
        if (grain.userId === this.userId) {                            // 285
          Grains.update({ _id: grainId, userId: this.userId }, { $set: { title: newTitle } });
        } else {                                                       //
          if (!globalDb.userHasIdentity(this.userId, identityId)) {    // 288
            throw new Meteor.Error(403, "Current user does not have identity " + identityId);
          }                                                            //
                                                                       //
          var token = ApiTokens.findOne({                              // 292
            grainId: grainId,                                          // 293
            objectId: { $exists: false },                              // 294
            "owner.user.identityId": identityId                        // 295
          }, {                                                         //
            sort: { created: 1 }                                       // 297
          });                                                          //
          if (token) {                                                 // 299
            ApiTokens.update(token._id, { $set: { "owner.user.title": newTitle } });
          }                                                            //
        }                                                              //
      }                                                                //
    }                                                                  //
  },                                                                   //
                                                                       //
  privatizeGrain: function (grainId) {                                 // 307
    check(grainId, String);                                            // 308
    if (this.userId) {                                                 // 309
      Grains.update({ _id: grainId, userId: this.userId }, { $set: { "private": true } });
    }                                                                  //
  },                                                                   //
                                                                       //
  inviteUsersToGrain: function (origin, identityId, grainId, title, roleAssignment, contacts, message) {
    var _this2 = this;                                                 //
                                                                       //
    if (!this.isSimulation) {                                          // 316
      var _ret2 = (function () {                                       //
        check(origin, String);                                         // 317
        check(identityId, String);                                     // 318
        check(grainId, String);                                        // 319
        check(title, String);                                          // 320
        check(roleAssignment, roleAssignmentPattern);                  // 321
        check(contacts, [{                                             // 322
          _id: String,                                                 // 324
          isDefault: Match.Optional(Boolean),                          // 325
          profile: Match.ObjectIncluding({                             // 326
            service: String,                                           // 327
            name: String,                                              // 328
            intrinsicName: String                                      // 329
          })                                                           //
        }]);                                                           //
        check(message, { text: String, html: String });                // 333
        if (!_this2.userId) {                                          // 334
          throw new Meteor.Error(403, "Must be logged in to share by email.");
        }                                                              //
                                                                       //
        if (!globalDb.userHasIdentity(_this2.userId, identityId)) {    // 338
          throw new Meteor.Error(403, "Not an identity of the current user: " + identityId);
        }                                                              //
                                                                       //
        if (contacts.length === 0) {                                   // 342
          throw new Meteor.Error(400, "No contacts were provided.");   // 343
        }                                                              //
                                                                       //
        var accountId = _this2.userId;                                 // 346
        var identity = globalDb.getIdentity(identityId);               // 347
        var sharerDisplayName = identity.profile.name;                 // 348
        var outerResult = { successes: [], failures: [] };             // 349
        contacts.forEach(function (contact) {                          // 350
          if (contact.isDefault && contact.profile.service === "email") {
            var emailAddress = contact.profile.intrinsicName;          // 352
            var result = SandstormPermissions.createNewApiToken(globalDb, { identityId: identityId, accountId: accountId }, grainId, "email invitation for " + emailAddress, roleAssignment, { webkey: { forSharing: true } });
            var _url = origin + "/shared/" + result.token;             // 357
            var html = message.html + "<br><br>" + "<a href='" + _url + "' style='display:inline-block;text-decoration:none;" + "font-family:sans-serif;width:200px;min-height:30px;line-height:30px;" + "border-radius:4px;text-align:center;background:#428bca;color:white'>" + "Open Shared Grain</a><div style='font-size:8pt;font-style:italic;color:gray'>" + "Note: If you forward this email to other people, they will be able to access " + "the share as well. To prevent this, remove the button before forwarding.</div>";
            try {                                                      // 365
              SandstormEmail.send({                                    // 366
                to: emailAddress,                                      // 367
                from: "Sandstorm server <no-reply@" + HOSTNAME + ">",  // 368
                subject: sharerDisplayName + " has invited you to join a grain: " + title,
                text: message.text + "\n\nFollow this link to open the shared grain:\n\n" + _url + "\n\nNote: If you forward this email to other people, they will be able to " + "access the share as well. To prevent this, remove the link before forwarding.",
                html: html                                             // 373
              });                                                      //
            } catch (e) {                                              //
              outerResult.failures.push({ contact: contact, error: e.toString() });
            }                                                          //
          } else {                                                     //
            var result = SandstormPermissions.createNewApiToken(globalDb, { identityId: identityId, accountId: accountId }, grainId, "direct invitation to " + contact.profile.intrinsicName, roleAssignment, { user: { identityId: contact._id, title: title } });
            try {                                                      // 383
              var _identity = Meteor.users.findOne({ _id: contact._id });
              var emailAddress = SandstormDb.getVerifiedEmails(_identity)[0];
              var _url2 = origin + "/shared/" + result.token;          // 386
              if (emailAddress) {                                      // 387
                var intrinsicName = contact.profile.intrinsicName;     // 388
                var loginNote = undefined;                             // 389
                if (contact.profile.service === "google") {            // 390
                  loginNote = "Google account with address " + emailAddress;
                } else if (contact.profile.service === "github") {     //
                  loginNote = "Github account with username " + intrinsicName;
                } else if (contact.profile.service === "email") {      //
                  loginNote = "email address " + intrinsicName;        // 395
                } else {                                               //
                  throw new Meteor.Error(500, "Unknown service to email share link.");
                }                                                      //
                                                                       //
                var html = message.html + "<br><br>" + "<a href='" + _url2 + "' style='display:inline-block;text-decoration:none;" + "font-family:sans-serif;width:200px;min-height:30px;line-height:30px;" + "border-radius:4px;text-align:center;background:#428bca;color:white'>" + "Open Shared Grain</a><div style='font-size:8pt;font-style:italic;color:gray'>" + "Note: You will need to log in with your " + loginNote + " to access this grain.";
                SandstormEmail.send({                                  // 407
                  to: emailAddress,                                    // 408
                  from: "Sandstorm server <no-reply@" + HOSTNAME + ">",
                  subject: sharerDisplayName + " has invited you to join a grain: " + title,
                  text: message.text + "\n\nFollow this link to open the shared grain:\n\n" + _url2 + "\n\nNote: You will need to log in with your " + loginNote + " to access this grain.",
                  html: html                                           // 414
                });                                                    //
              } else {                                                 //
                outerResult.failures.push({ contact: contact, warning: "User does not have a " + "verified email, so notification of this share was not sent to them. Please " + "manually share " + _url2 + " with them." });
              }                                                        //
            } catch (e) {                                              //
              outerResult.failures.push({ contact: contact, error: e.toString(),
                warning: "Share succeeded, but there was an error emailing the user. Please " + "manually share " + url + " with them." });
            }                                                          //
          }                                                            //
        });                                                            //
                                                                       //
        return {                                                       // 429
          v: outerResult                                               //
        };                                                             //
      })();                                                            //
                                                                       //
      if (typeof _ret2 === "object") return _ret2.v;                   //
    }                                                                  //
  }                                                                    //
});                                                                    //
                                                                       //
if (Meteor.isClient) {                                                 // 434
  (function () {                                                       //
    Tracker.autorun(function () {                                      // 435
      // We need to keep track of certain data about each grain we can view.
      // TODO(cleanup): Do these in GrainView to avoid spurious resubscribes.
      var grains = globalGrains.get();                                 // 438
      grains.forEach(function (grain) {                                // 439
        grain.depend();                                                // 440
        var grainId = grain.grainId();                                 // 441
        if (grainId) {                                                 // 442
          Meteor.subscribe("grainTopBar", grainId);                    // 443
          if (grain.isOwner()) {                                       // 444
            Meteor.subscribe("packageByGrainId", grainId);             // 445
          }                                                            //
                                                                       //
          var token = grain.token();                                   // 448
          if (token) {                                                 // 449
            Meteor.subscribe("tokenInfo", token);                      // 450
          }                                                            //
        }                                                              //
      });                                                              //
    });                                                                //
                                                                       //
    Template.layout.events({                                           // 456
      "click .incognito-button": function (event) {                    // 457
        console.log(event);                                            // 458
        var grains = globalGrains.get();                               // 459
        var token = event.currentTarget.getAttribute("data-token");    // 460
        if (token) {                                                   // 461
          grains.forEach(function (grain) {                            // 462
            if (grain.token() == token) {                              // 463
              grain.doNotRevealIdentity();                             // 464
            }                                                          //
          });                                                          //
        } else {                                                       //
          console.error("Interstitial prompt answered, but no token present?");
        }                                                              //
      }                                                                //
    });                                                                //
                                                                       //
    var promptNewTitle = function () {                                 // 473
      var grain = getActiveGrain(globalGrains.get());                  // 474
      if (grain) {                                                     // 475
        var _prompt = "Set new title:";                                // 476
        if (!grain.isOwner()) {                                        // 477
          _prompt = "Set a new personal title: (does not change the owner's title for this grain)";
        }                                                              //
                                                                       //
        var title = window.prompt(_prompt, grain.title());             // 481
        if (title) {                                                   // 482
          grain.setTitle(title);                                       // 483
        }                                                              //
      }                                                                //
    };                                                                 //
                                                                       //
    Template.grainTitle.events({                                       // 488
      click: function (event) {                                        // 489
        promptNewTitle();                                              // 490
      },                                                               //
                                                                       //
      keydown: function (event) {                                      // 493
        if (event.keyCode === 13 || event.keyCode === 32) {            // 494
          // Allow space or enter to trigger renaming the grain - Firefox doesn't treat enter on the
          // focused element as click().                               //
          promptNewTitle();                                            // 497
          event.preventDefault();                                      // 498
        }                                                              //
      }                                                                //
    });                                                                //
                                                                       //
    Template.grainDeleteButton.events({                                // 503
      "click button": function (event) {                               // 504
        var grains = globalGrains.get();                               // 505
        var activeIndex = activeGrainIndex(grains);                    // 506
        var activeGrain = grains[activeIndex];                         // 507
        var newActiveIndex = activeIndex == grains.length - 1 ? activeIndex - 1 : activeIndex;
        if (activeGrain.isOwner()) {                                   // 509
          if (window.confirm("Really delete this grain?")) {           // 510
            Meteor.call("deleteGrain", activeGrain.grainId());         // 511
            // TODO: extract globalGrains into a class that has a "close" method for closing the active view
            activeGrain.destroy();                                     // 513
            if (grains.length == 1) {                                  // 514
              globalGrains.set([]);                                    // 515
              Router.go("root"); // will redirect to the right place   // 516
            } else {                                                   //
                grains.splice(activeIndex, 1);                         // 518
                grains[newActiveIndex].setActive(true);                // 519
                globalGrains.set(grains);                              // 520
                Router.go(grains[newActiveIndex].route());             // 521
              }                                                        //
          }                                                            //
        } else {                                                       //
          if (window.confirm("Really forget this grain?")) {           // 525
            var identityId = activeGrain.identityId();                 // 526
            if (identityId) {                                          // 527
              Meteor.call("forgetGrain", activeGrain.grainId(), identityId);
            }                                                          //
            // TODO: extract globalGrains into a class that has a "close" method for closing the active view
            activeGrain.destroy();                                     // 531
            if (grains.length == 1) {                                  // 532
              globalGrains.set([]);                                    // 533
              Router.go("root"); // will redirect to the right place   // 534
            } else {                                                   //
                grains.splice(activeIndex, 1);                         // 536
                grains[newActiveIndex].setActive(true);                // 537
                globalGrains.set(grains);                              // 538
                Router.go(grains[newActiveIndex].route());             // 539
              }                                                        //
          }                                                            //
        }                                                              //
      }                                                                //
    });                                                                //
                                                                       //
    Template.grainDebugLogButton.events({                              // 546
      "click button": function (event) {                               // 547
        this.reset();                                                  // 548
        var activeGrain = getActiveGrain(globalGrains.get());          // 549
        window.open("/grainlog/" + activeGrain.grainId(), "_blank", "menubar=no,status=no,toolbar=no,width=700,height=700");
      }                                                                //
    });                                                                //
                                                                       //
    Template.grainBackupButton.onCreated(function () {                 // 555
      this.isLoading = new ReactiveVar(false);                         // 556
    });                                                                //
                                                                       //
    Template.grainBackupButton.helpers({                               // 559
      isLoading: function () {                                         // 560
        return Template.instance().isLoading.get();                    // 561
      }                                                                //
    });                                                                //
                                                                       //
    Template.grainBackupButton.events({                                // 565
      "click button": function (event, template) {                     // 566
        template.isLoading.set(true);                                  // 567
        this.reset();                                                  // 568
        var activeGrain = getActiveGrain(globalGrains.get());          // 569
        Meteor.call("backupGrain", activeGrain.grainId(), function (err, id) {
          template.isLoading.set(false);                               // 571
          if (err) {                                                   // 572
            alert("Backup failed: " + err); // TODO(someday): make this better UI
          } else {                                                     //
              // Firefox for some reason decides to kill all websockets when we try to download the file
              // by navigating there. So we're left doing a dirty hack to get around the popup blocker.
              var isFirefox = typeof InstallTrigger !== "undefined";   // 577
                                                                       //
              if (isFirefox) {                                         // 579
                var save = document.createElement("a");                // 580
                save.href = "/downloadBackup/" + id;                   // 581
                                                                       //
                save.download = activeGrain.title() + ".zip";          // 583
                var _event = document.createEvent("MouseEvents");      // 584
                _event.initMouseEvent("click", true, false, window, 0, 0, 0, 0, 0, false, false, false, false, 0, null);
                save.dispatchEvent(_event);                            // 589
              } else {                                                 //
                window.location = "/downloadBackup/" + id;             // 591
              }                                                        //
            }                                                          //
        });                                                            //
      }                                                                //
    });                                                                //
                                                                       //
    Template.grainRestartButton.events({                               // 598
      "click button": function (event) {                               // 599
        this.reset();                                                  // 600
        var activeGrain = getActiveGrain(globalGrains.get());          // 601
        var grainId = activeGrain.grainId();                           // 602
                                                                       //
        Meteor.call("shutdownGrain", grainId, function (err) {         // 604
          if (err) {                                                   // 605
            alert("Restart failed: " + err); // TODO(someday): make this better UI
          } else {                                                     //
              var _frames = document.getElementsByClassName("grain-frame");
              for (var i = 0; i < _frames.length; i++) {               // 609
                var frame = _frames[i];                                // 610
                if (frame.dataset.grainid == grainId) {                // 611
                  frame.src = frame.src;                               // 612
                }                                                      //
              }                                                        //
            }                                                          //
        });                                                            //
      }                                                                //
    });                                                                //
                                                                       //
    function selectElementContents(element) {                          // 620
      if (document.body.createTextRange) {                             // 621
        var range = document.body.createTextRange();                   // 622
        range.moveToElementText(element);                              // 623
        range.select();                                                // 624
      } else if (window.getSelection) {                                //
        var selection = window.getSelection();                         // 626
        var range = document.createRange();                            // 627
        range.selectNodeContents(element);                             // 628
        selection.removeAllRanges();                                   // 629
        selection.addRange(range);                                     // 630
      }                                                                //
    }                                                                  //
                                                                       //
    function selectTargetContents(event) {                             // 634
      event.preventDefault();                                          // 635
      selectElementContents(event.currentTarget);                      // 636
    }                                                                  //
                                                                       //
    Template.grainApiTokenPopup.events({                               // 639
      "click .copy-me": selectTargetContents,                          // 640
      "focus .copy-me": selectTargetContents,                          // 641
      "submit .newApiToken": function (event) {                        // 642
        event.preventDefault();                                        // 643
        var activeGrain = getActiveGrain(globalGrains.get());          // 644
        var grainId = activeGrain.grainId();                           // 645
        activeGrain.setGeneratedApiToken("pending");                   // 646
        var roleList = document.getElementById("api-token-role");      // 647
        // TODO(cleanup): avoid using global ids; select a child of the current template instead
        var assignment = { allAccess: null };                          // 649
        if (roleList && roleList.selectedIndex > 0) {                  // 650
          assignment = { roleId: roleList.selectedIndex - 1 };         // 651
        }                                                              //
                                                                       //
        Meteor.call("newApiToken", { identityId: activeGrain.identityId() }, grainId, document.getElementById("api-token-petname").value, assignment, { webkey: { forSharing: false } }, function (error, result) {
          if (error) {                                                 // 658
            activeGrain.setGeneratedApiToken(undefined);               // 659
            window.alert("Failed to create token.\n" + error);         // 660
            console.error(error.stack);                                // 661
          } else {                                                     //
            activeGrain.setGeneratedApiToken(result.endpointUrl + "#" + result.token);
          }                                                            //
        });                                                            //
      },                                                               //
                                                                       //
      "click #resetApiToken": function (event) {                       // 668
        var activeGrain = getActiveGrain(globalGrains.get());          // 669
        activeGrain.setGeneratedApiToken(undefined);                   // 670
      },                                                               //
                                                                       //
      "click button.revoke-token": function (event) {                  // 673
        Meteor.call("updateApiToken", event.currentTarget.getAttribute("data-token-id"), { revoked: true });
      },                                                               //
                                                                       //
      "click .token-petname": function (event) {                       // 678
        // TODO(soon): Find a less-annoying way to get this input, perhaps by allowing the user
        //   to edit the petname in place.                             //
        var petname = window.prompt("Set new label:", this.petname);   // 681
        if (petname) {                                                 // 682
          Meteor.call("updateApiToken", event.currentTarget.getAttribute("data-token-id"), { petname: petname });
        }                                                              //
      }                                                                //
    });                                                                //
                                                                       //
    Template.grainSharePopup.events({                                  // 689
      "click .copy-me": selectTargetContents,                          // 690
      "focus .copy-me": selectTargetContents,                          // 691
      "click #share-grain-popup-closer": function (event) {            // 692
        Session.set("show-share-grain", false);                        // 693
      },                                                               //
                                                                       //
      "click button.who-has-access": function (event, instance) {      // 696
        event.preventDefault();                                        // 697
        var closer = globalTopbar.addItem({                            // 698
          name: "who-has-access",                                      // 699
          template: Template.whoHasAccess,                             // 700
          popupTemplate: Template.whoHasAccessPopup,                   // 701
          data: new ReactiveVar(instance.data),                        // 702
          startOpen: true,                                             // 703
          onDismiss: function () {                                     // 704
            return "remove";                                           // 705
          }                                                            //
        });                                                            //
      },                                                               //
                                                                       //
      "click #privatize-grain": function (event) {                     // 710
        Meteor.call("privatizeGrain", getActiveGrain(globalGrains.get()).grainId());
      },                                                               //
                                                                       //
      "click a.open-non-anonymously": function (event) {               // 714
        event.preventDefault();                                        // 715
        globalTopbar.closePopup();                                     // 716
        getActiveGrain(globalGrains.get()).reset();                    // 717
        getActiveGrain(globalGrains.get()).openSession();              // 718
      }                                                                //
    });                                                                //
                                                                       //
    Template.shareWithOthers.onRendered(function () {                  // 722
      this.find("[role=tab]").focus();                                 // 723
    });                                                                //
                                                                       //
    var activateTargetTab = function (event, instance) {               // 726
      // Deactivate all tabs and all tab panels.                       //
      instance.findAll("ul[role=tablist]>li[role=tab]").forEach(function (element) {
        element.setAttribute("aria-selected", false);                  // 729
      });                                                              //
                                                                       //
      instance.findAll(".tabpanel").forEach(function (element) {       // 732
        element.setAttribute("aria-hidden", true);                     // 733
      });                                                              //
                                                                       //
      // Activate the tab header the user selected.                    //
      event.currentTarget.setAttribute("aria-selected", true);         // 737
      // Show the corresponding tab panel.                             //
      var idToShow = event.currentTarget.getAttribute("aria-controls");
      var tabPanelToShow = instance.find("#" + idToShow);              // 740
      tabPanelToShow.setAttribute("aria-hidden", false);               // 741
    };                                                                 //
                                                                       //
    Template.shareWithOthers.events({                                  // 744
      "click #send-invite-tab-header": activateTargetTab,              // 745
      "click #shareable-link-tab-header": activateTargetTab,           // 746
      "keydown [role=tab]": function (event, template) {               // 747
        if (event.keyCode == 38 || event.keyCode == 40) {              // 748
          // up and down arrows                                        //
          event.preventDefault();                                      // 749
        }                                                              //
                                                                       //
        var $focus = $(template.find(":focus"));                       // 752
        var $items = template.$("[role=tab]:visible");                 // 753
        var focusIndex = $items.index($focus);                         // 754
        var newFocusIndex = undefined;                                 // 755
        if (event.keyCode == 37) {                                     // 756
          // left arrow                                                //
          event.preventDefault();                                      // 757
          newFocusIndex = focusIndex - 1;                              // 758
          if (newFocusIndex == -1) {                                   // 759
            newFocusIndex = $items.length - 1;                         // 760
          }                                                            //
        } else if (event.keyCode == 39) {                              //
          // right arrow                                               //
          event.preventDefault();                                      // 763
          newFocusIndex = focusIndex + 1;                              // 764
          if (newFocusIndex >= $items.length) {                        // 765
            newFocusIndex = 0;                                         // 766
          }                                                            //
        } else if (event.keyCode == 13) {                              //
          // Enter key                                                 //
          event.preventDefault();                                      // 769
          activateTargetTab(event, template);                          // 770
        }                                                              //
                                                                       //
        if (newFocusIndex != null) {                                   // 773
          $items.attr("tabindex", "-1");                               // 774
          var $newFocus = $($items[newFocusIndex]);                    // 775
          $newFocus.attr("tabindex", "0");                             // 776
          $newFocus.focus();                                           // 777
        }                                                              //
      }                                                                //
    });                                                                //
                                                                       //
    Template.shareableLinkTab.events({                                 // 782
      "change .share-token-role": function (event, instance) {         // 783
        var success = instance.completionState.get().success;          // 784
        if (success) {                                                 // 785
          var roleList = event.target;                                 // 786
          var assignment = undefined;                                  // 787
          if (roleList) {                                              // 788
            assignment = { roleId: roleList.selectedIndex };           // 789
          } else {                                                     //
            assignment = { none: null };                               // 791
          }                                                            //
                                                                       //
          Meteor.call("updateApiToken", success.id, { roleAssignment: assignment }, function (error) {
            if (error) {                                               // 795
              console.error(error.stack);                              // 796
            }                                                          //
          });                                                          //
        }                                                              //
      },                                                               //
                                                                       //
      "change .label": function (event, instance) {                    // 802
        var success = instance.completionState.get().success;          // 803
        if (success) {                                                 // 804
          var label = event.target.value;                              // 805
          Meteor.call("updateApiToken", success.id, { petname: label }, function (error) {
            if (error) {                                               // 807
              console.error(error.stack);                              // 808
            }                                                          //
          });                                                          //
        }                                                              //
      },                                                               //
                                                                       //
      "submit form.new-share-token": function (event, instance) {      // 814
        event.preventDefault();                                        // 815
        if (!instance.completionState.get().clear) {                   // 816
          return;                                                      // 817
        }                                                              //
                                                                       //
        var currentGrain = getActiveGrain(globalGrains.get());         // 820
        var grainId = currentGrain.grainId();                          // 821
        var roleList = event.target.getElementsByClassName("share-token-role")[0];
        var assignment = undefined;                                    // 823
        if (roleList) {                                                // 824
          assignment = { roleId: roleList.selectedIndex };             // 825
        } else {                                                       //
          assignment = { none: null };                                 // 827
        }                                                              //
                                                                       //
        instance.completionState.set({ pending: true });               // 830
        Meteor.call("newApiToken", { identityId: currentGrain.identityId() }, grainId, event.target.getElementsByClassName("label")[0].value, assignment, { webkey: { forSharing: true } }, function (error, result) {
          if (error) {                                                 // 835
            console.error(error.stack);                                // 836
          } else {                                                     //
            result.url = getOrigin() + "/shared/" + result.token;      // 838
            instance.completionState.set({ success: result });         // 839
            // On the next render, .copy-me will exist, and we should focus it then.
            Meteor.defer(function () {                                 // 841
              var element = instance.find(".copy-me");                 // 842
              element.focus();                                         // 843
              selectElementContents(element);                          // 844
            });                                                        //
          }                                                            //
        });                                                            //
      },                                                               //
                                                                       //
      "click .reset-share-token": function (event, instance) {         // 850
        instance.completionState.set({ clear: true });                 // 851
        instance.find("form").reset();                                 // 852
        instance.find("form option[data-default-selected=true]").selected = true;
      }                                                                //
    });                                                                //
                                                                       //
    Template.grainPowerboxOfferPopup.events({                          // 857
      "click button.dismiss": function (event) {                       // 858
        var data = Template.instance().data.get();                     // 859
        data.onDismiss();                                              // 860
      },                                                               //
                                                                       //
      "click .copy-me": selectTargetContents,                          // 863
      "focus .copy-me": selectTargetContents                           // 864
    });                                                                //
                                                                       //
    Template.grainSharePopup.helpers({                                 // 867
      incognito: function () {                                         // 868
        return !Accounts.getCurrentIdentityId();                       // 869
      },                                                               //
                                                                       //
      currentTokenUrl: function () {                                   // 872
        return getOrigin() + "/shared/" + getActiveGrain(globalGrains.get()).token();
      },                                                               //
                                                                       //
      currentGrain: function () {                                      // 876
        return getActiveGrain(globalGrains.get());                     // 877
      }                                                                //
    });                                                                //
                                                                       //
    Template.grainView.helpers({                                       // 881
      unpackedGrainState: function () {                                // 882
        return mapGrainStateToTemplateData(this);                      // 883
      },                                                               //
                                                                       //
      identityPickerData: function () {                                // 886
        var identities = SandstormDb.getUserIdentityIds(Meteor.user()).map(function (id) {
          var identity = Meteor.users.findOne({ _id: id });            // 888
          if (identity) {                                              // 889
            SandstormDb.fillInProfileDefaults(identity);               // 890
            SandstormDb.fillInIntrinsicName(identity);                 // 891
            SandstormDb.fillInPictureUrl(identity);                    // 892
            return identity;                                           // 893
          }                                                            //
        });                                                            //
                                                                       //
        var grain = getActiveGrain(globalGrains.get());                // 897
        return { identities: identities,                               // 898
          onPicked: function (identityId) {                            // 899
            grain.revealIdentity(identityId);                          // 899
          } };                                                         //
      }                                                                //
    });                                                                //
                                                                       //
    Template.grain.helpers({                                           // 903
      currentGrain: function () {                                      // 904
        return getActiveGrain(globalGrains.get());                     // 905
      },                                                               //
                                                                       //
      isOwner: function () {                                           // 908
        var current = getActiveGrain(globalGrains.get());              // 909
        return current && current.isOwner();                           // 910
      },                                                               //
                                                                       //
      grainSize: function () {                                         // 913
        var current = getActiveGrain(globalGrains.get());              // 914
        var sizeBytes = current && current.size();                     // 915
        if (sizeBytes !== undefined) {                                 // 916
          return "(" + prettySize(sizeBytes) + ")";                    // 917
        } else {                                                       //
          return "";                                                   // 919
        }                                                              //
      },                                                               //
                                                                       //
      setGrainWindowTitle: function () {                               // 923
        var current = getActiveGrain(globalGrains.get());              // 924
        if (current) {                                                 // 925
          current.updateDocumentTitle();                               // 926
        }                                                              //
      },                                                               //
                                                                       //
      displayWebkeyButton: function () {                               // 930
        // TODO: figure out what this should do                        //
        return Meteor.userId() || !this.oldSharingModel;               // 932
      },                                                               //
                                                                       //
      showPowerboxOffer: function () {                                 // 935
        var current = getActiveGrain(globalGrains.get());              // 936
        return current && current.showPowerboxOffer();                 // 937
      },                                                               //
                                                                       //
      powerboxOfferData: function () {                                 // 940
        var current = getActiveGrain(globalGrains.get());              // 941
        return current && current.powerboxOfferData();                 // 942
      },                                                               //
                                                                       //
      showPowerboxRequest: function () {                               // 945
        var current = getActiveGrain(globalGrains.get());              // 946
        return current && current.showPowerboxRequest();               // 947
      },                                                               //
                                                                       //
      powerboxRequestData: function () {                               // 950
        var current = getActiveGrain(globalGrains.get());              // 951
        return current && current.powerboxRequestData();               // 952
      },                                                               //
                                                                       //
      cancelPowerboxRequest: function () {                             // 955
        return function () {                                           // 956
          var current = getActiveGrain(globalGrains.get());            // 957
          current.setPowerboxRequest(undefined);                       // 958
          return "remove";                                             // 959
        };                                                             //
      }                                                                //
    });                                                                //
                                                                       //
    Template.grainTitle.helpers({                                      // 964
      title: function () {                                             // 965
        var grain = getActiveGrain(globalGrains.get());                // 966
        return grain && grain.title() || "Untitled grain";             // 967
      }                                                                //
    });                                                                //
                                                                       //
    Template.grainApiTokenPopup.helpers({                              // 971
      displayToken: function () {                                      // 972
        return !this.revoked && !this.expiresIfUnused && !this.parentToken;
      },                                                               //
                                                                       //
      existingTokens: function () {                                    // 976
        var current = getActiveGrain(globalGrains.get());              // 977
        return current && ApiTokens.find({                             // 978
          grainId: current.grainId(),                                  // 979
          identityId: current.identityId(),                            // 980
          forSharing: { $ne: true },                                   // 981
          $or: [{ owner: { webkey: null } }, { owner: { $exists: false } }],
          expiresIfUnused: null                                        // 986
        });                                                            //
      },                                                               //
                                                                       //
      generatedApiToken: function () {                                 // 990
        var current = getActiveGrain(globalGrains.get());              // 991
        return current && current.generatedApiToken();                 // 992
      },                                                               //
                                                                       //
      generatedApiTokenPending: function () {                          // 995
        var current = getActiveGrain(globalGrains.get());              // 996
        return (current && current.generatedApiToken()) == "pending";  // 997
      },                                                               //
                                                                       //
      viewInfo: function () {                                          // 1000
        var current = getActiveGrain(globalGrains.get());              // 1001
        return current && current.viewInfo();                          // 1002
      }                                                                //
    });                                                                //
                                                                       //
    Template.whoHasAccessPopup.onCreated(function () {                 // 1006
      var _this = this;                                                // 1007
      this.subscribe("contactProfiles");                               // 1008
      var currentGrain = getActiveGrain(globalGrains.get());           // 1009
      _this.identityId = currentGrain.identityId();                    // 1010
      _this.grainId = currentGrain.grainId();                          // 1011
      _this.transitiveShares = new ReactiveVar(null);                  // 1012
      _this.downstreamTokensById = new ReactiveVar({});                // 1013
      this.resetTransitiveShares = function () {                       // 1014
        Meteor.call("transitiveShares", _this.identityId, _this.grainId, function (error, downstream) {
          if (error) {                                                 // 1017
            console.error(error.stack);                                // 1018
          } else {                                                     //
            (function () {                                             //
              var downstreamTokensById = {};                           // 1020
              var sharesByRecipient = {};                              // 1021
              downstream.forEach(function (token) {                    // 1022
                downstreamTokensById[token._id] = token;               // 1023
                if (Match.test(token.owner, { user: Match.ObjectIncluding({ identityId: String }) })) {
                  var recipient = token.owner.user.identityId;         // 1025
                  if (!sharesByRecipient[recipient]) {                 // 1026
                    sharesByRecipient[recipient] = {                   // 1027
                      recipient: recipient,                            // 1028
                      dedupedShares: [],                               // 1029
                      allShares: []                                    // 1030
                    };                                                 //
                  }                                                    //
                                                                       //
                  sharesByRecipient[recipient].allShares.push(token);  // 1034
                  var dedupedShares = sharesByRecipient[recipient].dedupedShares;
                  if (!dedupedShares.some(function (share) {           // 1036
                    return share.identityId === token.identityId;      //
                  })) {                                                //
                    dedupedShares.push(token);                         // 1037
                  }                                                    //
                }                                                      //
              });                                                      //
                                                                       //
              var result = _.values(sharesByRecipient);                // 1042
              if (result.length == 0) {                                // 1043
                result = { empty: true };                              // 1044
              }                                                        //
                                                                       //
              _this.transitiveShares.set(result);                      // 1047
              _this.downstreamTokensById.set(downstreamTokensById);    // 1048
            })();                                                      //
          }                                                            //
        });                                                            //
      };                                                               //
                                                                       //
      this.resetTransitiveShares();                                    // 1053
    });                                                                //
                                                                       //
    Template.whoHasAccessPopup.events({                                // 1056
      "change .share-token-role": function (event, instance) {         // 1057
        var roleList = event.target;                                   // 1058
        var assignment = undefined;                                    // 1059
        if (roleList) {                                                // 1060
          assignment = { roleId: roleList.selectedIndex };             // 1061
        } else {                                                       //
          assignment = { none: null };                                 // 1063
        }                                                              //
                                                                       //
        Meteor.call("updateApiToken", roleList.getAttribute("data-token-id"), { roleAssignment: assignment }, function (error) {
          if (error) {                                                 // 1068
            console.error(error.stack);                                // 1069
          }                                                            //
        });                                                            //
      },                                                               //
                                                                       //
      "click button.revoke-token": function (event, instance) {        // 1074
        Meteor.call("updateApiToken", event.currentTarget.getAttribute("data-token-id"), { revoked: true });
        instance.resetTransitiveShares();                              // 1077
      },                                                               //
                                                                       //
      "click button.revoke-access": function (event, instance) {       // 1080
        var recipient = event.currentTarget.getAttribute("data-recipient");
        var transitiveShares = instance.transitiveShares.get();        // 1082
        var tokensById = instance.downstreamTokensById.get();          // 1083
        var recipientShares = _.findWhere(transitiveShares, { recipient: recipient });
        var currentIdentityId = Accounts.getCurrentIdentityId();       // 1085
        var recipientTokens = _.where(recipientShares.allShares, { identityId: currentIdentityId });
                                                                       //
        // Two cases:                                                  //
        // 1. All of the tokens are direct shares. Easy. Just revoke them.
        // 2. Some of the links are child tokens. For each, we walk up the chain of parents to the
        //    root. Then we walk downwards to see whether any other identities would be immediately
        //    affected by revoking those roots. We then collect that data and display it in a
        //    confirmation dialog.                                     //
                                                                       //
        var directTokens = [];                                         // 1095
        // Direct shares from the current identity to the recipient.   //
                                                                       //
        var rootTokens = [];                                           // 1098
        // Roots of chains of child tokens starting at the current identity and leading to the
        // recipient.                                                  //
                                                                       //
        recipientTokens.forEach(function (token) {                     // 1102
          if (token.parentToken) {                                     // 1103
            var currentToken = token;                                  // 1104
            do {                                                       // 1105
              currentToken = tokensById[token.parentToken];            // 1106
            } while (currentToken.parentToken);                        //
                                                                       //
            rootTokens.push(currentToken);                             // 1109
          } else {                                                     //
            directTokens.push(token);                                  // 1111
          }                                                            //
        });                                                            //
                                                                       //
        if (rootTokens.length > 0) {                                   // 1115
          var _ret5 = (function () {                                   //
            // Some of the shares are not direct.                      //
                                                                       //
            var tokensByParent = {};                                   // 1118
            for (var id in babelHelpers.sanitizeForInObject(tokensById)) {
              var token = tokensById[id];                              // 1120
              if (token.parentToken) {                                 // 1121
                if (!tokensByParent[token.parentToken]) {              // 1122
                  tokensByParent[token.parentToken] = [];              // 1123
                }                                                      //
                                                                       //
                tokensByParent[token.parentToken].push(token);         // 1126
              }                                                        //
            }                                                          //
                                                                       //
            var otherAffectedIdentities = {};                          // 1130
            var tokenStack = rootTokens.slice(0);                      // 1131
            while (tokenStack.length > 0) {                            // 1132
              var current = tokenStack.pop();                          // 1133
              if (Match.test(current.owner, { user: Match.ObjectIncluding({ identityId: String }) })) {
                if (current.owner.user.identityId != recipient) {      // 1136
                  otherAffectedIdentities[current.owner.user.identityId] = true;
                }                                                      //
              } else {                                                 //
                if (tokensByParent[current._id]) {                     // 1140
                  var children = tokensByParent[current._id];          // 1141
                  children.forEach(function (child) {                  // 1142
                    tokenStack.push(child);                            // 1143
                  });                                                  //
                }                                                      //
              }                                                        //
            }                                                          //
                                                                       //
            var recipientIdentity = ContactProfiles.findOne({ _id: recipient });
            var recipientName = recipientIdentity && recipientIdentity.profile.name || "Unknown User";
            var singular = rootTokens.length == 1;                     // 1152
                                                                       //
            var tokenLabels = _.pluck(rootTokens, "petname").map(function (petname) {
              return petname || "Unlabeled Link";                      //
            }).map(function (petname) {                                //
              return "\"" + petname + "\"";                            //
            }).join(", ");                                             //
                                                                       //
            var confirmText = "This will revoke the following sharing link" + (singular ? "" : "s") + ":\n\n    " + tokenLabels + "\n\n";
                                                                       //
            var othersNote = "(No signed-in user other than " + recipientName + " has opened " + (singular ? "this link" : "these links") + " yet.)";
            if (Object.keys(otherAffectedIdentities).length > 0) {     // 1164
              var othersNames = Object.keys(otherAffectedIdentities).map(function (identityId) {
                return ContactProfiles.findOne({ _id: identityId });   //
              }).map(function (identity) {                             //
                return identity && identity.profile.name || "Unknown User";
              }).join(", ");                                           //
              othersNote = (singular ? "This link has" : "These links have") + " also been opened by:\n\n    " + othersNames;
            }                                                          //
                                                                       //
            if (window.confirm(confirmText + othersNote)) {            // 1173
              rootTokens.forEach(function (token) {                    // 1174
                Meteor.call("updateApiToken", token._id, { revoked: true });
              });                                                      //
            } else {                                                   //
              // Cancel.                                               //
              return {                                                 // 1179
                v: undefined                                           //
              };                                                       //
            }                                                          //
          })();                                                        //
                                                                       //
          if (typeof _ret5 === "object") return _ret5.v;               //
        }                                                              //
                                                                       //
        directTokens.forEach(function (token) {                        // 1183
          Meteor.call("updateApiToken", token._id, { revoked: true });
        });                                                            //
                                                                       //
        instance.resetTransitiveShares();                              // 1187
      },                                                               //
                                                                       //
      "click .token-petname": function (event, instance) {             // 1190
        // TODO(soon): Find a less-annoying way to get this input, perhaps by allowing the user
        //   to edit the petname in place.                             //
        var petname = window.prompt("Set new label:", this.petname);   // 1193
        if (petname) {                                                 // 1194
          Meteor.call("updateApiToken", event.currentTarget.getAttribute("data-token-id"), { petname: petname });
        }                                                              //
                                                                       //
        instance.resetTransitiveShares();                              // 1199
      }                                                                //
    });                                                                //
                                                                       //
    function isEmptyPermissionSet(permissionSet) {                     // 1203
      if (!permissionSet) {                                            // 1204
        return true;                                                   // 1205
      }                                                                //
                                                                       //
      for (var ii = 0; ii < permissionSet.length; ++ii) {              // 1208
        if (permissionSet[ii]) {                                       // 1209
          return false;                                                // 1210
        }                                                              //
      }                                                                //
                                                                       //
      return true;                                                     // 1214
    }                                                                  //
                                                                       //
    Template.whoHasAccessPopup.helpers({                               // 1217
      existingShareTokens: function () {                               // 1218
        if (Meteor.userId()) {                                         // 1219
          return ApiTokens.find({                                      // 1220
            grainId: Template.instance().grainId,                      // 1221
            identityId: Template.instance().identityId,                // 1222
            forSharing: true,                                          // 1223
            $or: [{ owner: { webkey: null } }, { owner: { $exists: false } }]
          }).fetch();                                                  //
        }                                                              //
      },                                                               //
                                                                       //
      isCurrentIdentity: function () {                                 // 1232
        if (this.identityId === Accounts.getCurrentIdentityId()) {     // 1233
          return true;                                                 // 1234
        }                                                              //
      },                                                               //
                                                                       //
      getPetname: function () {                                        // 1238
        if (this.petname) {                                            // 1239
          return this.petname;                                         // 1240
        } else {                                                       //
          return "Unlabeled Link";                                     // 1242
        }                                                              //
      },                                                               //
                                                                       //
      displayName: function (identityId) {                             // 1246
        var identity = ContactProfiles.findOne({ _id: identityId });   // 1247
        if (!identity) {                                               // 1248
          identity = Meteor.users.findOne({ _id: identityId });        // 1249
        }                                                              //
                                                                       //
        if (identity) {                                                // 1252
          return identity.profile.name;                                // 1253
        } else {                                                       //
          return "Unknown User (" + identityId.slice(0, 16) + ")";     // 1255
        }                                                              //
      },                                                               //
                                                                       //
      transitiveShares: function () {                                  // 1259
        return Template.instance().transitiveShares.get();             // 1260
      },                                                               //
                                                                       //
      indexedRoles: function () {                                      // 1263
        var result = [];                                               // 1264
        var instance = Template.instance();                            // 1265
        var currentGrain = getActiveGrain(globalGrains.get());         // 1266
        var roles = currentGrain.viewInfo().roles;                     // 1267
        for (var ii = 0; ii < roles.length; ++ii) {                    // 1268
          result.push({ idx: ii, title: roles[ii].title, verbPhrase: roles[ii].verbPhrase });
        }                                                              //
                                                                       //
        return result;                                                 // 1272
      },                                                               //
                                                                       //
      roleText: function () {                                          // 1275
        if (this.verbPhrase) {                                         // 1276
          return this.verbPhrase.defaultText;                          // 1277
        } else {                                                       //
          return "is " + this.title.defaultText;                       // 1279
        }                                                              //
      },                                                               //
                                                                       //
      hasCustomRole: function (token) {                                // 1283
        var role = token.roleAssignment;                               // 1284
        if ("roleId" in role && isEmptyPermissionSet(role.addPermissions) && isEmptyPermissionSet(role.removePermissions)) {
          return false;                                                // 1288
        }                                                              //
                                                                       //
        return true;                                                   // 1291
      },                                                               //
                                                                       //
      hasCurrentRole: function (token) {                               // 1294
        var role = token.roleAssignment;                               // 1295
        if ("roleId" in role && role.roleId == this.idx && isEmptyPermissionSet(role.addPermissions) && isEmptyPermissionSet(role.removePermissions)) {
          return true;                                                 // 1299
        }                                                              //
                                                                       //
        return false;                                                  // 1302
      },                                                               //
                                                                       //
      displayToken: function () {                                      // 1305
        return !this.revoked && !this.expiresIfUnused && !this.parentToken;
      },                                                               //
                                                                       //
      viewInfo: function () {                                          // 1309
        var activeGrain = getActiveGrain(globalGrains.get());          // 1310
        return activeGrain && activeGrain.viewInfo();                  // 1311
      }                                                                //
    });                                                                //
                                                                       //
    Template.selectRole.helpers({                                      // 1315
      roleText: function () {                                          // 1316
        if (this.verbPhrase) {                                         // 1317
          return this.verbPhrase.defaultText;                          // 1318
        } else {                                                       //
          return "is " + this.title.defaultText;                       // 1320
        }                                                              //
      }                                                                //
    });                                                                //
                                                                       //
    Template.shareableLinkTab.onCreated(function () {                  // 1325
      this.completionState = new ReactiveVar({ clear: true });         // 1326
    });                                                                //
                                                                       //
    Template.emailInviteTab.onCreated(function () {                    // 1329
      this.completionState = new ReactiveVar({ clear: true });         // 1330
      this.contacts = new ReactiveVar([]);                             // 1331
    });                                                                //
                                                                       //
    Template.shareableLinkTab.helpers({                                // 1334
      completionState: function () {                                   // 1335
        var instance = Template.instance();                            // 1336
        return instance.completionState.get();                         // 1337
      }                                                                //
    });                                                                //
    Template.emailInviteTab.helpers({                                  // 1340
      completionState: function () {                                   // 1341
        var instance = Template.instance();                            // 1342
        return instance.completionState.get();                         // 1343
      },                                                               //
                                                                       //
      contacts: function () {                                          // 1346
        return Template.instance().contacts;                           // 1347
      }                                                                //
    });                                                                //
                                                                       //
    Template.emailInviteTab.events({                                   // 1351
      "submit form.email-invite": function (event, instance) {         // 1352
        event.preventDefault();                                        // 1353
        return false;                                                  // 1354
      },                                                               //
                                                                       //
      "click form.email-invite button": function (event, instance) {   // 1357
        event.preventDefault();                                        // 1358
        if (!instance.completionState.get().clear) {                   // 1359
          return;                                                      // 1360
        }                                                              //
                                                                       //
        var grainId = instance.data.grainId;                           // 1363
        var title = instance.data.title;                               // 1364
                                                                       //
        var roleList = instance.find(".share-token-role");             // 1366
        var assignment = undefined;                                    // 1367
        if (roleList) {                                                // 1368
          assignment = { roleId: roleList.selectedIndex };             // 1369
        } else {                                                       //
          assignment = { none: null };                                 // 1371
        }                                                              //
                                                                       //
        var message = instance.find(".personal-message").value;        // 1374
        instance.completionState.set({ pending: true });               // 1375
                                                                       //
        var currentGrain = getActiveGrain(globalGrains.get());         // 1377
                                                                       //
        // HTML-escape the message.                                    //
        var div = document.createElement("div");                       // 1380
        div.appendChild(document.createTextNode(message));             // 1381
        var htmlMessage = div.innerHTML.replace(/\n/g, "<br>");        // 1382
                                                                       //
        var contacts = instance.contacts.get();                        // 1384
        var emails = instance.find("input.emails");                    // 1385
        var emailsValue = emails.value;                                // 1386
        if (emailsValue) {                                             // 1387
          if (emailsValue.match(/.+\@.+\..+/)) {                       // 1388
            contacts.push({                                            // 1389
              _id: emailsValue,                                        // 1390
              profile: {                                               // 1391
                service: "email",                                      // 1392
                name: emailsValue,                                     // 1393
                intrinsicName: emailsValue,                            // 1394
                pictureUrl: "/email.svg"                               // 1395
              },                                                       //
              isDefault: true                                          // 1397
            });                                                        //
            instance.contacts.set(contacts);                           // 1399
            emails.value = "";                                         // 1400
          } else {                                                     //
            instance.completionState.set({ error: "Invalid text in contact input box: " + emailsValue });
            return;                                                    // 1404
          }                                                            //
        }                                                              //
                                                                       //
        Meteor.call("inviteUsersToGrain", getOrigin(), currentGrain.identityId(), grainId, title, assignment, contacts, { text: message, html: htmlMessage }, function (error, result) {
          if (error) {                                                 // 1411
            instance.completionState.set({ error: error.toString() });
          } else {                                                     //
            if (result.failures.length > 0) {                          // 1414
              var _message = "Failed to send to: ";                    // 1415
              for (var ii = 0; ii < result.failures.length; ++ii) {    // 1416
                console.error(result.failures[ii].error);              // 1417
                if (ii != 0) {                                         // 1418
                  _message += ", ";                                    // 1419
                }                                                      //
                                                                       //
                _message += result.failures[ii].contact.profile.name;  // 1422
                var warning = result.failures[ii].warning;             // 1423
                if (warning) {                                         // 1424
                  _message += ". " + warning;                          // 1425
                }                                                      //
              }                                                        //
                                                                       //
              instance.completionState.set({ error: _message });       // 1429
            } else {                                                   //
              instance.completionState.set({ success: "success" });    // 1431
            }                                                          //
          }                                                            //
        });                                                            //
      },                                                               //
                                                                       //
      "click .reset-invite": function (event, instance) {              // 1437
        instance.contacts.set([]);                                     // 1438
        instance.completionState.set({ clear: true });                 // 1439
        instance.find("form").reset();                                 // 1440
        instance.find("form option[data-default-selected=true]").selected = true;
      },                                                               //
                                                                       //
      "click .start-over-invite": function (event, instance) {         // 1444
        instance.completionState.set({ clear: true });                 // 1445
      }                                                                //
    });                                                                //
                                                                       //
    Template.grainPowerboxOfferPopup.helpers({                         // 1449
      powerboxOfferUrl: function () {                                  // 1450
        var offer = Template.instance().data.get().offer;              // 1451
        return offer && offer.url;                                     // 1452
      }                                                                //
    });                                                                //
                                                                       //
    // Send a keep-alive for each grain every now and then.            //
    Meteor.setInterval(function () {                                   // 1457
      var grains = globalGrains.get();                                 // 1458
      if (!grains) return;                                             // 1459
                                                                       //
      // Meteor has an exponential backoff of up to 5 minutes for reconnect. This is unnacceptable
      // for us, since we rely on Sessions being re-established in under 60s.
      if (Meteor.status().status === "waiting") {                      // 1463
        console.log("Sandstorm is trying to reconnect...");            // 1464
        Meteor.reconnect();                                            // 1465
      }                                                                //
                                                                       //
      grains.forEach(function (grain) {                                // 1468
        if (grain.sessionId()) {                                       // 1469
          // TODO(soon):  Investigate what happens in background tabs.  Maybe arrange to re-open the
          //   app if it dies while in the background.                 //
          Meteor.call("keepSessionAlive", grain.sessionId(), function (error, result) {
            // Sessions will automatically resume if possible, otherwise they will refresh.
          });                                                          //
        }                                                              //
      });                                                              //
    }, 60000);                                                         //
                                                                       //
    // Message handler for Sandstorm's client-side postMessage API.    //
    Meteor.startup(function () {                                       // 1480
      var messageListener = function (event) {                         // 1481
        if (event.origin === getOrigin()) {                            // 1482
          // Meteor likes to postMessage() to itself sometimes, so we ignore these messages.
          return;                                                      // 1484
        }                                                              //
                                                                       //
        // Look up the grain index of which grain this postmessage came from, so we can map behavior
        // into that particular grain's state                          //
        var grains = globalGrains.get();                               // 1489
        var senderGrainIndex = grainOriginToIndex(grains, event.origin);
        if (senderGrainIndex == -1) {                                  // 1491
          // We got a postMessage from an origin that is not a grain we currently believe is open.
          // Ignore it. (It may be aimed at some other message listener registered elsewhere...)
          return;                                                      // 1494
        }                                                              //
                                                                       //
        var senderGrain = grains[senderGrainIndex];                    // 1497
                                                                       //
        if (event.data.setPath || event.data.setPath === "") {         // 1499
          var path = event.data.setPath || "/";                        // 1500
          check(path, String);                                         // 1501
          check(path.charAt(0), "/");                                  // 1502
          // TODO(security): More sanitization of this path. E.g. reject "/../../".
          senderGrain.setPath(path);                                   // 1504
        } else if (event.data.startSharing) {                          //
          // Allow the current grain to request that the "Share Access" menu be shown.
          // Only show this popup if no other popup is currently active.
          // TODO(security): defend against malicious apps spamming this call, blocking all other UI.
          var currentGrain = getActiveGrain(globalGrains.get());       // 1509
          if (senderGrain === currentGrain && !globalTopbar.isPopupOpen()) {
            globalTopbar.openPopup("share");                           // 1511
          }                                                            //
        } else if (event.data.setTitle || event.data.setTitle === "") {
          senderGrain.setFrameTitle(event.data.setTitle);              // 1514
        } else if (event.data.renderTemplate) {                        //
          var _ret6 = (function () {                                   //
            // Request creation of a single-use template with a privileged API token.
            // Why?  Apps should not be able to obtain capabilities-as-keys to
            // themselves directly, because those can be leaked through an
            // arbitrary bit stream or covert channel.  However, apps often need a
            // way to provide instructions to users to copy/paste with some
            // privileged token contained within.  By providing this templating in
            // the platform, we can ensure that the token is only visible to the
            // shell's origin.                                         //
            var call = event.data.renderTemplate;                      // 1524
            check(call, Object);                                       // 1525
            var rpcId = call.rpcId;                                    // 1526
            try {                                                      // 1527
              check(call, { rpcId: String, template: String, petname: Match.Optional(String),
                roleAssignment: Match.Optional(roleAssignmentPattern),
                forSharing: Match.Optional(Boolean) });                // 1530
            } catch (error) {                                          //
              event.source.postMessage({ rpcId: rpcId, error: error.toString() }, event.origin);
              return {                                                 // 1533
                v: undefined                                           //
              };                                                       //
            }                                                          //
                                                                       //
            var template = call.template;                              // 1536
            var petname = "connected external app";                    // 1537
            if (call.petname) {                                        // 1538
              petname = call.petname;                                  // 1539
            }                                                          //
                                                                       //
            var assignment = { allAccess: null };                      // 1542
            if (call.roleAssignment) {                                 // 1543
              assignment = call.roleAssignment;                        // 1544
            }                                                          //
                                                                       //
            var forSharing = call.forSharing ? call.forSharing : false;
            // Tokens expire by default in 5 minutes from generation date
            var selfDestructDuration = 5 * 60 * 1000;                  // 1549
                                                                       //
            var provider = undefined;                                  // 1551
            if (Router.current().route.getName() === "shared") {       // 1552
              provider = { rawParentToken: Router.current().params.token };
            } else {                                                   //
              provider = { identityId: senderGrain.identityId() };     // 1555
            }                                                          //
                                                                       //
            var owner = {                                              // 1558
              webkey: {                                                // 1559
                forSharing: forSharing,                                // 1560
                expiresIfUnusedDuration: selfDestructDuration          // 1561
              }                                                        //
            };                                                         //
                                                                       //
            Meteor.call("newApiToken", provider, senderGrain.grainId(), petname, assignment, owner, function (error, result) {
              if (error) {                                             // 1567
                event.source.postMessage({ rpcId: rpcId, error: error.toString() }, event.origin);
              } else {                                                 //
                var tokenId = result.token;                            // 1570
                // Generate random key id2.                            //
                var id2 = Random.secret();                             // 1572
                // Store apitoken id1 and template in session storage in the offer
                // template namespace under key id2.                   //
                var key = "offerTemplate" + id2;                       // 1575
                var renderedTemplate = template.replace(/\$API_TOKEN/g, tokenId).replace(/\$API_HOST/g, makeWildcardHost("api"));
                sessionStorage.setItem(key, JSON.stringify({           // 1578
                  token: tokenId,                                      // 1579
                  renderedTemplate: renderedTemplate,                  // 1580
                  expires: Date.now() + selfDestructDuration           // 1581
                }));                                                   //
                sessionStorage.setItem("apiHost", makeWildcardHost("api"));
                                                                       //
                // Send message to event.source with URL containing id2
                templateLink = window.location.origin + "/offer-template.html#" + id2;
                event.source.postMessage({ rpcId: rpcId, uri: templateLink }, event.origin);
              }                                                        //
            });                                                        //
          })();                                                        //
                                                                       //
          if (typeof _ret6 === "object") return _ret6.v;               //
        } else if (event.data.powerboxRequest) {                       //
          var powerboxRequest = event.data.powerboxRequest;            // 1592
          check(powerboxRequest, {                                     // 1593
            // TODO: be more strict, and check more fields, once the test apps are more conformant
            rpcId: Match.Any,                                          // 1595
            query: Match.Optional([Object]), // TODO: be more precise, and also make non-optional once tests are updated to provide a query
            saveLabel: Match.Optional(String)                          // 1597
          });                                                          //
                                                                       //
          var powerboxRequestInfo = {                                  // 1600
            source: event.source,                                      // 1601
            origin: event.origin,                                      // 1602
            // We'll need these to post a reply message                //
                                                                       //
            rpcId: powerboxRequest.rpcId,                              // 1605
            query: powerboxRequest.query,                              // 1606
            saveLabel: powerboxRequest.saveLabel,                      // 1607
            // These data come from the grain                          //
                                                                       //
            grainId: senderGrain.grainId(),                            // 1610
            identityId: senderGrain.identityId(),                      // 1611
            // Attach grain context to the request.                    //
                                                                       //
            onCompleted: function () {                                 // 1614
              globalTopbar.closePopup();                               // 1614
            }                                                          //
          };                                                           //
          // Allow the grain to close the popup when we've completed the request.
          var requestContext = new SandstormPowerboxRequest(globalDb, powerboxRequestInfo);
          senderGrain.setPowerboxRequest(requestContext);              // 1618
          globalTopbar.openPopup("request");                           // 1619
        } else {                                                       //
          console.log("postMessage from app not understood: ", event.data);
          console.log(event);                                          // 1622
        }                                                              //
      };                                                               //
                                                                       //
      window.addEventListener("message", messageListener, false);      // 1626
    });                                                                //
  })();                                                                //
}                                                                      //
                                                                       //
if (Meteor.isClient) {                                                 // 1630
  function maybeScrollLog() {                                          // 1631
    var elem = document.getElementById("grainLog");                    // 1632
    if (elem) {                                                        // 1633
      // The log already exists. It's about to be updated. Check if it's scrolled to the bottom
      // before the update.                                            //
      if (elem.scrollHeight - elem.scrollTop === elem.clientHeight) {  // 1636
        // Indeed, so we want to scroll it back to the bottom after the update.
        Tracker.afterFlush(function () {                               // 1638
          scrollLogToBottom(elem);                                     // 1638
        });                                                            //
      }                                                                //
    } else {                                                           //
      // No element exists yet, but it's probably about to be created, in which case we definitely
      // want to scroll it.                                            //
      Tracker.afterFlush(function () {                                 // 1643
        var elem2 = document.getElementById("grainLog");               // 1644
        if (elem2) scrollLogToBottom(elem2);                           // 1645
      });                                                              //
    }                                                                  //
  }                                                                    //
                                                                       //
  function scrollLogToBottom(elem) {                                   // 1650
    elem.scrollTop = elem.scrollHeight;                                // 1651
  }                                                                    //
}                                                                      //
                                                                       //
function makeGrainIdActive(grainId) {                                  // 1655
  var grains = globalGrains.get();                                     // 1656
  for (var i = 0; i < grains.length; i++) {                            // 1657
    var grain = grains[i];                                             // 1658
    if (grain.grainId() === grainId) {                                 // 1659
      if (!grain.isActive()) {                                         // 1660
        grain.setActive(true);                                         // 1661
      }                                                                //
    } else {                                                           //
      if (grain.isActive()) {                                          // 1664
        grain.setActive(false);                                        // 1665
      }                                                                //
    }                                                                  //
  }                                                                    //
}                                                                      //
                                                                       //
function getActiveGrain(grains) {                                      // 1671
  var idx = activeGrainIndex(grains);                                  // 1672
  return idx == -1 ? undefined : grains[idx];                          // 1673
}                                                                      //
                                                                       //
function activeGrainIndex(grains) {                                    // 1676
  for (var i = 0; i < grains.length; i++) {                            // 1677
    if (grains[i].isActive()) {                                        // 1678
      return i;                                                        // 1679
    }                                                                  //
  }                                                                    //
                                                                       //
  return -1;                                                           // 1683
}                                                                      //
                                                                       //
function grainOriginToIndex(grains, origin) {                          // 1686
  for (var i = 0; i < grains.length; i++) {                            // 1687
    if (grains[i].origin() === origin) {                               // 1688
      return i;                                                        // 1689
    }                                                                  //
  }                                                                    //
                                                                       //
  return -1;                                                           // 1693
}                                                                      //
                                                                       //
function grainIdToIndex(grains, grainId) {                             // 1696
  for (var i = 0; i < grains.length; i++) {                            // 1697
    var grain = grains[i];                                             // 1698
    if (grains[i].grainId() === grainId) {                             // 1699
      return i;                                                        // 1700
    }                                                                  //
  }                                                                    //
                                                                       //
  return -1;                                                           // 1704
}                                                                      //
                                                                       //
function mapGrainStateToTemplateData(grainState) {                     // 1707
  var templateData = {                                                 // 1708
    grainId: grainState.grainId(),                                     // 1709
    active: grainState.isActive(),                                     // 1710
    title: grainState.title(),                                         // 1711
    error: grainState.error(),                                         // 1712
    appOrigin: grainState.origin(),                                    // 1713
    hasNotLoaded: !grainState.hasLoaded(),                             // 1714
    sessionId: grainState.sessionId(),                                 // 1715
    originalPath: grainState._originalPath,                            // 1716
    interstitial: grainState.shouldShowInterstitial(),                 // 1717
    token: grainState.token(),                                         // 1718
    viewInfo: grainState.viewInfo()                                    // 1719
  };                                                                   //
  return templateData;                                                 // 1721
}                                                                      //
                                                                       //
GrainLog = new Mongo.Collection("grainLog");                           // 1724
// Pseudo-collection created by subscribing to "grainLog", implemented in proxy.js.
                                                                       //
Router.map(function () {                                               // 1727
  this.route("apps", {                                                 // 1728
    path: "/apps",                                                     // 1729
    waitOn: function () {                                              // 1730
      return globalSubs;                                               // 1730
    },                                                                 //
                                                                       //
    data: function () {                                                // 1732
      if (!this.ready()) return;                                       // 1733
      if (!Meteor.userId() && !Meteor.loggingIn()) {                   // 1734
        Router.go("root", {}, { replaceState: true });                 // 1735
      }                                                                //
                                                                       //
      return new SandstormAppList(globalDb, globalQuotaEnforcer);      // 1738
    }                                                                  //
  });                                                                  //
  this.route("newGrainRedirect", {                                     // 1741
    // The path /grain/new used to be where you would go to create new grains.
    // Its functionality has been superceded by the apps route, so redirect in
    // case anyone has the old link saved somewhere.                   //
    path: "/grain/new",                                                // 1745
    onBeforeAction: function () {                                      // 1746
      Router.go("apps", {}, { replaceState: true });                   // 1747
    }                                                                  //
  });                                                                  //
  this.route("grains", {                                               // 1750
    path: "/grain",                                                    // 1751
    waitOn: function () {                                              // 1752
      return globalSubs;                                               // 1752
    },                                                                 //
                                                                       //
    data: function () {                                                // 1754
      if (!this.ready()) return;                                       // 1755
      if (!Meteor.userId() && !Meteor.loggingIn()) {                   // 1756
        Router.go("root", {}, { replaceState: true });                 // 1757
      }                                                                //
                                                                       //
      return new SandstormGrainListPage(globalDb, globalQuotaEnforcer);
    }                                                                  //
  });                                                                  //
  this.route("grain", {                                                // 1763
    path: "/grain/:grainId/:path(.*)?",                                // 1764
    loadingTemplate: "loadingNoMessage",                               // 1765
                                                                       //
    waitOn: function () {                                              // 1767
      return globalSubs;                                               // 1768
    },                                                                 //
                                                                       //
    onBeforeAction: function () {                                      // 1771
      // Run this hook only once.                                      //
      if (this.state.get("beforeActionHookRan")) {                     // 1773
        return this.next();                                            // 1773
      }                                                                //
                                                                       //
      this.state.set("beforeActionHookRan", true);                     // 1775
      var grainId = this.params.grainId;                               // 1776
      var path = "/" + (this.params.path || "") + (this.originalUrl.match(/[#?].*$/) || "");
      var grains = globalGrains.get();                                 // 1778
      var grainIndex = grainIdToIndex(grains, grainId);                // 1779
      if (grainIndex == -1) {                                          // 1780
        // The element we need to attach our Blaze view to may not exist yet.
        // In that case, defer creating the GrainView until we're sure it's
        // had a chance to render.                                     //
        var openView = (function () {                                  // 1784
          function openView() {                                        // 1784
            var mainContentElement = document.querySelector("body>.main-content");
            if (mainContentElement) {                                  // 1786
              var _grains = globalGrains.get();                        // 1787
              var grainToOpen = new GrainView(grainId, path, undefined, mainContentElement);
              grainToOpen.openSession();                               // 1789
              grainIndex = _grains.push(grainToOpen) - 1;              // 1790
              globalGrains.set(_grains);                               // 1791
              makeGrainIdActive(grainId);                              // 1792
            } else {                                                   //
              Meteor.defer(openView);                                  // 1794
            }                                                          //
          }                                                            //
                                                                       //
          return openView;                                             //
        })();                                                          //
                                                                       //
        openView();                                                    // 1798
      } else {                                                         //
        makeGrainIdActive(grainId);                                    // 1800
      }                                                                //
                                                                       //
      this.next();                                                     // 1803
    },                                                                 //
                                                                       //
    onStop: function () {                                              // 1806
      this.state.set("beforeActionHookRan", undefined);                // 1807
      globalGrains.get().forEach(function (grain) {                    // 1808
        if (grain.isActive()) {                                        // 1809
          grain.setActive(false);                                      // 1810
        }                                                              //
      });                                                              //
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("/shared/:token/:path(.*)?", {                            // 1816
    name: "shared",                                                    // 1817
    template: "grain",                                                 // 1818
    loadingTemplate: "loadingNoMessage",                               // 1819
                                                                       //
    waitOn: function () {                                              // 1821
      return [Meteor.subscribe("devPackages"), Meteor.subscribe("tokenInfo", this.params.token), Meteor.subscribe("grainsMenu")];
    },                                                                 //
                                                                       //
    // This subscription gives us the data we need for deciding whether to automatically reveal
    // our identity.                                                   //
    // TODO(soon): Subscribe to contacts instead.                      //
    onBeforeAction: function () {                                      // 1833
      // Run this hook only once. We could accomplish the same thing by using the `onRun()` hook
      // and waiting for `this.ready()`, but for some reason that fails in the case when a user
      // logs in while visiting a /shared/ link.                       //
      if (this.state.get("beforeActionHookRan")) {                     // 1837
        return this.next();                                            // 1837
      }                                                                //
                                                                       //
      this.state.set("beforeActionHookRan", true);                     // 1839
                                                                       //
      var token = this.params.token;                                   // 1841
      var path = "/" + (this.params.path || "") + (this.originalUrl.match(/[#?].*$/) || "");
      var hash = this.params.hash;                                     // 1843
                                                                       //
      var tokenInfo = TokenInfo.findOne({ _id: token });               // 1845
      if (tokenInfo && tokenInfo.invalidToken) {                       // 1846
        this.state.set("invalidToken", true);                          // 1847
      } else {                                                         //
        (function () {                                                 //
          var grainId = tokenInfo.grainId;                             // 1849
          var grains = globalGrains.get();                             // 1850
          var grainIndex = grainIdToIndex(grains, grainId);            // 1851
          if (grainIndex == -1) {                                      // 1852
            var openView = (function () {                              // 1853
              function openView() {                                    // 1853
                var mainContentElement = document.querySelector("body>.main-content");
                if (mainContentElement) {                              // 1855
                  var _grains2 = globalGrains.get();                   // 1856
                  var grainToOpen = new GrainView(grainId, path, tokenInfo, mainContentElement);
                  grainToOpen.openSession();                           // 1858
                  grainIndex = _grains2.push(grainToOpen) - 1;         // 1859
                  globalGrains.set(_grains2);                          // 1860
                  makeGrainIdActive(grainId);                          // 1861
                } else {                                               //
                  Meteor.defer(openView);                              // 1863
                }                                                      //
              }                                                        //
                                                                       //
              return openView;                                         //
            })();                                                      //
                                                                       //
            openView();                                                // 1867
          } else {                                                     //
            makeGrainIdActive(grainId);                                // 1869
          }                                                            //
        })();                                                          //
      }                                                                //
                                                                       //
      this.next();                                                     // 1873
    },                                                                 //
                                                                       //
    action: function () {                                              // 1876
      if (this.state.get("invalidToken")) {                            // 1877
        this.render("invalidToken", { data: { token: this.params.token } });
      } else {                                                         //
        this.render();                                                 // 1880
      }                                                                //
    },                                                                 //
                                                                       //
    onStop: function () {                                              // 1884
      this.state.set("beforeActionHookRan", undefined);                // 1885
      this.state.set("invalidToken", undefined);                       // 1886
      globalGrains.get().forEach(function (grain) {                    // 1887
        if (grain.isActive()) {                                        // 1888
          grain.setActive(false);                                      // 1889
        }                                                              //
      });                                                              //
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("grainLog", {                                             // 1895
    path: "/grainlog/:grainId",                                        // 1896
    layoutTemplate: "noLayout",                                        // 1897
                                                                       //
    waitOn: function () {                                              // 1899
      return [Meteor.subscribe("grainTopBar", this.params.grainId), Meteor.subscribe("grainLog", this.params.grainId)];
    },                                                                 //
                                                                       //
    data: function () {                                                // 1906
      if (this.ready()) {                                              // 1907
        maybeScrollLog();                                              // 1908
        var grain = Grains.findOne(this.params.grainId);               // 1909
        return {                                                       // 1910
          title: grain ? grain.title : "(deleted grain)",              // 1911
          // jscs:disable requireCamelCaseOrUpperCaseIdentifiers       //
          html: AnsiUp.ansi_to_html(GrainLog.find({}, { $sort: { _id: 1 } }).map(function (entry) {
            return entry.text;                                         // 1914
          }).join(""), { use_classes: true })                          //
        };                                                             //
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  // jscs:enable requireCamelCaseOrUpperCaseIdentifiers                //
  this.route("appDetails", {                                           // 1922
    path: "/apps/:appId",                                              // 1923
    template: "appDetails",                                            // 1924
    waitOn: function () {                                              // 1925
      return globalSubs.concat(Meteor.subscribe("appIndex", this.params.appId));
    },                                                                 //
                                                                       //
    data: function () {                                                // 1930
      return new SandstormAppDetails(globalDb, globalQuotaEnforcer, this.params.appId);
    }                                                                  //
  });                                                                  //
});                                                                    //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=grain.js.map
