(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// shared/shell.js                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
// This file implements the common shell components such as the top bar.
// It also covers the root page.                                       //
                                                                       //
var getNamesFromIdentityIds = function (identityIds) {                 // 20
  check(identityIds, [String]);                                        // 21
  if (identityIds.length === 0) {                                      // 22
    return [];                                                         // 23
  }                                                                    //
                                                                       //
  var identities = Meteor.users.find({                                 // 26
    _id: { $in: identityIds }                                          // 27
  });                                                                  //
  return identities.map(function (identity) {                          // 29
    return { name: identity.profile.name };                            // 30
  });                                                                  //
};                                                                     //
                                                                       //
browseHome = function () {                                             // 34
  Router.go("root");                                                   // 35
};                                                                     //
                                                                       //
getOrigin = function () {                                              // 38
  return document.location.protocol + "//" + document.location.host;   // 39
};                                                                     //
                                                                       //
if (Meteor.isClient) {                                                 // 42
  // Subscribe to basic grain information first and foremost, since    //
  // without it we might e.g. redirect to the wrong place on login.    //
  globalSubs = [Meteor.subscribe("grainsMenu"), Meteor.subscribe("userPackages"), Meteor.subscribe("devPackages"), Meteor.subscribe("credentials"), Meteor.subscribe("accountIdentities")];
                                                                       //
  Tracker.autorun(function () {                                        // 53
    var me = Meteor.user();                                            // 54
    if (me) {                                                          // 55
      if (me.profile) {                                                // 56
        Meteor.subscribe("identityProfile", me._id);                   // 57
      }                                                                //
                                                                       //
      if (me.loginIdentities) {                                        // 60
        me.loginIdentities.forEach(function (identity) {               // 61
          Meteor.subscribe("identityProfile", identity.id);            // 62
        });                                                            //
      }                                                                //
                                                                       //
      if (me.nonloginIdentities) {                                     // 66
        me.nonloginIdentities.forEach(function (identity) {            // 67
          Meteor.subscribe("identityProfile", identity.id);            // 68
        });                                                            //
      }                                                                //
    }                                                                  //
  });                                                                  //
}                                                                      //
                                                                       //
if (Meteor.isServer) {                                                 // 75
  // Set up browser policy.                                            //
  //                                                                   //
  // Note that by default (when the browser-policy package is added), Content-Security-Policy will
  // already be set to same-origin-only for everything except XHR and WebSocket. Eval is disabled,
  // but inline script are enabled.                                    //
  //                                                                   //
  // TODO(security): Consider disallowing inline scripts. Currently this forces Meteor to do an
  //   extra round trip on startup to fetch server settings. That seems like something that could
  //   be fixed in Meteor (e.g. embed the settings as JSON text rather than script). Startup time
  //   is incredibly important, and Meteor's templating system (which we use to render all our
  //   HTML) already does a good job of protecting us from script injection, so right now I think
  //   we should favor avoiding the round trip.                        //
  //                                                                   //
  // TODO(someday): Detect when an app tries to navigate off-site using CSP's violation reporting
  //   feature. Ask the user whether they want to open the link in a new tab. This is an annoying
  //   prompt, but if we just open the tab directly then apps can trivially leak by opening tabs
  //   with encoded URLs to an evil server. Although, this attack would be very detectable to the
  //   user, so maybe it's not a big deal...                           //
                                                                       //
  BrowserPolicy.framing.disallow(); // Disallow framing of the UI.     // 95
  BrowserPolicy.content.allowFrameOrigin(getWildcardOrigin());         // 96
                                                                       //
  // Allow anything to be loaded from the static asset host.           //
  var Url = Npm.require("url");                                        // 99
  var staticAssetHost = Url.parse(process.env.ROOT_URL).protocol + "//" + globalDb.makeWildcardHost("static");
  BrowserPolicy.content.allowImageOrigin(staticAssetHost);             // 102
  BrowserPolicy.content.allowScriptOrigin(staticAssetHost);            // 103
  BrowserPolicy.content.allowFontOrigin(staticAssetHost);              // 104
  BrowserPolicy.content.allowConnectOrigin(staticAssetHost);           // 105
                                                                       //
  Meteor.publish("grainsMenu", function () {                           // 107
    var _this = this;                                                  //
                                                                       //
    if (this.userId) {                                                 // 108
      if (Meteor.settings["public"].quotaEnabled) {                    // 109
        (function () {                                                 //
          // Hack: Fire off an asynchronous update to the user's storage usage whenever they open the
          //   front page.                                             //
          // TODO(someday): Implement the ability to reactively subscribe to storage usage from the
          //   back-end?                                               //
          var userId = _this.userId;                                   // 114
          globalBackend.cap().getUserStorageUsage(userId).then(function (results) {
            inMeteor(function () {                                     // 116
              Meteor.users.update(userId, { $set: { storageUsage: parseInt(results.size) } });
            });                                                        //
          })["catch"](function (err) {                                 //
            if (err.kjType !== "unimplemented") {                      // 120
              console.error(err.stack);                                // 121
            }                                                          //
          });                                                          //
        })();                                                          //
      }                                                                //
                                                                       //
      var identityIds = SandstormDb.getUserIdentityIds(globalDb.getUser(this.userId));
      return [UserActions.find({ userId: this.userId }), Grains.find({ userId: this.userId }), ApiTokens.find({ "owner.user.identityId": { $in: identityIds } })];
    } else {                                                           //
      return [];                                                       // 133
    }                                                                  //
  });                                                                  //
                                                                       //
  Meteor.publish("sessions", function (sessionId) {                    // 137
    // sessionId itself should be secret enough, but they are also not meant to be shared, so as
    // a backup we only publish the session to its owner. Note that `userId` can be null if the
    // user is not logged in or is using incognito mode.               //
    check(sessionId, String);                                          // 141
    return Sessions.find({ _id: sessionId, $or: [{ userId: this.userId }, { userId: null }] });
  });                                                                  //
                                                                       //
  Meteor.publish("devPackages", function () {                          // 145
    return DevPackages.find();                                         // 146
  });                                                                  //
                                                                       //
  Meteor.publish("notifications", function () {                        // 149
    return Notifications.find({ userId: this.userId }, { fields: { timestamp: 1, text: 1, grainId: 1, userId: 1, isUnread: 1, appUpdates: 1,
        admin: 1, referral: 1, mailingListBonus: 1 } });               // 152
  });                                                                  //
                                                                       //
  Meteor.publish("notificationGrains", function (notificationIds) {    // 155
    // Since publishes can't be reactive, we leave it to the client to subscribe to both
    // "notifications" and "notificationGrains" reactively.            //
    check(notificationIds, [String]);                                  // 158
    var notifications = Notifications.find({                           // 159
      _id: { $in: notificationIds },                                   // 160
      userId: this.userId                                              // 161
    }, {                                                               //
      fields: { grainId: 1 }                                           // 163
    });                                                                //
                                                                       //
    var grainIds = notifications.map(function (row) {                  // 166
      return row.grainId;                                              // 167
    });                                                                //
                                                                       //
    return Grains.find({ _id: { $in: grainIds } }, { fields: { title: 1 } });
  });                                                                  //
                                                                       //
  Meteor.publish("hasUsers", function () {                             // 173
    var _this2 = this;                                                 //
                                                                       //
    // Publish pseudo-collection which tells the client if there are any users at all.
    //                                                                 //
    // TODO(cleanup):  This seems overcomplicated.  Does Meteor have a better way?
    var cursor = Meteor.users.find();                                  // 177
    if (cursor.count() > 0) {                                          // 178
      this.added("hasUsers", "hasUsers", { hasUsers: true });          // 179
    } else {                                                           //
      (function () {                                                   //
        var handle = cursor.observeChanges({                           // 181
          added: function (id) {                                       // 182
            _this2.added("hasUsers", "hasUsers", { hasUsers: true });  // 183
            handle.stop();                                             // 184
            handle = null;                                             // 185
          }                                                            //
        });                                                            //
        _this2.onStop(function () {                                    // 188
          if (handle) handle.stop();                                   // 189
        });                                                            //
      })();                                                            //
    }                                                                  //
                                                                       //
    this.ready();                                                      // 193
  });                                                                  //
                                                                       //
  Meteor.publish("referralInfoPseudo", function () {                   // 196
    var _this3 = this;                                                 //
                                                                       //
    // This publishes a pseudo-collection called referralInfo whose documents have the following
    // form:                                                           //
    //                                                                 //
    // - id: (String) same as the User._id of an identity this user has referred
    // - name: (String) the profile.name from that identity            //
    // - completed: (Boolean) if this referral is complete             //
                                                                       //
    //  If the user is not logged in, then we have no referralInfo.    //
    if (!this.userId) {                                                // 205
      return [];                                                       // 206
    }                                                                  //
                                                                       //
    // Implementation note:                                            //
    //                                                                 //
    // This pseudo-collection is populated very differently for (1) the completed: false case versus
    // the (2) completed: true case.                                   //
                                                                       //
    // Case 1. Publish information about not-yet-complete referrals.   //
    var notCompletedReferralIdentitiesCursor = Meteor.users.find({     // 215
      referredBy: this.userId,                                         // 216
      "profile.name": { $exists: true }                                // 217
    }, {                                                               //
      fields: {                                                        // 219
        _id: 1,                                                        // 220
        referredBy: 1,                                                 // 221
        "profile.name": 1                                              // 222
      }                                                                //
    });                                                                //
    var notCompletedReferralIdentitiesHandle = notCompletedReferralIdentitiesCursor.observeChanges({
      // The added function gets called with the id of Bob when Alice refers Bob.
      added: function (id, fields) {                                   // 227
        _this3.added("referralInfo", id, { name: fields.profile.name, completed: false });
      },                                                               //
      // The removed function gets called when Bob is no longer an uncompleted referral.  Note that
      // this will get more complicated once we support sending completed referrals to the client.
      removed: function (id) {                                         // 232
        _this3.removed("referralInfo", id);                            // 233
      },                                                               //
      // The modified function gets called when Bob's profile.name changed.
      modified: function (id, fields) {                                // 236
        _this3.modified("referralInfo", id, { name: fields.profile.name, completed: false });
      }                                                                //
    });                                                                //
                                                                       //
    // Case 2. Handle completed referrals.                             //
    //                                                                 //
    // - Do a query for the current list of completed identities.      //
    //                                                                 //
    // - Every time we see a new such identity, we create a query that watches that one identity in
    //   case its profile.name changes.                                //
    //                                                                 //
    // - Also watch the first query, since the list of completed identities might change.
    var handleForProfileNameByIdentityId = {};                         // 249
    var stopWatchingAllIdentities = function () {                      // 250
      Object.keys(handleForProfileNameByIdentityId).forEach(function (identityId) {
        stopWatchingIdentity(identityId);                              // 252
      });                                                              //
    };                                                                 //
                                                                       //
    var stopWatchingIdentity = function (identityId) {                 // 256
      var handleForProfileName = handleForProfileNameByIdentityId[identityId];
      if (handleForProfileName) {                                      // 258
        _this3.removed("referralInfo", identityId);                    // 259
        handleForProfileName.stop();                                   // 260
        // delete is safe because we iterate across `Object.keys()` which returns a copy.
        delete handleForProfileNameByIdentityId[identityId];           // 262
      }                                                                //
    };                                                                 //
                                                                       //
    var watchIdentityAndPublishReferralSuccess = function (identityId) {
      var handleForProfileName = handleForProfileNameByIdentityId[identityId];
      if (handleForProfileName) {                                      // 268
        return;                                                        // 269
      }                                                                //
                                                                       //
      handleForProfileName = Meteor.users.find({                       // 272
        _id: identityId                                                // 273
      }, {                                                             //
        fields: {                                                      // 275
          "profile.name": 1                                            // 276
        }                                                              //
      }).observeChanges({                                              //
        added: function (id, fields) {                                 // 279
          _this3.added("referralInfo", id, { name: fields.profile.name, completed: true });
        },                                                             //
                                                                       //
        changed: function (id, fields) {                               // 283
          _this3.changed("referralInfo", id, { name: fields.profile.name, completed: true });
        },                                                             //
                                                                       //
        removed: function (id) {                                       // 287
          stopWatchingIdentity(id);                                    // 288
        }                                                              //
      });                                                              //
                                                                       //
      handleForProfileNameByIdentityId[identityId] = handleForProfileName;
    };                                                                 //
                                                                       //
    var completedIdentityIdsHandle = Meteor.users.find({               // 295
      _id: this.userId,                                                // 296
      referredIdentityIds: { $exists: true }                           // 297
    }, {                                                               //
      fields: {                                                        // 299
        referredIdentityIds: true                                      // 300
      }                                                                //
    }).observeChanges({                                                //
      // `added` gets called when a user gets their first completed referral.
      added: function (id, fields) {                                   // 304
        for (var i = 0; i < fields.referredIdentityIds.length; i++) {  // 305
          // Unconditionally mark these as successful referrals and start watching.
          watchIdentityAndPublishReferralSuccess(fields.referredIdentityIds[i]);
        }                                                              //
      },                                                               //
      // `changed` gets called when a user adds/removes referredIdentityIds, usually when a
      // referral becomes complete.                                    //
      changed: function (id, fields) {                                 // 313
        // Two major tasks.                                            //
        //                                                             //
        // 1. Look for identityIds to unsubscribe from & send removed notices to the client.
        //                                                             //
        // 2. Look for identityIds to subscribe to.                    //
                                                                       //
        // Task 1. Unsubscribe where needed.                           //
        var referredIdentityIdsAsObject = {};                          // 321
        fields.referredIdentityIds.forEach(function (i) {              // 322
          referredIdentityIdsAsObject[i] = true;                       // 322
        });                                                            //
                                                                       //
        Object.keys(handleForProfileNameByIdentityId).forEach(function (identityId) {
          // If the handle doesn't show up in the new list of referredIdentityIds, then remove
          // info from the client & stop it on the server & make it null.
          var handleForProfileName = handleForProfileNameByIdentityId[identityId];
          if (referredIdentityIdsAsObject.hasOwnProperty(identityId)) {
            stopWatchingIdentity(identityId);                          // 329
          }                                                            //
        });                                                            //
                                                                       //
        // Task 2. Subscribe where needed.                             //
        for (var i = 0; i < fields.referredIdentityIds.length; i++) {  // 334
          // The watch... function will avoid double-creating subscriptions, so this is safe.
          watchIdentityAndPublishReferralSuccess(fields.referredIdentityIds[i]);
        }                                                              //
      },                                                               //
      // `removed` gets called when a User suddenly has no referredIdentityIds.
      removed: function () {                                           // 340
        // Remove all data from client; stop all handles.              //
        stopWatchingAllIdentities();                                   // 342
      }                                                                //
    });                                                                //
                                                                       //
    // With cases 1 and 2 handled, register a cleanup function, then declare victory.
    this.onStop(function () {                                          // 347
      stopWatchingAllIdentities();                                     // 348
      notCompletedReferralIdentitiesHandle.stop();                     // 349
      completedIdentityIdsHandle.stop();                               // 350
    });                                                                //
                                                                       //
    self.ready();                                                      // 353
  });                                                                  //
                                                                       //
  Meteor.publish("backers", function () {                              // 356
    var _this4 = this;                                                 //
                                                                       //
    var backers = Assets.getText("backers.txt");                       // 357
    var anonCount = 0;                                                 // 358
    var counter = 0;                                                   // 359
                                                                       //
    backers.split("\n").forEach(function (name) {                      // 361
      name = name.trim();                                              // 362
      if (name === "") {                                               // 363
        ++anonCount;                                                   // 364
      } else {                                                         //
        _this4.added("backers", counter++, { name: name });            // 366
      }                                                                //
    });                                                                //
                                                                       //
    // Text file ends in \n but that shouldn't count.                  //
    --anonCount;                                                       // 371
                                                                       //
    this.added("backers", "anonymous", { count: anonCount - 1 });      // 373
                                                                       //
    this.ready();                                                      // 375
  });                                                                  //
}                                                                      //
                                                                       //
var makeAccountSettingsUi = function () {                              // 379
  return new SandstormAccountSettingsUi(globalTopbar, globalDb, window.location.protocol + "//" + makeWildcardHost("static"));
};                                                                     //
                                                                       //
var MONTHS = ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
                                                                       //
var formatInCountdown = function (template, countdownDatetime) {       // 386
  var diff = countdownDatetime.getTime() - Date.now();                 // 387
                                                                       //
  var units = {                                                        // 389
    day: 86400000,                                                     // 390
    hour: 3600000,                                                     // 391
    minute: 60000,                                                     // 392
    second: 1000                                                       // 393
  };                                                                   //
                                                                       //
  for (var unit in babelHelpers.sanitizeForInObject(units)) {          // 396
    // If it's more than one full unit away, then we'll print in terms of this unit. This does
    // mean that we write e.g. "1 minute" for the whole range between 2 minutes and 1 minute, but
    // whatever, this is typical of these sorts of displays.           //
    if (diff >= units[unit]) {                                         // 400
      var count = Math.floor(diff / units[unit]);                      // 401
      // Update next 1ms after the point where `count` would change.   //
      setTopBarTimeout(template, diff - count * units[unit] + 1);      // 403
      return {                                                         // 404
        text: "in " + count + " " + unit + (count > 1 ? "s" : ""),     // 405
        className: "countdown-" + unit                                 // 406
      };                                                               //
    }                                                                  //
  }                                                                    //
                                                                       //
  // We're within a second of the countdown, or past it.               //
  if (diff < -3600000) {                                               // 412
    // Notification appears stale.                                     //
    return null;                                                       // 414
  } else {                                                             //
    setTopBarTimeout(template, diff + 3600001);                        // 416
    return { text: "any moment", className: "countdown-now" };         // 417
  }                                                                    //
};                                                                     //
                                                                       //
var formatAccountExpires = function () {                               // 421
  var expires = Meteor.user().expires;                                 // 422
  return expires && expires.toLocaleTimeString() || null;              // 423
};                                                                     //
                                                                       //
var formatAccountExpiresIn = function (template, currentDatetime) {    // 426
  // TODO(someday): formatInCountdown will set the interval to match account expiration time, and
  // completely overwrite the previous interval for $IN_COUNTDOWN      //
  var user = Meteor.user() || {};                                      // 429
  var expires = user.expires || null;                                  // 430
  if (!expires) {                                                      // 431
    return null;                                                       // 432
  } else {                                                             //
    return formatInCountdown(template, expires, currentDatetime);      // 434
  }                                                                    //
};                                                                     //
                                                                       //
var setTopBarTimeout = function (template, delay) {                    // 438
  Meteor.clearTimeout(template.timeout);                               // 439
  template.timeout = Meteor.setTimeout(function () {                   // 440
    template.timer.changed();                                          // 441
  }, delay);                                                           //
                                                                       //
  // Make sure we re-run when the timeout triggers.                    //
  template.timer.depend();                                             // 445
};                                                                     //
                                                                       //
var determineAppName = function (grainId) {                            // 448
  // Returns:                                                          //
  //                                                                   //
  // - The current app title, if we can determine it, or               //
  //                                                                   //
  // - The empty string "", if we can't determine the current app title.
  var params = "";                                                     // 454
                                                                       //
  // Try our hardest to find the package's name, falling back on the default if needed.
  if (grainId) {                                                       // 457
    var grain = Grains.findOne({ _id: grainId });                      // 458
    if (grain && grain.packageId) {                                    // 459
      var thisPackage = Packages.findOne({ _id: grain.packageId });    // 460
      if (thisPackage) {                                               // 461
        params = SandstormDb.appNameFromPackage(thisPackage);          // 462
      }                                                                //
    }                                                                  //
  }                                                                    //
                                                                       //
  return params;                                                       // 467
};                                                                     //
                                                                       //
var billingPromptState = new ReactiveVar(null);                        // 470
                                                                       //
var showBillingPrompt = function (reason, next) {                      // 472
  billingPromptState.set({                                             // 473
    reason: reason,                                                    // 474
    db: globalDb,                                                      // 475
    topbar: globalTopbar,                                              // 476
    accountsUi: globalAccountsUi,                                      // 477
    onComplete: function () {                                          // 478
      billingPromptState.set(null);                                    // 479
      if (next) next();                                                // 480
    }                                                                  //
  });                                                                  //
};                                                                     //
                                                                       //
var ifQuotaAvailable = function (next) {                               // 485
  var reason = isUserOverQuota(Meteor.user());                         // 486
  if (reason) {                                                        // 487
    if (window.BlackrockPayments) {                                    // 488
      showBillingPrompt(reason, function () {                          // 489
        // If the user successfully raised their quota, continue the operation.
        if (!isUserOverQuota(Meteor.user())) {                         // 491
          next();                                                      // 492
        }                                                              //
      });                                                              //
    } else {                                                           //
      alert("You are out of storage space. Please delete some things and try again.");
    }                                                                  //
  } else {                                                             //
    next();                                                            // 499
  }                                                                    //
};                                                                     //
                                                                       //
var ifPlanAllowsCustomApps = function (next) {                         // 503
  if (globalDb.isDemoUser() || globalDb.isUninvitedFreeUser()) {       // 504
    if (window.BlackrockPayments) {                                    // 505
      showBillingPrompt("customApp", function () {                     // 506
        // If the user successfully chose a plan, continue the operation.
        if (!globalDb.isDemoUser() && !globalDb.isUninvitedFreeUser()) {
          next();                                                      // 509
        }                                                              //
      });                                                              //
    } else {                                                           //
      alert("Sorry, demo users cannot upload custom apps.");           // 513
    }                                                                  //
  } else {                                                             //
    next();                                                            // 516
  }                                                                    //
};                                                                     //
                                                                       //
var isDemoExpired = function () {                                      // 520
  var user = Meteor.user();                                            // 521
  if (!user) return false;                                             // 522
  var expires = user.expires;                                          // 523
  if (!expires) return false;                                          // 524
  expires = expires.getTime() - Date.now();                            // 525
  if (expires <= 0) return true;                                       // 526
  var comp = Tracker.currentComputation;                               // 527
  if (expires && comp) {                                               // 528
    Meteor.setTimeout(comp.invalidate.bind(comp), expires);            // 529
  }                                                                    //
                                                                       //
  return false;                                                        // 532
};                                                                     //
                                                                       //
// export: called by sandstorm-accounts-ui/login_buttons.js            //
logoutSandstorm = function () {                                        // 536
  Meteor.logout(function () {                                          // 537
    sessionStorage.removeItem("linkingIdentityLoginToken");            // 538
    Accounts._loginButtonsSession.closeDropdown();                     // 539
    globalTopbar.closePopup();                                         // 540
    var openGrains = globalGrains.get();                               // 541
    openGrains.forEach(function (grain) {                              // 542
      grain.destroy();                                                 // 543
    });                                                                //
                                                                       //
    globalGrains.set([]);                                              // 546
    Router.go("root");                                                 // 547
  });                                                                  //
};                                                                     //
                                                                       //
// export: this is also used by grain.js                               //
makeDateString = function (date) {                                     // 552
  if (!date) {                                                         // 553
    return "";                                                         // 554
  }                                                                    //
                                                                       //
  var result = undefined;                                              // 557
                                                                       //
  var now = new Date();                                                // 559
  var diff = now.valueOf() - date.valueOf();                           // 560
                                                                       //
  if (diff < 86400000 && now.getDate() === date.getDate()) {           // 562
    result = date.toLocaleTimeString();                                // 563
  } else {                                                             //
    result = MONTHS[date.getMonth()] + " " + date.getDate() + " ";     // 565
                                                                       //
    if (now.getFullYear() !== date.getFullYear()) {                    // 567
      result = date.getFullYear() + " " + result;                      // 568
    }                                                                  //
  }                                                                    //
                                                                       //
  return result;                                                       // 572
};                                                                     //
                                                                       //
// export: used in sandstorm-ui-grainlist                              //
prettySize = function (size) {                                         // 576
  var suffix = "B";                                                    // 577
  if (size >= 1000000000) {                                            // 578
    size = size / 1000000000;                                          // 579
    suffix = "GB";                                                     // 580
  } else if (size >= 1000000) {                                        //
    size = size / 1000000;                                             // 582
    suffix = "MB";                                                     // 583
  } else if (size >= 1000) {                                           //
    size = size / 1000;                                                // 585
    suffix = "kB";                                                     // 586
  }                                                                    //
                                                                       //
  return size.toPrecision(3) + suffix;                                 // 589
};                                                                     //
                                                                       //
// export: used in shared/demo.js                                      //
launchAndEnterGrainByPackageId = function (packageId) {                // 593
  var action = UserActions.findOne({ packageId: packageId });          // 594
  if (!action) {                                                       // 595
    alert("Somehow, you seem to have attempted to launch a package you have not installed.");
    return;                                                            // 597
  } else {                                                             //
    launchAndEnterGrainByActionId(action._id, null, null);             // 599
  }                                                                    //
};                                                                     //
                                                                       //
// export: used in sandstorm-ui-app-details                            //
launchAndEnterGrainByActionId = function (actionId, devPackageId, devIndex) {
  // Note that this takes a devPackageId and a devIndex as well. If provided,
  // they override the actionId.                                       //
  var packageId = undefined;                                           // 607
  var command = undefined;                                             // 608
  var appTitle = undefined;                                            // 609
  var nounPhrase = undefined;                                          // 610
  if (devPackageId) {                                                  // 611
    var devPackage = DevPackages.findOne(devPackageId);                // 612
    if (!devPackage) {                                                 // 613
      console.error("no such dev package: ", devPackageId);            // 614
      return;                                                          // 615
    }                                                                  //
                                                                       //
    var devAction = devPackage.manifest.actions[devIndex];             // 618
    packageId = devPackageId;                                          // 619
    command = devAction.command;                                       // 620
    appTitle = SandstormDb.appNameFromPackage(devPackage);             // 621
    nounPhrase = SandstormDb.nounPhraseForActionAndAppTitle(devAction, appTitle);
  } else {                                                             //
    var action = UserActions.findOne(actionId);                        // 624
    if (!action) {                                                     // 625
      console.error("no such action:", actionId);                      // 626
      return;                                                          // 627
    }                                                                  //
                                                                       //
    packageId = action.packageId;                                      // 630
    var pkg = Packages.findOne(packageId);                             // 631
    command = action.command;                                          // 632
    appTitle = SandstormDb.appNameFromPackage(pkg);                    // 633
    nounPhrase = SandstormDb.nounPhraseForActionAndAppTitle(action, appTitle);
  }                                                                    //
                                                                       //
  var title = "Untitled " + appTitle + " " + nounPhrase;               // 637
                                                                       //
  var identityId = Accounts.getCurrentIdentityId();                    // 639
                                                                       //
  // We need to ask the server to start a new grain, then browse to it.
  Meteor.call("newGrain", packageId, command, title, identityId, function (error, grainId) {
    if (error) {                                                       // 643
      console.error(error);                                            // 644
      alert(error.message);                                            // 645
    } else {                                                           //
      Router.go("grain", { grainId: grainId });                        // 647
    }                                                                  //
  });                                                                  //
};                                                                     //
                                                                       //
// export global - used in grain.js                                    //
globalQuotaEnforcer = {                                                // 653
  ifQuotaAvailable: ifQuotaAvailable,                                  // 654
  ifPlanAllowsCustomApps: ifPlanAllowsCustomApps                       // 655
};                                                                     //
                                                                       //
if (Meteor.isClient) {                                                 // 658
  HasUsers = new Mongo.Collection("hasUsers"); // dummy collection defined above
  Backers = new Mongo.Collection("backers"); // pseudo-collection defined above
  ReferralInfo = new Meteor.Collection("referralInfo"); // pseudo-collection
                                                                       //
  if (Meteor.settings["public"].quotaEnabled) {                        // 663
    window.testDisableQuotaClientSide = function () {                  // 664
      Meteor.settings["public"].quotaEnabled = false;                  // 665
    };                                                                 //
  }                                                                    //
                                                                       //
  Router.onRun(function () {                                           // 669
    // Close menus and popups any time we navigate.                    //
    globalTopbar.reset();                                              // 671
    this.next();                                                       // 672
  });                                                                  //
                                                                       //
  Template.layout.events({                                             // 675
    "click a": function (event) {                                      // 676
      // Close menus if a navigation link is clicked. Usually the Router.onRun(), above, will also
      // execute, but it will not in the case where the link points to the current page, yet we'd
      // really still like for the menus to close in such cases.       //
      if (!event.isDefaultPrevented()) {                               // 680
        globalTopbar.reset();                                          // 681
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.about.helpers({                                             // 686
    setDocumentTitle: function () {                                    // 687
      document.title = "About · " + globalDb.getServerTitle();         // 688
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.referrals.helpers({                                         // 692
    setDocumentTitle: function () {                                    // 693
      document.title = "Referral Program · " + globalDb.getServerTitle();
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.body.onRendered(function () {                               // 698
    // If we're on iOS, set a class name on <body> so we can use CSS styles to work around mobile
    // Safari's ridiculous iframe rendering behavior.                  //
    //                                                                 //
    // Note that this can't be done as a template helper because the <body> tag cannot have
    // attributes determined by helpers. This appears to be a Meteor bug related to the fact that
    // the <body> tag is not inside a <template>, but rather is itself equivalent to a <template>.
    if (/iPad|iPhone|iPod/.test(navigator.userAgent)) {                // 705
      document.body.className = "ios";                                 // 706
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.layout.onCreated(function () {                              // 710
    this.timer = new Tracker.Dependency();                             // 711
    var resizeTracker = this.resizeTracker = new Tracker.Dependency();
    this.resizeFunc = function () {                                    // 713
      resizeTracker.changed();                                         // 714
    };                                                                 //
                                                                       //
    window.addEventListener("resize", this.resizeFunc, false);         // 717
  });                                                                  //
                                                                       //
  Template.layout.onDestroyed(function () {                            // 720
    Meteor.clearTimeout(this.timeout);                                 // 721
    window.removeEventListener("resize", this.resizeFunc, false);      // 722
  });                                                                  //
                                                                       //
  Template.referrals.helpers({                                         // 725
    isPaid: function () {                                              // 726
      return Meteor.user() && Meteor.user().plan && Meteor.user().plan !== "free";
    },                                                                 //
                                                                       //
    notYetCompleteReferralNames: function () {                         // 730
      return ReferralInfo.find({ completed: false });                  // 731
    },                                                                 //
                                                                       //
    completeReferralNames: function () {                               // 734
      return ReferralInfo.find({ completed: true });                   // 735
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.layout.helpers({                                            // 739
    adminAlertIsTooLarge: function () {                                // 740
      Template.instance().resizeTracker.depend();                      // 741
      var setting = Settings.findOne({ _id: "adminAlert" });           // 742
      if (!setting || !setting.value) {                                // 743
        return false;                                                  // 744
      }                                                                //
      // TODO(someday): 10 and 850 are just magic values that estimate the size of the font and the
      // number of pixels everything else in the topbar respectively. This should really be
      // calculated based on actual sizes of things in the topbar.     //
      return window.innerWidth - setting.value.length * 10 < 850;      // 749
    },                                                                 //
                                                                       //
    adminAlert: function () {                                          // 752
      var setting = Settings.findOne({ _id: "adminAlert" });           // 753
      if (!setting || !setting.value) {                                // 754
        return null;                                                   // 755
      }                                                                //
                                                                       //
      var text = setting.value;                                        // 758
                                                                       //
      var alertTimeSetting = Settings.findOne({ _id: "adminAlertTime" });
      var alertTime = alertTimeSetting && alertTimeSetting.value;      // 761
                                                                       //
      var alertUrlSetting = Settings.findOne({ _id: "adminAlertUrl" });
      var alertUrl = alertUrlSetting ? alertUrlSetting.value.trim() : null;
      if (!alertUrl) alertUrl = null;                                  // 765
                                                                       //
      var template = Template.instance();                              // 767
      var param = undefined;                                           // 768
      var className = undefined;                                       // 769
      if (text.indexOf("$TIME") !== -1) {                              // 770
        if (!alertTime) return null;                                   // 771
        text = text.replace("$TIME", alertTime.toLocaleTimeString());  // 772
      }                                                                //
                                                                       //
      if (text.indexOf("$DATE") !== -1) {                              // 775
        if (!alertTime) return null;                                   // 776
        text = text.replace("$DATE", alertTime.toLocaleDateString());  // 777
      }                                                                //
                                                                       //
      if (text.indexOf("$IN_COUNTDOWN") !== -1) {                      // 780
        if (!alertTime) return null;                                   // 781
        param = formatInCountdown(template, alertTime);                // 782
        if (!param) return null;                                       // 783
        text = text.replace("$IN_COUNTDOWN", param.text);              // 784
        className = param.className;                                   // 785
      }                                                                //
                                                                       //
      if (text.indexOf("$ACCOUNT_EXPIRES_IN") !== -1) {                // 788
        param = formatAccountExpiresIn(template);                      // 789
        if (!param) return null;                                       // 790
        text = text.replace("$ACCOUNT_EXPIRES_IN", param.text);        // 791
        className = param.className;                                   // 792
      }                                                                //
                                                                       //
      if (text.indexOf("$ACCOUNT_EXPIRES") !== -1) {                   // 795
        param = formatAccountExpires();                                // 796
        if (!param) return null;                                       // 797
        text = text.replace("$ACCOUNT_EXPIRES", param);                // 798
      }                                                                //
                                                                       //
      if (text.indexOf("$APPNAME") !== -1) {                           // 801
        text = text.replace("$APPNAME", determineAppName(this.grainId));
      }                                                                //
                                                                       //
      if (alertUrl && alertUrl.indexOf("$APPNAME") !== -1) {           // 805
        alertUrl = alertUrl.replace("$APPNAME", determineAppName(this.grainId));
      }                                                                //
                                                                       //
      return {                                                         // 809
        text: text,                                                    // 810
        className: className,                                          // 811
        alertUrl: alertUrl                                             // 812
      };                                                               //
    },                                                                 //
                                                                       //
    billingPromptState: function () {                                  // 816
      return billingPromptState.get();                                 // 817
    },                                                                 //
                                                                       //
    demoExpired: isDemoExpired,                                        // 820
    canUpgradeDemo: function () {                                      // 821
      return Meteor.settings["public"].allowUninvited;                 // 822
    },                                                                 //
                                                                       //
    globalAccountsUi: function () {                                    // 825
      return globalAccountsUi;                                         // 826
    },                                                                 //
                                                                       //
    identityUser: function () {                                        // 829
      var user = Meteor.user();                                        // 830
      return user && user.profile;                                     // 831
    },                                                                 //
                                                                       //
    showAccountButtons: function () {                                  // 834
      return Meteor.user() && !Meteor.loggingIn() && !isDemoUser();    // 835
    },                                                                 //
                                                                       //
    accountButtonsData: function () {                                  // 838
      return { isAdmin: globalDb.isAdmin() };                          // 839
    },                                                                 //
                                                                       //
    firstLogin: function () {                                          // 842
      return credentialsSubscription.ready() && !isDemoUser() && !Meteor.loggingIn() && Meteor.user() && !Meteor.user().hasCompletedSignup;
    },                                                                 //
                                                                       //
    accountSettingsUi: function () {                                   // 847
      return makeAccountSettingsUi();                                  // 848
    },                                                                 //
                                                                       //
    firstTimeBillingPromptState: function () {                         // 851
      // Should we show the first-time billing plan selector?          //
                                                                       //
      // Don't show if billing is not enabled.                         //
      if (!window.BlackrockPayments) return;                           // 855
                                                                       //
      var user = Meteor.user();                                        // 857
                                                                       //
      // Don't show if not logged in.                                  //
      if (!user) return;                                               // 860
                                                                       //
      // Don't show if not in the experiment.                          //
      if (!user.experiments || user.experiments.firstTimeBillingPrompt !== "test") return;
                                                                       //
      // Don't show if the user has selected a plan already.           //
      if (user.plan && !Session.get("firstTimeBillingPromptOpen")) return;
                                                                       //
      // Only show to account users (not identities).                  //
      if (!user.loginIdentities) return;                               // 869
                                                                       //
      // Don't show to demo users.                                     //
      if (user.expires) return;                                        // 872
                                                                       //
      // Don't show when viewing another user's grain. We don't want to scare people away from
      // logging in to collaborate.                                    //
      var route = Router.current().route.getName();                    // 876
      if (route === "shared") return;                                  // 877
      if (route === "grain") {                                         // 878
        if (_.some(globalGrains.get(), function (grain) {              // 879
          return grain.isActive() && !grain.isOwner();                 // 880
        })) {                                                          //
                                                                       //
          return;                                                      // 883
        }                                                              //
      }                                                                //
                                                                       //
      // Don't let the plan chooser disappear instantly once user.plan is set.
      Session.set("firstTimeBillingPromptOpen", true);                 // 888
                                                                       //
      // OK, show it.                                                  //
      return {                                                         // 891
        db: globalDb,                                                  // 892
        topbar: globalTopbar,                                          // 893
        accountsUi: globalAccountsUi,                                  // 894
        onComplete: function () {                                      // 895
          Session.set("firstTimeBillingPromptOpen", false);            // 896
        }                                                              //
      };                                                               //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.layout.events({                                             // 902
    "click #demo-expired .logout": function (event) {                  // 903
      logoutSandstorm();                                               // 904
    }                                                                  //
  });                                                                  //
                                                                       //
  credentialsSubscription = Meteor.subscribe("credentials");           // 908
                                                                       //
  Template.registerHelper("dateString", makeDateString);               // 910
  Template.registerHelper("hideNavbar", function () {                  // 911
    // Hide navbar if user is not logged in, since they can't go anywhere with it.
    return !Meteor.userId() || isDemoExpired();                        // 913
  });                                                                  //
                                                                       //
  Template.registerHelper("shrinkNavbar", function () {                // 916
    // Shrink the navbar if the user clicked the button to do so.      //
    return Session.get("shrink-navbar");                               // 918
  });                                                                  //
                                                                       //
  Template.registerHelper("quotaEnabled", function () {                // 921
    return Meteor.settings["public"].quotaEnabled;                     // 922
  });                                                                  //
                                                                       //
  Template.root.helpers({                                              // 925
    storageUsage: function () {                                        // 926
      return Meteor.userId() ? prettySize(Meteor.user().storageUsage || 0) : undefined;
    },                                                                 //
                                                                       //
    storageQuota: function () {                                        // 930
      var plan = globalDb.getMyPlan();                                 // 931
      return plan ? prettySize(plan.storage) : undefined;              // 932
    },                                                                 //
                                                                       //
    overQuota: function () {                                           // 935
      return !window.BlackrockPayments && isUserOverQuota(Meteor.user());
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.root.events({                                               // 940
    "click .uninstall-app-button": function (event) {                  // 941
      // TODO(cleanup): This event handler is no longer used, but the new UI does not yet implement
      //   uninstall. Leave this code here for reference until it does.
      var appId = event.currentTarget.getAttribute("data-appid");      // 944
      if (window.confirm("Really uninstall this app?")) {              // 945
        UserActions.find({ appId: appId, userId: Meteor.userId() }).forEach(function (action) {
          UserActions.remove(action._id);                              // 947
        });                                                            //
                                                                       //
        Meteor.call("deleteUnusedPackages", appId);                    // 950
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.notificationsPopup.helpers({                                // 955
    notifications: function () {                                       // 956
      Meteor.call("readAllNotifications");                             // 957
      return Notifications.find({ userId: Meteor.userId() }, { sort: { timestamp: -1 } }).map(function (row) {
        var grain = Grains.findOne({ _id: row.grainId });              // 959
        if (grain) {                                                   // 960
          row.grainTitle = grain.title;                                // 961
        }                                                              //
                                                                       //
        return row;                                                    // 964
      });                                                              //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.notifications.helpers({                                     // 969
    notificationCount: function () {                                   // 970
      return Notifications.find({ userId: Meteor.userId(), isUnread: true }).count();
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.notificationsPopup.events({                                 // 975
    "click #notification-dropdown": function (event) {                 // 976
      return false;                                                    // 977
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.notificationItem.helpers({                                  // 981
    isAppUpdates: function () {                                        // 982
      return !!this.appUpdates;                                        // 983
    },                                                                 //
                                                                       //
    notificationTitle: function () {                                   // 986
      if (this.admin) {                                                // 987
        return "Notification from System";                             // 988
      } else if (this.appUpdates) {                                    //
        return "App updates are available";                            // 990
      } else if (this.referral || this.mailingListBonus) {             //
        return false;                                                  // 992
      }                                                                //
                                                                       //
      return this.grainTitle + " is backgrounded";                     // 995
    },                                                                 //
                                                                       //
    titleHelperText: function () {                                     // 998
      if (this.admin) {                                                // 999
        return "Dismiss this system notification";                     // 1000
      } else if (this.referral) {                                      //
        return "Dismiss this referral notification";                   // 1002
      } else {                                                         //
        return "Stop the background app";                              // 1004
      }                                                                //
    },                                                                 //
                                                                       //
    dismissText: function () {                                         // 1008
      if (this.admin && this.admin.type === "reportStats") {           // 1009
        return false;                                                  // 1010
      } else if (this.referral) {                                      //
        return "Dismiss";                                              // 1012
      }                                                                //
                                                                       //
      return "Cancel";                                                 // 1015
    },                                                                 //
                                                                       //
    adminLink: function () {                                           // 1018
      return this.admin && this.admin.action;                          // 1019
    },                                                                 //
                                                                       //
    appUpdatesList: function () {                                      // 1022
      return _.values(this.appUpdates);                                // 1023
    },                                                                 //
                                                                       //
    paidUser: function () {                                            // 1026
      var plan = Meteor.user().plan;                                   // 1027
      return plan && plan !== "free";                                  // 1028
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.notificationItem.events({                                   // 1032
    "click .cancel-notification": function (event) {                   // 1033
      Meteor.call("dismissNotification", this._id);                    // 1034
      return false;                                                    // 1035
    },                                                                 //
                                                                       //
    "click .accept-notification": function (event) {                   // 1038
      var _this5 = this;                                               //
                                                                       //
      if (this.appUpdates) {                                           // 1039
        Meteor.call("updateApps", this.appUpdates, function (err) {    // 1040
          // TODO(someday): if (err)                                   //
          Meteor.call("dismissNotification", _this5._id);              // 1042
        });                                                            //
      }                                                                //
                                                                       //
      return false;                                                    // 1046
    },                                                                 //
                                                                       //
    "click .dismiss-notification": function (event) {                  // 1049
      Meteor.call("dismissNotification", this._id);                    // 1050
    }                                                                  //
  });                                                                  //
                                                                       //
  Meteor.startup(function () {                                         // 1054
    // Tell app authors how to run JS in the context of the grain-frame.
    console.log("%cApp authors: To understand the grain-frame in Sandstorm and how to find " + "logs and perform troubleshooting, see: " + "\n- https://docs.sandstorm.io/en/latest/developing/path/ " + "\n- https://docs.sandstorm.io/en/latest/using/top-bar/ " + "\n- https://docs.sandstorm.io/en/latest/developing/troubleshooting/ " + "\n" + "\nWhen debugging, make sure you execute Javascript " + "in the context of the 'grain-frame' IFRAME. References: " + "\n- https://stackoverflow.com/questions/3275816/debugging-iframes-with-chrome-developer-tools " + "\n- https://developer.mozilla.org/en-US/docs/Tools/Working_with_iframes " + "\n" + "\nWe can also provide personal assistance! Get in touch: https://sandstorm.io/community", "font-size: large; background-color: yellow;");
                                                                       //
    Meteor.subscribe("notifications");                                 // 1071
                                                                       //
    Meteor.autorun(function () {                                       // 1073
      Meteor.subscribe("notificationGrains", Notifications.find().map(function (row) {
        return row._id;                                                // 1076
      }));                                                             //
    });                                                                //
  });                                                                  //
                                                                       //
  Meteor.methods({                                                     // 1082
    dismissNotification: function (notificationId) {                   // 1083
      // Client-side simulation of dismissNotification.                //
      Notifications.remove({ _id: notificationId });                   // 1085
    }                                                                  //
  });                                                                  //
}                                                                      //
                                                                       //
Router.configure({                                                     // 1090
  layoutTemplate: "layout",                                            // 1091
  notFoundTemplate: "notFound",                                        // 1092
  loadingTemplate: "loading"                                           // 1093
});                                                                    //
                                                                       //
if (Meteor.isClient) {                                                 // 1096
  Router.onBeforeAction("loading");                                    // 1097
}                                                                      //
                                                                       //
function getBuildInfo() {                                              // 1100
  var build = Meteor.settings && Meteor.settings["public"] && Meteor.settings["public"].build;
  var isNumber = typeof build === "number";                            // 1102
  if (!build) {                                                        // 1103
    build = "(unknown)";                                               // 1104
  } else if (isNumber) {                                               //
    build = String(Math.floor(build / 1000)) + "." + String(build % 1000);
  }                                                                    //
                                                                       //
  return {                                                             // 1109
    build: build,                                                      // 1110
    isUnofficial: !isNumber                                            // 1111
  };                                                                   //
}                                                                      //
                                                                       //
var promptForFile = function (input, callback) {                       // 1115
  // TODO(cleanup): Share code with "upload picture" and other upload buttons.
  function listener(e) {                                               // 1117
    input.removeEventListener("change", listener);                     // 1118
    callback(e.currentTarget.files[0]);                                // 1119
  }                                                                    //
                                                                       //
  input.addEventListener("change", listener);                          // 1122
  input.click();                                                       // 1123
};                                                                     //
                                                                       //
var startUpload = function (file, endpoint, onComplete) {              // 1126
  // TODO(cleanup): Use Meteor's HTTP, although this may require sending them a PR to support
  //   progress callbacks (and officially document that binary input is accepted).
                                                                       //
  Session.set("uploadStatus", "Uploading");                            // 1130
                                                                       //
  var xhr = new XMLHttpRequest();                                      // 1132
                                                                       //
  xhr.onreadystatechange = function () {                               // 1134
    if (xhr.readyState == 4) {                                         // 1135
      console.log(xhr.status);                                         // 1136
      if (xhr.status == 200) {                                         // 1137
        Session.set("uploadProgress", 0);                              // 1138
        onComplete(xhr.responseText);                                  // 1139
      } else {                                                         //
        Session.set("uploadError", {                                   // 1141
          status: xhr.status,                                          // 1142
          statusText: xhr.statusText,                                  // 1143
          response: xhr.responseText                                   // 1144
        });                                                            //
      }                                                                //
    }                                                                  //
  };                                                                   //
                                                                       //
  if (xhr.upload) {                                                    // 1150
    xhr.upload.addEventListener("progress", function (progressEvent) {
      Session.set("uploadProgress", Math.floor(progressEvent.loaded / progressEvent.total * 100));
    });                                                                //
  }                                                                    //
                                                                       //
  xhr.open("POST", endpoint, true);                                    // 1157
  xhr.send(file);                                                      // 1158
                                                                       //
  Router.go("uploadStatus");                                           // 1160
};                                                                     //
                                                                       //
restoreBackup = function (file) {                                      // 1163
  // This function is global so tests can call it                      //
  startUpload(file, "/uploadBackup", function (response) {             // 1165
    Session.set("uploadStatus", "Unpacking");                          // 1166
    var identityId = Accounts.getCurrentIdentityId();                  // 1167
    Meteor.call("restoreGrain", response, identityId, function (err, grainId) {
      if (err) {                                                       // 1169
        console.log(err);                                              // 1170
        Session.set("uploadStatus", undefined);                        // 1171
        Session.set("uploadError", {                                   // 1172
          status: "",                                                  // 1173
          statusText: err.message                                      // 1174
        });                                                            //
      } else {                                                         //
        Router.go("grain", { grainId: grainId });                      // 1177
      }                                                                //
    });                                                                //
  });                                                                  //
};                                                                     //
                                                                       //
promptRestoreBackup = function (input) {                               // 1183
  promptForFile(input, restoreBackup);                                 // 1184
};                                                                     //
                                                                       //
uploadApp = function (file) {                                          // 1187
  // This function is global so tests can call it                      //
  Meteor.call("newUploadToken", function (err, token) {                // 1189
    if (err) {                                                         // 1190
      console.error(err);                                              // 1191
      alert(err.message);                                              // 1192
    } else {                                                           //
      startUpload(file, "/upload/" + token, function (response) {      // 1194
        Session.set("uploadStatus", undefined);                        // 1195
        Router.go("install", { packageId: response });                 // 1196
      });                                                              //
    }                                                                  //
  });                                                                  //
};                                                                     //
                                                                       //
promptUploadApp = function (input) {                                   // 1202
  promptForFile(input, uploadApp);                                     // 1203
};                                                                     //
                                                                       //
Router.map(function () {                                               // 1206
  this.route("root", {                                                 // 1207
    path: "/",                                                         // 1208
    waitOn: function () {                                              // 1209
      return [Meteor.subscribe("hasUsers"), Meteor.subscribe("grainsMenu")];
    },                                                                 //
                                                                       //
    data: function () {                                                // 1216
      // If the user is logged-in, and can create new grains, and      //
      // has no grains yet, then send them to "new".                   //
      if (this.ready() && Meteor.userId() && !Meteor.loggingIn()) {    // 1219
        if (globalDb.currentUserGrains({}, {}).count() === 0 && globalDb.currentUserApiTokens().count() === 0) {
          Router.go("apps", {}, { replaceState: true });               // 1222
        } else {                                                       //
          Router.go("grains", {}, { replaceState: true });             // 1224
        }                                                              //
      }                                                                //
                                                                       //
      return {                                                         // 1228
        needsAdminTokenLogin: this.ready() && !HasUsers.findOne("hasUsers") && !globalDb.allowDevAccounts(),
        build: getBuildInfo().build,                                   // 1230
        splashUrl: (Settings.findOne("splashUrl") || {}).value         // 1231
      };                                                               //
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("linkHandler", {                                          // 1236
    path: "/link-handler/:url",                                        // 1237
                                                                       //
    data: function () {                                                // 1239
      var url = this.params.url;                                       // 1240
      if (url.lastIndexOf("web+sandstorm:", 0) === 0) {                // 1241
        url = url.slice("web+sandstorm:".length);                      // 1242
      }                                                                //
      // TODO(cleanup):  Didn't use Router.go() because the url may contain a query term.
      document.location = "/install/" + url;                           // 1245
      return {};                                                       // 1246
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("about", {                                                // 1250
    path: "/about",                                                    // 1251
    data: function () {                                                // 1252
      var result = getBuildInfo();                                     // 1253
                                                                       //
      var backers = Session.get("backers");                            // 1255
      if (backers) {                                                   // 1256
        result.backers = backers.names;                                // 1257
        result.anonCount = backers.anonCount;                          // 1258
      } else {                                                         //
        HTTP.get("/sandstorm-backers.txt", function (err, response) {  // 1260
          var names = response.content.split("\n").sort(function (a, b) {
            return a.toLowerCase().localeCompare(b.toLowerCase());     // 1262
          });                                                          //
                                                                       //
          var anonCount = 0;                                           // 1265
          while (anonCount < names.length && names[anonCount] === "") {
            ++anonCount;                                               // 1267
          }                                                            //
                                                                       //
          names = names.slice(anonCount);                              // 1270
                                                                       //
          // File ends in trailing newline, but that last blank line does not represent an
          // anonymous contributor.                                    //
          --anonCount;                                                 // 1274
                                                                       //
          Session.set("backers", {                                     // 1276
            names: names,                                              // 1277
            anonCount: anonCount                                       // 1278
          });                                                          //
        });                                                            //
      }                                                                //
                                                                       //
      result.termsUrl = globalDb.getSetting("termsUrl");               // 1283
      result.privacyUrl = globalDb.getSetting("privacyUrl");           // 1284
                                                                       //
      return result;                                                   // 1286
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("uploadStatus", {                                         // 1290
    path: "/upload",                                                   // 1291
                                                                       //
    waitOn: function () {                                              // 1293
      return Meteor.subscribe("credentials");                          // 1294
    },                                                                 //
                                                                       //
    data: function () {                                                // 1297
      return {                                                         // 1298
        progress: Session.get("uploadProgress"),                       // 1299
        status: Session.get("uploadStatus"),                           // 1300
        error: Session.get("uploadError")                              // 1301
      };                                                               //
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("referrals", {                                            // 1306
    path: "/referrals",                                                // 1307
                                                                       //
    waitOn: function () {                                              // 1309
      return Meteor.subscribe("referralInfoPseudo");                   // 1310
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("account", {                                              // 1314
    path: "/account",                                                  // 1315
                                                                       //
    data: function () {                                                // 1317
      // Don't allow logged-out or demo users to visit the accounts page. There should be no way
      // for them to get there except for typing the URL manually. In theory showing the accounts
      // page to a demo user could make some sense for editing their profile, but we really do not
      // want them signing up for subscription plans!                  //
      if (!Meteor.user() && !Meteor.loggingIn() || globalDb.isDemoUser()) {
        Router.go("root", {}, { replaceState: true });                 // 1323
      } else {                                                         //
        return makeAccountSettingsUi();                                // 1325
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("accountUsage", {                                         // 1330
    path: "/account/usage"                                             // 1331
  });                                                                  //
});                                                                    //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=shell.js.map
