(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var Accounts = Package['accounts-base'].Accounts;
var AccountsServer = Package['accounts-base'].AccountsServer;
var SandstormDb = Package['sandstorm-db'].SandstormDb;
var SandstormPermissions = Package['sandstorm-permissions'].SandstormPermissions;

/* Package-scope variables */
var inMeteor, promiseToFuture, waitPromise, SANDSTORM_ALTHOME, SandstormBackend;

(function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// packages/sandstorm-backend/sandstorm-backend.js                                                                     //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
                                                                                                                       // 16
var Capnp = Npm.require("capnp");                                                                                      // 17
var Backend = Capnp.importSystem("sandstorm/backend.capnp").Backend;                                                   // 18
var Crypto = Npm.require("crypto");                                                                                    // 19
var Future = Npm.require("fibers/future");                                                                             // 20
var Promise = Npm.require("es6-promise").Promise;                                                                      // 21
                                                                                                                       // 22
var inMeteorInternal = Meteor.bindEnvironment(function (callback) {                                                    // 23
  callback();                                                                                                          // 24
});                                                                                                                    // 25
                                                                                                                       // 26
inMeteor = function (callback) {                                                                                       // 27
  // Calls the callback in a Meteor context.  Returns a Promise for its result.                                        // 28
  return new Promise(function (resolve, reject) {                                                                      // 29
    inMeteorInternal(function () {                                                                                     // 30
      try {                                                                                                            // 31
        resolve(callback());                                                                                           // 32
      } catch (err) {                                                                                                  // 33
        reject(err);                                                                                                   // 34
      }                                                                                                                // 35
    });                                                                                                                // 36
  });                                                                                                                  // 37
}                                                                                                                      // 38
                                                                                                                       // 39
promiseToFuture = function (promise) {                                                                                 // 40
  var result = new Future();                                                                                           // 41
  promise.then(result.return.bind(result), result.throw.bind(result));                                                 // 42
  return result;                                                                                                       // 43
}                                                                                                                      // 44
                                                                                                                       // 45
waitPromise = function (promise) {                                                                                     // 46
  return promiseToFuture(promise).wait();                                                                              // 47
}                                                                                                                      // 48
                                                                                                                       // 49
SANDSTORM_ALTHOME = Meteor.settings && Meteor.settings.home;                                                           // 50
                                                                                                                       // 51
SandstormBackend = function(db, backendCap) {                                                                          // 52
  this._db = db;                                                                                                       // 53
  this.runningGrains = {};                                                                                             // 54
  this._backendCap = backendCap                                                                                        // 55
}                                                                                                                      // 56
                                                                                                                       // 57
SandstormBackend.prototype.cap = function() {                                                                          // 58
  return this._backendCap;                                                                                             // 59
}                                                                                                                      // 60
                                                                                                                       // 61
SandstormBackend.prototype.deleteUser = function (userId) {                                                            // 62
  return waitPromise(this._backendCap.deleteUser());                                                                   // 63
}                                                                                                                      // 64
                                                                                                                       // 65
SandstormBackend.prototype.shutdownGrain = function (grainId, ownerId, keepSessions) {                                 // 66
  if (!keepSessions) {                                                                                                 // 67
    Sessions.remove({grainId: grainId});                                                                               // 68
    delete this.runningGrains[grainId];                                                                                // 69
  }                                                                                                                    // 70
                                                                                                                       // 71
  var grain = this._backendCap.getGrain(ownerId, grainId).supervisor;                                                  // 72
  return grain.shutdown().then(function () {                                                                           // 73
    grain.close();                                                                                                     // 74
    throw new Error("expected shutdown() to throw disconnected");                                                      // 75
  }, function (err) {                                                                                                  // 76
    grain.close();                                                                                                     // 77
    if (err.kjType !== "disconnected") {                                                                               // 78
      throw err;                                                                                                       // 79
    }                                                                                                                  // 80
  });                                                                                                                  // 81
}                                                                                                                      // 82
                                                                                                                       // 83
SandstormBackend.prototype.deleteGrain = function (grainId, ownerId) {                                                 // 84
  // We leave it up to the caller if they want to actually wait, but some don't so we report                           // 85
  // exceptions.                                                                                                       // 86
  return this._backendCap.deleteGrain(ownerId, grainId).catch(function (err) {                                         // 87
    console.error("problem deleting grain " + grainId + ":", err.message);                                             // 88
    throw err;                                                                                                         // 89
  });                                                                                                                  // 90
}                                                                                                                      // 91
                                                                                                                       // 92
                                                                                                                       // 93
SandstormBackend.prototype.openGrain = function (grainId, isRetry) {                                                   // 94
  // Create a Cap'n Proto connection to the given grain. Note that this function does not actually                     // 95
  // verify that the connection succeeded. Instead, if an RPC call to the connection fails, check                      // 96
  // shouldRestartGrain(). If it returns true, call continueGrain() and then openGrain()                               // 97
  // again with isRetry = true, and then retry.                                                                        // 98
  //                                                                                                                   // 99
  // Must be called in a Meteor context.                                                                               // 100
                                                                                                                       // 101
  if (isRetry) {                                                                                                       // 102
    // Since this is a retry, try starting the grain even if we think it's already running.                            // 103
    return this.continueGrain(grainId);                                                                                // 104
  } else {                                                                                                             // 105
    // Start the grain if it is not running.                                                                           // 106
    var runningGrain = this.runningGrains[grainId];                                                                    // 107
    if (runningGrain) {                                                                                                // 108
      return waitPromise(runningGrain);                                                                                // 109
    } else {                                                                                                           // 110
      return this.continueGrain(grainId);                                                                              // 111
    }                                                                                                                  // 112
  }                                                                                                                    // 113
}                                                                                                                      // 114
                                                                                                                       // 115
SandstormBackend.shouldRestartGrain = function (error, retryCount) {                                                   // 116
  // Given an error thrown by an RPC call to a grain, return whether or not it makes sense to try                      // 117
  // to restart the grain and retry. `retryCount` is the number of times that the request has                          // 118
  // already gone through this cycle (should be zero for the first call).                                              // 119
                                                                                                                       // 120
  return error.kjType === "disconnected" && retryCount < 1;                                                            // 121
}                                                                                                                      // 122
                                                                                                                       // 123
SandstormBackend.prototype.maybeRetryUseGrain = function (grainId, cb, retryCount, err) {                              // 124
  var self = this;                                                                                                     // 125
  if (SandstormBackend.shouldRestartGrain(err, retryCount)) {                                                          // 126
    return inMeteor(function () {                                                                                      // 127
      return cb(self.openGrain(grainId, true).supervisor)                                                              // 128
          .catch(self.maybeRetryUseGrain.bind(self, grainId, cb, retryCount + 1));                                     // 129
    });                                                                                                                // 130
  } else {                                                                                                             // 131
    throw err;                                                                                                         // 132
  }                                                                                                                    // 133
}                                                                                                                      // 134
                                                                                                                       // 135
SandstormBackend.prototype.useGrain = function (grainId, cb) {                                                         // 136
  // This will open a grain for you, handling restarts if needed, and call the passed function with                    // 137
  // the supervisor capability as the only parameter. The callback must return a promise that used                     // 138
  // the supervisor, so that we can check if a disconnect error occurred, and retry if possible.                       // 139
  // This function returns the same promise that your callback returns.                                                // 140
  //                                                                                                                   // 141
  // This function is NOT expected to be run in a meteor context.                                                      // 142
                                                                                                                       // 143
  var runningGrain = this.runningGrains[grainId];                                                                      // 144
  var self = this;                                                                                                     // 145
  if (runningGrain) {                                                                                                  // 146
    return runningGrain.then(function (grainInfo) {                                                                    // 147
      return cb(grainInfo.supervisor);                                                                                 // 148
    }).catch(self.maybeRetryUseGrain.bind(self, grainId, cb, 0));                                                      // 149
  } else {                                                                                                             // 150
    return inMeteor(function () {                                                                                      // 151
      return cb(self.openGrain(grainId, false).supervisor)                                                             // 152
          .catch(self.maybeRetryUseGrain.bind(self, grainId, cb, 0));                                                  // 153
    });                                                                                                                // 154
  }                                                                                                                    // 155
}                                                                                                                      // 156
                                                                                                                       // 157
SandstormBackend.prototype.continueGrain = function(grainId) {                                                         // 158
  var grain = Grains.findOne(grainId);                                                                                 // 159
  if (!grain) {                                                                                                        // 160
    throw new Meteor.Error(404, "Grain Not Found", "Grain ID: " + grainId);                                            // 161
  }                                                                                                                    // 162
                                                                                                                       // 163
  // If a DevPackage with the same app ID is currently active, we let it override the installed                        // 164
  // package, so that the grain runs using the dev app.                                                                // 165
  var devPackage = DevPackages.findOne({appId: grain.appId});                                                          // 166
  var isDev;                                                                                                           // 167
  var pkg;                                                                                                             // 168
  if (devPackage) {                                                                                                    // 169
    isDev = true;                                                                                                      // 170
    pkg = devPackage;                                                                                                  // 171
  } else {                                                                                                             // 172
    pkg = Packages.findOne(grain.packageId);                                                                           // 173
    if (!pkg) {                                                                                                        // 174
      throw new Meteor.Error(500, "Grain's package not installed",                                                     // 175
                             "Package ID: " + grain.packageId);                                                        // 176
    }                                                                                                                  // 177
  }                                                                                                                    // 178
  var manifest = pkg.manifest;                                                                                         // 179
  var packageId = pkg._id;                                                                                             // 180
                                                                                                                       // 181
  if (!("continueCommand" in manifest)) {                                                                              // 182
    throw new Meteor.Error(500, "Package manifest defines no continueCommand.",                                        // 183
                           "Package ID: " + packageId);                                                                // 184
  }                                                                                                                    // 185
                                                                                                                       // 186
  return this.startGrainInternal(                                                                                      // 187
      packageId, grainId, grain.userId, manifest.continueCommand, false, isDev);                                       // 188
}                                                                                                                      // 189
                                                                                                                       // 190
SandstormBackend.prototype.startGrainInternal = function(packageId, grainId, ownerId, command, isNew, isDev) {         // 191
  // Starts the grain supervisor.  Must be executed in a Meteor context.  Blocks until grain is                        // 192
  // started. Returns a promise for an object containing two fields: `owner` (the ID of the owning                     // 193
  // user) and `supervisor` (the supervisor capability).                                                               // 194
                                                                                                                       // 195
  if (isUserExcessivelyOverQuota(Meteor.users.findOne(ownerId))) {                                                     // 196
    throw new Meteor.Error(402,                                                                                        // 197
                           ("Cannot start grain because owner's storage is exhausted.\n" +                             // 198
                            "Please ask them to upgrade."));                                                           // 199
  }                                                                                                                    // 200
                                                                                                                       // 201
  // Ugly: Stay backwards-compatible with old manifests that had "executablePath" and "args" rather                    // 202
  //   than just "argv".                                                                                               // 203
  if ("args" in command) {                                                                                             // 204
    if (!("argv" in command)) {                                                                                        // 205
      command.argv = command.args;                                                                                     // 206
    }                                                                                                                  // 207
    delete command.args;                                                                                               // 208
  }                                                                                                                    // 209
  if ("executablePath" in command) {                                                                                   // 210
    if (!("deprecatedExecutablePath" in command)) {                                                                    // 211
      command.deprecatedExecutablePath = command.executablePath;                                                       // 212
    }                                                                                                                  // 213
    delete command.executablePath;                                                                                     // 214
  }                                                                                                                    // 215
                                                                                                                       // 216
  var whenReady = this._backendCap.startGrain(ownerId, grainId, packageId, command, isNew, isDev)                      // 217
      .then(function (results) {                                                                                       // 218
    return {                                                                                                           // 219
      owner: ownerId,                                                                                                  // 220
      supervisor: results.supervisor                                                                                   // 221
    };                                                                                                                 // 222
  });                                                                                                                  // 223
                                                                                                                       // 224
  this.runningGrains[grainId] = whenReady;                                                                             // 225
  return waitPromise(whenReady);                                                                                       // 226
}                                                                                                                      // 227
                                                                                                                       // 228
SandstormBackend.prototype.updateLastActive = function(grainId, userId, identityId) {                                  // 229
  // Update the lastActive date on the grain, any relevant API tokens, and the user,                                   // 230
  // and also update the user's storage usage.                                                                         // 231
                                                                                                                       // 232
  var storagePromise = undefined;                                                                                      // 233
  if (Meteor.settings.public.quotaEnabled) {                                                                           // 234
    storagePromise = globalBackend._backendCap.getUserStorageUsage(userId);                                            // 235
  }                                                                                                                    // 236
                                                                                                                       // 237
  var now = new Date();                                                                                                // 238
  if (Grains.update(grainId, {$set: {lastUsed: now}}) === 0) {                                                         // 239
    // Grain must have been deleted. Ignore.                                                                           // 240
    return;                                                                                                            // 241
  }                                                                                                                    // 242
  if (userId) {                                                                                                        // 243
    Meteor.users.update(userId, {$set: {lastActive: now}});                                                            // 244
  }                                                                                                                    // 245
  if (identityId) {                                                                                                    // 246
    Meteor.users.update({_id: identityId}, {$set: {lastActive: now}});                                                 // 247
    // Update any API tokens that match this user/grain pairing as well                                                // 248
    ApiTokens.update({"grainId": grainId, "owner.user.identityId": identityId},                                        // 249
        {$set: {"owner.user.lastUsed": now }});                                                                        // 250
  }                                                                                                                    // 251
                                                                                                                       // 252
  if (Meteor.settings.public.quotaEnabled) {                                                                           // 253
    try {                                                                                                              // 254
      var ownerId = Grains.findOne(grainId).userId;                                                                    // 255
      var size = parseInt(waitPromise(storagePromise).size);                                                           // 256
      Meteor.users.update(ownerId, {$set: {storageUsage: size}});                                                      // 257
      // TODO(security): Consider actively killing grains if the user is excessively over quota?                       // 258
      //   Otherwise a constantly-active grain could consume arbitrary space without being stopped.                    // 259
    } catch (err) {                                                                                                    // 260
      if (err.kjType !== "unimplemented") {                                                                            // 261
        console.error("error getting user storage usage:", err.stack);                                                 // 262
      }                                                                                                                // 263
    }                                                                                                                  // 264
  }                                                                                                                    // 265
}                                                                                                                      // 266
                                                                                                                       // 267
                                                                                                                       // 268
function generateSessionId(grainId, userId, salt) {                                                                    // 269
  var sessionParts = [grainId, salt];                                                                                  // 270
  if (userId) {                                                                                                        // 271
    sessionParts.push(userId);                                                                                         // 272
  }                                                                                                                    // 273
  var sessionInput = sessionParts.join(":");                                                                           // 274
  return Crypto.createHash("sha256").update(sessionInput).digest("hex");                                               // 275
}                                                                                                                      // 276
                                                                                                                       // 277
SandstormBackend.prototype.openSessionInternal = function (grainId, userId, identityId, title, apiToken, cachedSalt) {
                                                                                                                       // 279
  // Start the grain if it is not running. This is an optimization: if we didn't start it here,                        // 280
  // it would start on the first request to the session host, but we'd like to get started before                      // 281
  // the round trip.                                                                                                   // 282
  var runningGrain = this.runningGrains[grainId];                                                                      // 283
  var grainInfo;                                                                                                       // 284
  if (runningGrain) {                                                                                                  // 285
    grainInfo = waitPromise(runningGrain);                                                                             // 286
  } else {                                                                                                             // 287
    grainInfo = this.continueGrain(grainId);                                                                           // 288
  }                                                                                                                    // 289
                                                                                                                       // 290
  this.updateLastActive(grainId, userId, identityId);                                                                  // 291
                                                                                                                       // 292
  cachedSalt = cachedSalt || Random.id(22);                                                                            // 293
  var sessionId = generateSessionId(grainId, userId, cachedSalt);                                                      // 294
  var session = Sessions.findOne({_id: sessionId});                                                                    // 295
  if (session) {                                                                                                       // 296
    // TODO(someday): also do some more checks for anonymous sessions (sessions without a userId).                     // 297
    if ((session.identityId && session.identityId !== identityId) ||                                                   // 298
        (session.grainId !== grainId)) {                                                                               // 299
      var e = new Meteor.Error(500, "Duplicate SessionId");                                                            // 300
      console.error(e);                                                                                                // 301
      throw e;                                                                                                         // 302
    } else {                                                                                                           // 303
      return {sessionId: session._id, title: title, grainId: grainId, hostId: session.hostId, salt: cachedSalt};       // 304
    }                                                                                                                  // 305
  }                                                                                                                    // 306
                                                                                                                       // 307
  session = {                                                                                                          // 308
    _id: sessionId,                                                                                                    // 309
    grainId: grainId,                                                                                                  // 310
    hostId: Crypto.createHash("sha256").update(sessionId).digest("hex").slice(0, 32),                                  // 311
    timestamp: new Date().getTime(),                                                                                   // 312
    hasLoaded: false                                                                                                   // 313
  };                                                                                                                   // 314
                                                                                                                       // 315
  if (userId) {                                                                                                        // 316
    session.identityId = identityId;                                                                                   // 317
    session.userId = userId;                                                                                           // 318
  } else if (apiToken) {                                                                                               // 319
    session.hashedToken = apiToken._id;                                                                                // 320
  } else {                                                                                                             // 321
    // Must be old-style sharing, i.e. !grain.private.                                                                 // 322
  }                                                                                                                    // 323
                                                                                                                       // 324
  Sessions.insert(session);                                                                                            // 325
                                                                                                                       // 326
  return {sessionId: session._id, title: title, grainId: grainId, hostId: session.hostId, salt: cachedSalt};           // 327
}                                                                                                                      // 328
                                                                                                                       // 329
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['sandstorm-backend'] = {
  SandstormBackend: SandstormBackend
};

})();

//# sourceMappingURL=sandstorm-backend.js.map
