(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var SandstormDb = Package['sandstorm-db'].SandstormDb;

/* Package-scope variables */
var ContactProfiles;

(function(){

////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                            //
// packages/sandstorm-contacts/contacts-server.js                                             //
//                                                                                            //
////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                              //
// Sandstorm - Personal Cloud Sandbox                                                         // 1
// Copyright (c) 2016 Sandstorm Development Group, Inc. and contributors                      // 2
// All rights reserved.                                                                       // 3
//                                                                                            // 4
// Licensed under the Apache License, Version 2.0 (the "License");                            // 5
// you may not use this file except in compliance with the License.                           // 6
// You may obtain a copy of the License at                                                    // 7
//                                                                                            // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                               // 9
//                                                                                            // 10
// Unless required by applicable law or agreed to in writing, software                        // 11
// distributed under the License is distributed on an "AS IS" BASIS,                          // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                   // 13
// See the License for the specific language governing permissions and                        // 14
// limitations under the License.                                                             // 15
                                                                                              // 16
Meteor.publish("contactProfiles", function () {                                               // 17
  var db = this.connection.sandstormDb;                                                       // 18
  var self = this;                                                                            // 19
                                                                                              // 20
  // We maintain a map from identity IDs to live query handles that track profile changes.    // 21
  var contactIdentities = {};                                                                 // 22
                                                                                              // 23
  var self = this;                                                                            // 24
  function addIdentityOfContact(contact) {                                                    // 25
    if (!(contact.identityId in contactIdentities)) {                                         // 26
      var user = Meteor.users.findOne({_id: contact.identityId});                             // 27
      if (user) {                                                                             // 28
        SandstormDb.fillInProfileDefaults(user);                                              // 29
        SandstormDb.fillInIntrinsicName(user);                                                // 30
        var filteredUser = _.pick(user, "_id", "profile");                                    // 31
        self.added("contactProfiles", user._id, filteredUser);                                // 32
      }                                                                                       // 33
      contactIdentities[contact.identityId] =                                                 // 34
        Meteor.users.find({_id: contact.identityId}, {fields: {profile: 1}}).observeChanges({
          changed: function (id, fields) {                                                    // 36
            self.changed("contactProfiles", id, fields);                                      // 37
          }                                                                                   // 38
        });                                                                                   // 39
    }                                                                                         // 40
  }                                                                                           // 41
  var cursor = db.collections.contacts.find({ownerId: this.userId});                          // 42
                                                                                              // 43
  var handle = cursor.observe({                                                               // 44
    added: function (contact) {                                                               // 45
      addIdentityOfContact(contact);                                                          // 46
    },                                                                                        // 47
    changed: function (contact) {                                                             // 48
      addIdentityOfContact(contact);                                                          // 49
    },                                                                                        // 50
    removed: function (contact) {                                                             // 51
      self.removed("contactProfiles", contact.identityId);                                    // 52
      contactIdentities[contact.identityId].stop();                                           // 53
      delete contactIdentities[contact.identityId];                                           // 54
    },                                                                                        // 55
  });                                                                                         // 56
  this.ready();                                                                               // 57
                                                                                              // 58
  this.onStop(function() {                                                                    // 59
    handle.stop();                                                                            // 60
    Object.keys(contactIdentities).forEach(function(identityId) {                             // 61
      contactIdentities[identityId].stop();                                                   // 62
      delete contactIdentities[identityId];                                                   // 63
    });                                                                                       // 64
  });                                                                                         // 65
});                                                                                           // 66
                                                                                              // 67
////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['sandstorm-contacts'] = {
  ContactProfiles: ContactProfiles
};

})();

//# sourceMappingURL=sandstorm-contacts.js.map
