(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var MongoInternals = Package.mongo.MongoInternals;
var Mongo = Package.mongo.Mongo;
var Random = Package.random.Random;
var check = Package.check.check;
var Match = Package.check.Match;
var _ = Package.underscore._;
var Accounts = Package['accounts-base'].Accounts;
var AccountsServer = Package['accounts-base'].AccountsServer;
var HTTP = Package.http.HTTP;
var HTTPInternals = Package.http.HTTPInternals;

/* Package-scope variables */
var Packages, DevPackages, UserActions, Grains, RoleAssignments, Contacts, Sessions, SignupKeys, ActivityStats, DeleteStats, FileTokens, ApiTokens, Notifications, StatsTokens, Misc, Settings, Migrations, StaticAssets, AssetUploadTokens, Plans, AppIndex, KeybaseProfiles, isDemoUser, isSignedUp, isSignedUpOrDemo, successfulReferralsCount, getUserQuota, isUserOverQuota, isUserExcessivelyOverQuota, isAdmin, isAdminById, findAdminUserForToken, matchWildcardHost, makeWildcardHost, getWildcardOrigin, allowDevAccounts, sendReferralProgramNotification, roleAssignmentPattern, SandstormDb, i, appName, addStaticAsset, verifiedEmails, unverifiedEmails, userPictureUrl, fetchPicture;

(function(){

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/db.js                                                                               //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Sandstorm - Personal Cloud Sandbox                                                                        // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                     // 2
// All rights reserved.                                                                                      // 3
//                                                                                                           // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                           // 5
// you may not use this file except in compliance with the License.                                          // 6
// You may obtain a copy of the License at                                                                   // 7
//                                                                                                           // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                              // 9
//                                                                                                           // 10
// Unless required by applicable law or agreed to in writing, software                                       // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                         // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                  // 13
// See the License for the specific language governing permissions and                                       // 14
// limitations under the License.                                                                            // 15
                                                                                                             // 16
// This file defines the database schema.                                                                    // 17
                                                                                                             // 18
// Useful for debugging: Set the env variable LOG_MONGO_QUERIES to have the server write every               // 19
// query it makes, so you can see if it's doing queries too often, etc.                                      // 20
if (Meteor.isServer && process.env.LOG_MONGO_QUERIES) {                                                      // 21
  var oldFind = Mongo.Collection.prototype.find;                                                             // 22
  Mongo.Collection.prototype.find = function () {                                                            // 23
    console.log(this._prefix, arguments);                                                                    // 24
    return oldFind.apply(this, arguments);                                                                   // 25
  }                                                                                                          // 26
}                                                                                                            // 27
                                                                                                             // 28
// Helper so that we don't have to if (Meteor.isServer) before declaring indexes.                            // 29
if (Meteor.isServer) {                                                                                       // 30
  Mongo.Collection.prototype.ensureIndexOnServer = Mongo.Collection.prototype._ensureIndex;                  // 31
} else {                                                                                                     // 32
  Mongo.Collection.prototype.ensureIndexOnServer = function () {};                                           // 33
}                                                                                                            // 34
                                                                                                             // 35
// TODO(soon): Systematically go through this file and add ensureIndexOnServer() as needed.                  // 36
                                                                                                             // 37
// Users = new Mongo.Collection("users");                                                                    // 38
// The users collection is special and can be accessed through `Meteor.users`.                               // 39
// See https://docs.meteor.com/#/full/meteor_users.                                                          // 40
//                                                                                                           // 41
// There are two distinct types of entries in the users collection: identities and accounts. An              // 42
// identity contains personal profile information and typically includes some intrinsic method for           // 43
// authenticating as the owner of that information.                                                          // 44
//                                                                                                           // 45
// An account is an owner of app actions, grains, contacts, notifications, and payment info.                 // 46
// Each account can have multiple identities linked to it. To log in as an account you must first            // 47
// authenticate as one of its linked identities.                                                             // 48
//                                                                                                           // 49
// Every user contains the following fields:                                                                 // 50
//   _id: Unique string ID. For accounts, this is random. For identities, this is the globally               // 51
//        stable SHA-256 ID of this identity, hex-encoded.                                                   // 52
//   createdAt: Date when this entry was added to the collection.                                            // 53
//   lastActive: Date of the user's most recent interaction with this Sandstorm server.                      // 54
//   services: Object containing login data used by Meteor authentication services.                          // 55
//   expires: Date when this user should be deleted. Only present for demo users.                            // 56
//   upgradedFromDemo: If present, the date when this user was upgraded from being a demo user.              // 57
//                     TODO(cleanup): Unlike other dates in our database, this is stored as a number         // 58
//                     rather than as a Date object. We should fix that.                                     // 59
//   appDemoId: If present and non-null, then the user is a demo user who arrived via an /appdemo/           // 60
//              link. This field contains the app ID of the app that the user started out demoing.           // 61
//              Unlike the `expires` field, this field is not cleared when the user upgrades from            // 62
//              being a demo user.                                                                           // 63
//                                                                                                           // 64
// Identity users additionally contain the following fields:                                                 // 65
//   profile: Object containing the data that will be shared with users and grains that come into            // 66
//            contact with this identity. Includes the following fields:                                     // 67
//       service: String containing the name of this identity's authentication method.                       // 68
//       name: String containing the chosen display name of the identity.                                    // 69
//       handle: String containing the identity's preferred handle.                                          // 70
//       picture: _id into the StaticAssets table for the identity's picture. If not present,                // 71
//                an identicon will be used.                                                                 // 72
//       pronoun: One of "male", "female", "neutral", or "robot".                                            // 73
//   unverifiedEmail: If present, a string containing an email address specified by the user.                // 74
//   referredBy: ID of the Account that referred this Identity.                                              // 75
//                                                                                                           // 76
// Account users additionally contain the following fields:                                                  // 77
//   loginIdentities: Array of identity objects, each of which may include the following fields.             // 78
//       id: The globally-stable SHA-256 ID of this identity, hex-encoded.                                   // 79
//   nonloginIdentities: Array of identity objects, of the same form as `loginIdentities`. We use            // 80
//                       a separate array here so that we can use a Mongo index to enforce the               // 81
//                       invariant that an identity only be a login identity for a single account.           // 82
//   primaryEmail: String containing this account's primary email address. Must be a verified adress         // 83
//                 of one of this account's linked identities. Call SandstormDb.getUserEmails()              // 84
//                 to do this checking automatically.                                                        // 85
//   isAdmin: Boolean indicating whether this account is allowed to access the Sandstorm admin panel.        // 86
//   signupKey: If this is an invited user, then this field contains their signup key.                       // 87
//   signupNote: If the user was invited through a link, then this field contains the note that the          // 88
//               inviter admin attached to the key.                                                          // 89
//   signupEmail: If the user was invited by email, then this field contains the email address that          // 90
//                the invite was sent to.                                                                    // 91
//   hasCompletedSignup: True if this account has confirmed its profile and agreed to this server's          // 92
//                       terms of service.                                                                   // 93
//   plan: _id of an entry in the Plans table which determines the user's quota.                             // 94
//   planBonus: {storage, compute, grains} bonus amounts to add to the user's plan. The payments             // 95
//              module writes data here; we merely read it. Missing fields should be treated as              // 96
//              zeroes. Does not yet include referral bonus, which is calculated separately.                 // 97
//              TODO(cleanup): Use for referral bonus too.                                                   // 98
//   storageUsage: Number of bytes this user is currently storing.                                           // 99
//   payments: Object defined by payments module, if loaded.                                                 // 100
//   dailySentMailCount: Number of emails sent by this user today; used to limit spam.                       // 101
//   referredByComplete: ID of the Account that referred this Account. If this is set, we                    // 102
//                        stop writing new referredBy values onto Identities for this account.               // 103
//   referredCompleteDate: The Date at which the completed referral occurred.                                // 104
//   referredIdentityIds: List of Identity IDs that this Account has referred. This is used for              // 105
//                        reliably determining which Identity's names are safe to display.                   // 106
//   experiments: Object where each field is an experiment that the user is in, and each value               // 107
//           is the parameters for that experiment. Typically, the value simply names which                  // 108
//           experiment group which the user is in, where "control" is one group. If an experiment           // 109
//           is not listed, then the user should not be considered at all for the purpose of that            // 110
//           experiment. Each experiment may define a point in time where users not already in the           // 111
//           experiment may be added to it and assigned to a group (for example, at user creation            // 112
//           time). Current experiments:                                                                     // 113
//       firstTimeBillingPrompt: Value is "control" or "test". Users are assigned to groups at               // 114
//               account creation on servers where billing is enabled (i.e. Oasis). Users in the             // 115
//               test group will see a plan selection dialog and asked to make an explitic choice            // 116
//               (possibly "free") before they can create grains (but not when opening someone               // 117
//               else's shared grain). The goal of the experiment is to determine whether this               // 118
//               prompt scares users away -- and also whether it increases paid signups.                     // 119
//   stashedOldUser: A complete copy of this user from before the accounts/identities migration.             // 120
//                   TODO(cleanup): Delete this field once we're sure it's safe to do so.                    // 121
                                                                                                             // 122
Meteor.users.ensureIndexOnServer("services.google.email", {sparse: 1});                                      // 123
Meteor.users.ensureIndexOnServer("services.github.emails.email", {sparse: 1});                               // 124
Meteor.users.ensureIndexOnServer("services.email.email", {unique: 1, sparse: 1});                            // 125
Meteor.users.ensureIndexOnServer("loginIdentities.id", {unique: 1, sparse: 1});                              // 126
Meteor.users.ensureIndexOnServer("nonloginIdentities.id", {sparse: 1});                                      // 127
Meteor.users.ensureIndexOnServer("services.google.id", {unique: 1, sparse: 1});                              // 128
Meteor.users.ensureIndexOnServer("services.github.id", {unique: 1, sparse: 1});                              // 129
                                                                                                             // 130
// TODO(cleanup): This index is obsolete; delete it.                                                         // 131
Meteor.users.ensureIndexOnServer("identities.id", {unique: 1, sparse: 1});                                   // 132
                                                                                                             // 133
Packages = new Mongo.Collection("packages");                                                                 // 134
// Packages which are installed or downloading.                                                              // 135
//                                                                                                           // 136
// Each contains:                                                                                            // 137
//   _id:  128-bit prefix of SHA-256 hash of spk file, hex-encoded.                                          // 138
//   status:  String.  One of "download", "verify", "unpack", "analyze", "ready", "failed", "delete"         // 139
//   progress:  Float.  -1 = N/A, 0-1 = fractional progress (e.g. download percentage),                      // 140
//       >1 = download byte count.                                                                           // 141
//   error:  If status is "failed", error message string.                                                    // 142
//   manifest:  If status is "ready", the package manifest.  See "Manifest" in package.capnp.                // 143
//   appId:  If status is "ready", the application ID string.  Packages representing different               // 144
//       versions of the same app have the same appId.  The spk tool defines the app ID format               // 145
//       and can cryptographically verify that a package belongs to a particular app ID.                     // 146
//   shouldCleanup:  If true, a reference to this package was recently dropped, and the package              // 147
//       collector should at some point check whether there are any other references and, if not,            // 148
//       delete the package.                                                                                 // 149
//   url:  When status is "download", the URL from which the SPK can be obtained, if provided.               // 150
//   isAutoUpdated: This package was downloaded as part of an auto-update. We shouldn't clean it up          // 151
//     even if it has no users.                                                                              // 152
//   authorPgpKeyFingerprint: Verified PGP key fingerprint (SHA-1, hex, all-caps) of the app                 // 153
//     packager.                                                                                             // 154
                                                                                                             // 155
DevPackages = new Mongo.Collection("devpackages");                                                           // 156
// List of packages currently made available via the dev tools running on the local machine.                 // 157
// This is normally empty; the only time it is non-empty is when a developer is using the spk tool           // 158
// on the local machine to publish an under-development app to this server. That should only ever            // 159
// happen on developers' desktop machines.                                                                   // 160
//                                                                                                           // 161
// While a dev package is published, it automatically appears as installed by every user of the              // 162
// server, and it overrides all packages with the same application ID. If any instances of those             // 163
// packages are currently open, they are killed and reset on publish.                                        // 164
//                                                                                                           // 165
// When the dev tool disconnects, the package is automatically unpublished, and any open instances           // 166
// are again killed and refreshed.                                                                           // 167
//                                                                                                           // 168
// Each contains:                                                                                            // 169
//   _id:  The package ID string (as with Packages._id).                                                     // 170
//   appId: The app ID this package is intended to override (as with Packages.appId).                        // 171
//   timestamp:  Time when the package was last updated. If this changes while the package is                // 172
//     published, all running instances are reset. This is used e.g. to reset the app each time              // 173
//     changes are made to the source code.                                                                  // 174
//   manifest:  The app's manifest, as with Packages.manifest.                                               // 175
                                                                                                             // 176
UserActions = new Mongo.Collection("userActions");                                                           // 177
// List of actions that each user has installed which create new grains.  Each app may install               // 178
// some number of actions (usually, one).                                                                    // 179
//                                                                                                           // 180
// Each contains:                                                                                            // 181
//   _id:  random                                                                                            // 182
//   userId:  User who has installed this action.                                                            // 183
//   packageId:  Package used to run this action.                                                            // 184
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.                      // 185
//   appTitle:  Same as Packages.findOne(packageId).manifest.appTitle; denormalized so                       // 186
//       that clients can access it without subscribing to the Packages collection.                          // 187
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for                  // 188
//       searchability.                                                                                      // 189
//   appMarketingVersion:  Human-readable presentation of the app version, e.g. "2.9.17"                     // 190
//   title: JSON-encoded LocalizedText title for this action, e.g.                                           // 191
//       `{defaultText: "New Spreadsheet"}`.                                                                 // 192
//   nounPhrase: JSON-encoded LocalizedText describing what is created when this action is run.              // 193
//   command:  Manifest.Command to run this action (see package.capnp).                                      // 194
                                                                                                             // 195
Grains = new Mongo.Collection("grains");                                                                     // 196
// Grains belonging to users.                                                                                // 197
//                                                                                                           // 198
// Each contains:                                                                                            // 199
//   _id:  random                                                                                            // 200
//   packageId:  _id of the package of which this grain is an instance.                                      // 201
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.                      // 202
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for                  // 203
//       searchability.                                                                                      // 204
//   userId: The _id of the account that owns this grain.                                                    // 205
//   identityId: The identity with which the owning account prefers to open this grain.                      // 206
//   title:  Human-readable string title, as chosen by the user.                                             // 207
//   lastUsed:  Date when the grain was last used by a user.                                                 // 208
//   private: If true, then knowledge of `_id` does not suffice to open this grain.                          // 209
//   cachedViewInfo: The JSON-encoded result of `UiView.getViewInfo()`, cached from the most recent          // 210
//                   time a session to this grain was opened.                                                // 211
//                                                                                                           // 212
// The following fields *might* also exist. These are temporary hacks used to implement e-mail and           // 213
// web publishing functionality without powerbox support; they will be replaced once the powerbox            // 214
// is implemented.                                                                                           // 215
//   publicId:  An id used to publicly identify this grain. Used e.g. to route incoming e-mail and           // 216
//       web publishing. This field is initialized when first requested by the app.                          // 217
                                                                                                             // 218
RoleAssignments = new Mongo.Collection("roleAssignments");                                                   // 219
// *OBSOLETE* Before `user` was a variant of ApiTokenOwner, this collection was used to store edges          // 220
// in the permissions sharing graph. This functionality has been subsumed by the ApiTokens                   // 221
// collection.                                                                                               // 222
                                                                                                             // 223
Contacts = new Mongo.Collection("contacts");                                                                 // 224
// Edges in the social graph.                                                                                // 225
//                                                                                                           // 226
// If Alice has Bob as a contact, then she is allowed to see Bob's profile information and Bob               // 227
// will show up in her user-picker UI for actions like share-by-identity.                                    // 228
//                                                                                                           // 229
// Contacts are not symmetric. Bob might be one of Alice's contacts even if Alice is not one of              // 230
// Bob's.                                                                                                    // 231
//                                                                                                           // 232
// Each contains:                                                                                            // 233
//   _id: random                                                                                             // 234
//   ownerId: The accountId of the user account who owns this contact.                                       // 235
//   petname: Human-readable label chosen by and only visible to the owner. Uniquely identifies              // 236
//            the contact to the owner.                                                                      // 237
//   created: Date when this contact was created.                                                            // 238
//   identityId: The `_id` of the user whose contact info this contains.                                     // 239
                                                                                                             // 240
Sessions = new Mongo.Collection("sessions");                                                                 // 241
// UI sessions open to particular grains.  A new session is created each time a user opens a grain.          // 242
//                                                                                                           // 243
// Each contains:                                                                                            // 244
//   _id:  random                                                                                            // 245
//   grainId:  _id of the grain to which this session is connected.                                          // 246
//   hostId: ID part of the hostname from which this grain is being served. I.e. this replaces the           // 247
//       '*' in WILDCARD_HOST.                                                                               // 248
//   timestamp:  Time of last keep-alive message to this session.  Sessions time out after some              // 249
//       period.                                                                                             // 250
//   userId:  User ID of the user who owns this session.                                                     // 251
//   identityId:  Identity ID of the user who owns this session.                                             // 252
//   hashedToken: If the session is owned by an anonymous user, the _id of the entry in ApiTokens            // 253
//       that was used to open it. Note that for old-style sharing (i.e. when !grain.private),               // 254
//       anonymous users can get access without an API token and so neither userId nor hashedToken           // 255
//       are present.                                                                                        // 256
//   powerboxView: If present, this is a view that should be presented as part of a powerbox                 // 257
//       interaction.                                                                                        // 258
//     offer: The webkey that corresponds to cap that was passed to the `offer` RPC.                         // 259
//   viewInfo: The UiView.ViewInfo corresponding to the underlying UiSession. This isn't populated           // 260
//       until newSession is called on the UiView.                                                           // 261
//   permissions: The permissions for the current identity on this UiView. This isn't populated              // 262
//       until newSession is called on the UiView.                                                           // 263
//   hasLoaded: Marked as true by the proxy when the underlying UiSession has responded to its first         // 264
//       request                                                                                             // 265
                                                                                                             // 266
SignupKeys = new Mongo.Collection("signupKeys");                                                             // 267
// Invite keys which may be used by users to get access to Sandstorm.                                        // 268
//                                                                                                           // 269
// Each contains:                                                                                            // 270
//   _id:  random                                                                                            // 271
//   used:  Boolean indicating whether this key has already been consumed.                                   // 272
//   note:  Text note assigned when creating key, to keep track of e.g. whom the key was for.                // 273
//   email: If this key was sent as an email invite, the email address to which it was sent.                 // 274
                                                                                                             // 275
ActivityStats = new Mongo.Collection("activityStats");                                                       // 276
// Contains usage statistics taken on a regular interval. Each entry is a data point.                        // 277
//                                                                                                           // 278
// Each contains:                                                                                            // 279
//   timestamp: Date when measurements were taken.                                                           // 280
//   daily: Contains stats counts pertaining to the last day before the sample time.                         // 281
//   weekly: Contains stats counts pertaining to the last seven days before the sample time.                 // 282
//   monthly: Contains stats counts pertaining to the last thirty days before the timestamp.                 // 283
//                                                                                                           // 284
// Each of daily, weekly, and monthly contains:                                                              // 285
//   activeUsers: The number of unique users who have used a grain on the server in the time                 // 286
//       interval. Only counts logged-in users.                                                              // 287
//   demoUsers: Demo users.                                                                                  // 288
//   appDemoUsers: Users that came in through "app demo".                                                    // 289
//   activeGrains: The number of unique grains that have been used in the time interval.                     // 290
//   apps: An object indexed by app ID recording, for each app:                                              // 291
//       owners: Number of unique owners of this app (counting only grains that still exist).                // 292
//       sharedUsers: Number of users who have accessed other people's grains of this app (counting          // 293
//         only grains that still exist).                                                                    // 294
//       grains: Number of active grains of this app (that still exist).                                     // 295
//       deleted: Number of non-demo grains of this app that were deleted.                                   // 296
//       demoed: Number of demo grains created and expired.                                                  // 297
//       appDemoUsers: Number of app demos initiated with this app.                                          // 298
                                                                                                             // 299
DeleteStats = new Mongo.Collection("deleteStats");                                                           // 300
// Contains records of objects that were deleted, for stat-keeping purposes.                                 // 301
//                                                                                                           // 302
// Each contains:                                                                                            // 303
//   type: "grain" or "user" or "demoGrain" or "demoUser" or "appDemoUser"                                   // 304
//   lastActive: Date of the user's or grain's last activity.                                                // 305
//   appId: For type = "grain", the app ID of the grain. For type = "appDemoUser", the app ID they           // 306
//     arrived to demo. For others, undefined.                                                               // 307
                                                                                                             // 308
FileTokens = new Mongo.Collection("fileTokens");                                                             // 309
// Tokens corresponding to backup files that are currently stored on the server. A user receives             // 310
// a token when they create a backup file (either by uploading it, or by backing up one of their             // 311
// grains) and may use the token to read the file (either to download it, or to restore a new                // 312
// grain from it).                                                                                           // 313
//                                                                                                           // 314
// Each contains:                                                                                            // 315
//   _id:       The unguessable token string.                                                                // 316
//   name:      Suggested filename.                                                                          // 317
//   timestamp: File creation time. Used to figure out when the token and file should be wiped.              // 318
                                                                                                             // 319
ApiTokens = new Mongo.Collection("apiTokens");                                                               // 320
// Access tokens for APIs exported by apps.                                                                  // 321
//                                                                                                           // 322
// Originally API tokens were only used by external users through the HTTP API endpoint. However,            // 323
// now they are also used to implement SturdyRefs, not just held by external users, but also when            // 324
// an app holds a SturdyRef to another app within the same server. See the various `save()`,                 // 325
// `restore()`, and `drop()` methods in `grain.capnp` (on `SandstormApi`, `AppPersistent`, and               // 326
// `MainView`) -- the fields of type `Data` are API tokens.                                                  // 327
//                                                                                                           // 328
// Each contains:                                                                                            // 329
//   _id:       A SHA-256 hash of the token, base64-encoded.                                                 // 330
//   grainId:   The grain servicing this API. (Not present if the API isn't serviced by a grain.)            // 331
//   identityId: For UiView capabilities, this is the identity for which the view is attenuated.             // 332
//              That is, the UiView's newSession() method will intersect the requested permissions           // 333
//              with this identity's permissions before forwarding on to the underlying app. If              // 334
//              `identityId` is not present, then no identity attenuation is applied, i.e. this is           // 335
//              a raw UiView as implemented by the app. (The `roleAssignment` field, below, may              // 336
//              still apply. For non-UiView capabilities, `identityId` is never present. Note that           // 337
//              this is NOT the identity against which the `requiredPermissions` parameter of                // 338
//              `SandstormApi.restore()` is checked; that would be `owner.grain.introducerIdentity`.         // 339
//   accountId: For tokens where `identityId` is set, the `_id` (in the Users table) of the account          // 340
//              that created the token.                                                                      // 341
//   roleAssignment: If this API token represents a UiView, this field contains a JSON-encoded               // 342
//              Grain.ViewSharingLink.RoleAssignment representing the permissions it carries. These          // 343
//              permissions will be intersected with those held by `identityId` when the view is             // 344
//              opened.                                                                                      // 345
//   forSharing: If true, requests sent to the HTTP API endpoint with this token will be treated as          // 346
//              anonymous rather than as directly associated with `identityId`. This has no effect           // 347
//              on the permissions granted.                                                                  // 348
//   objectId:  If present, this token represents an arbitrary Cap'n Proto capability exported by            // 349
//              the app or its supervisor (whereas without this it strictly represents UiView).              // 350
//              sturdyRef is the JSON-encoded SupervisorObjectId (defined in `supervisor.capnp`).            // 351
//              Note that if the SupervisorObjectId contains an AppObjectId, that field is                   // 352
//              treated as type AnyPointer, and so encoded as a raw Cap'n Proto message.                     // 353
//   frontendRef: If present, this token actually refers to an object implemented by the front-end,          // 354
//              not a particular grain. (`grainId` and `identityId` are not set.) This is an object          // 355
//              containing exactly one of the following fields:                                              // 356
//       notificationHandle: A `Handle` for an ongoing notification, as returned by                          // 357
//                           `NotificationTarget.addOngoing`. The value is an `_id` from the                 // 358
//                           `Notifications` collection.                                                     // 359
//       ipNetwork: An IpNetwork capability that is implemented by the frontend. Eventually, this            // 360
//                  will be moved out of the frontend and into the backend, but we'll migrate the            // 361
//                  database when that happens. This field contains the boolean true to signify that         // 362
//                  it has been set.                                                                         // 363
//       ipInterface: Ditto IpNetwork, except it's an IpInterface.                                           // 364
//   parentToken: If present, then this token represents exactly the capability represented by               // 365
//              the ApiToken with _id = parentToken, except possibly (if it is a UiView) attenuated          // 366
//              by `roleAssignment` (if present). To facilitate permissions computations, if the             // 367
//              capability is a UiView, then `grainId` is set to the backing grain, `identityId`             // 368
//              is set to the identity that shared the view, and `accountId` is set to the account           // 369
//              that shared the view. Neither `objectId` nor `frontendRef` is present when                   // 370
//              `parentToken` is present.                                                                    // 371
//   petname:   Human-readable label for this access token, useful for identifying tokens for                // 372
//              revocation. This should be displayed when visualizing incoming capabilities to               // 373
//              the grain identified by `grainId`.                                                           // 374
//   created:   Date when this token was created.                                                            // 375
//   revoked:   If true, then this sturdyref has been revoked and can no longer be restored. It may          // 376
//              become un-revoked in the future.                                                             // 377
//   expires:   Optional expiration Date. If undefined, the token does not expire.                           // 378
//   owner:     A `ApiTokenOwner` (defined in `supervisor.capnp`, stored as a JSON object)                   // 379
//              as passed to the `save()` call that created this token. If not present, treat                // 380
//              as `webkey` (the default for `ApiTokenOwner`).                                               // 381
//   expiresIfUnused:                                                                                        // 382
//              Optional Date after which the token, if it has not been used yet, expires.                   // 383
//              This field should be cleared on a token's first use.                                         // 384
//   requirements: List of conditions which must hold for this token to be considered valid.                 // 385
//              Semantically, this list specifies the powers which were *used* to originally                 // 386
//              create the token. If any condition in the list becomes untrue, then the token must           // 387
//              be considered revoked, and all live refs and sturdy refs obtained transitively               // 388
//              through it must also become revoked. Each item is the JSON serialization of the              // 389
//              `MembraneRequirement` structure defined in `supervisor.capnp`.                               // 390
//                                                                                                           // 391
// It is important to note that a token's owner and provider are independent from each other. To             // 392
// illustrate, here is an approximate definition of ApiToken in pseudo Cap'n Proto schema language:          // 393
//                                                                                                           // 394
// struct ApiToken {                                                                                         // 395
//   owner :ApiTokenOwner;                                                                                   // 396
//   provider :union {                                                                                       // 397
//     grain :group {                                                                                        // 398
//       grainId :Text;                                                                                      // 399
//       union {                                                                                             // 400
//         uiView :group {                                                                                   // 401
//           identityId :Text;                                                                               // 402
//           roleAssignment :RoleAssignment;                                                                 // 403
//           forSharing :Bool;                                                                               // 404
//         }                                                                                                 // 405
//         objectId :SupervisorObjectId;                                                                     // 406
//       }                                                                                                   // 407
//     }                                                                                                     // 408
//     frontendRef :union {                                                                                  // 409
//        notificationHandle :Text;                                                                          // 410
//        ipNetwork :Bool;                                                                                   // 411
//        ipInterface :Bool;                                                                                 // 412
//     }                                                                                                     // 413
//     child :group {                                                                                        // 414
//       parentToken :Text;                                                                                  // 415
//       union {                                                                                             // 416
//         uiView :group {                                                                                   // 417
//           grainId :Text;                                                                                  // 418
//           identityId :Text;                                                                               // 419
//           roleAssignment :RoleAssignment = (allAccess = ());                                              // 420
//         }                                                                                                 // 421
//         other :Void;                                                                                      // 422
//       }                                                                                                   // 423
//     }                                                                                                     // 424
//   }                                                                                                       // 425
//   requirements: List(Supervisor.MembraneRequirement);                                                     // 426
//   ...                                                                                                     // 427
// }                                                                                                         // 428
                                                                                                             // 429
Notifications = new Mongo.Collection("notifications");                                                       // 430
// Notifications for a user.                                                                                 // 431
//                                                                                                           // 432
// Each contains:                                                                                            // 433
//   _id:          random                                                                                    // 434
//   grainId:      The grain originating this notification, if any.                                          // 435
//   userId:       The user receiving the notification.                                                      // 436
//   text:         The JSON-ified LocalizedText to display in the notification.                              // 437
//   isUnread:     Boolean indicating if this notification is unread.                                        // 438
//   timestamp:    Date when this notification was last updated                                              // 439
//   ongoing:      If present, this is an ongoing notification, and this field contains an                   // 440
//                 ApiToken referencing the `OngoingNotification` capability.                                // 441
//   admin:        If present, this is a notification intended for an admin.                                 // 442
//     action:     If present, this is a (string) link that the notification should direct the               // 443
//                 admin to.                                                                                 // 444
//     type:       The type of notification (currently only "reportStats").                                  // 445
//   appUpdates:   If present, this is an app update notification. It is an object with the appIds           // 446
//                 as keys.                                                                                  // 447
//     $appId:     The appId that has an outstanding update.                                                 // 448
//       packageId: The packageId that it will update to.                                                    // 449
//       name: The name of the app. (appTitle from package.manifest)                                         // 450
//       version: The app's version number. (appVersion from package.manifest)                               // 451
//       marketingVersion: String marketing version of this app. (appMarketingVersion from package.manifest)
//   referral:     If this boolean field is true, then treat this notification as a referral                 // 453
//                 notification. This causes text to be ignored, since we need custom logic.                 // 454
//   mailingListBonus: Like `referral`, but notify the user about the mailing list bonus. This is            // 455
//                 a one-time notification only to Oasis users who existed when the bonus program            // 456
//                 was implemented.                                                                          // 457
                                                                                                             // 458
StatsTokens = new Mongo.Collection("statsTokens");                                                           // 459
// Access tokens for the Stats collection                                                                    // 460
//                                                                                                           // 461
// These tokens are used for accessing the ActivityStats collection remotely                                 // 462
// (ie. from a dashboard webapp)                                                                             // 463
//                                                                                                           // 464
// Each contains:                                                                                            // 465
//   _id:       The token. At least 128 bits entropy (Random.id(22)).                                        // 466
                                                                                                             // 467
Misc = new Mongo.Collection("misc");                                                                         // 468
// Miscellaneous configuration and other settings                                                            // 469
//                                                                                                           // 470
// This table is currently only used for persisting BASE_URL from one session to the next,                   // 471
// but in general any miscellaneous settings should go in here                                               // 472
//                                                                                                           // 473
// Each contains:                                                                                            // 474
//   _id:       The name of the setting. eg. "BASE_URL"                                                      // 475
//   value:     The value of the setting.                                                                    // 476
                                                                                                             // 477
Settings = new Mongo.Collection("settings");                                                                 // 478
// Settings for this Sandstorm instance go here. They are configured through the adminSettings               // 479
// route. This collection differs from misc in that any admin user can update it through the admin           // 480
// interface.                                                                                                // 481
//                                                                                                           // 482
// Each contains:                                                                                            // 483
//   _id:       The name of the setting. eg. "MAIL_URL"                                                      // 484
//   value:     The value of the setting.                                                                    // 485
//   automaticallyReset: Sometimes the server needs to automatically reset a setting. When it does           // 486
//                       so, it will also write an object to this field indicating why the reset was         // 487
//                       needed. That object can have the following variants:                                // 488
//       baseUrlChangedFrom: The reset was due to BASE_URL changing. This field contains a string            // 489
//                           with the old BASE_URL.                                                          // 490
//                                                                                                           // 491
//   potentially other fields that are unique to the setting                                                 // 492
                                                                                                             // 493
Migrations = new Mongo.Collection("migrations");                                                             // 494
// This table tracks which migrations we have applied to this instance.                                      // 495
// It contains a single entry:                                                                               // 496
//   _id:       "migrations_applied"                                                                         // 497
//   value:     The number of migrations this instance has successfully completed.                           // 498
                                                                                                             // 499
StaticAssets = new Mongo.Collection("staticAssets");                                                         // 500
// Collection of static assets served up from the Sandstorm server's "static" host. We only                  // 501
// support relatively small assets: under 1MB each.                                                          // 502
//                                                                                                           // 503
// Each contains:                                                                                            // 504
//   _id:       Random ID; will be used in the URL.                                                          // 505
//   hash:      A base64-encoded SHA-256 hash of the data, used to de-dupe.                                  // 506
//   mimeType:  MIME type of the asset, suitable for Content-Type header.                                    // 507
//   encoding:  Either "gzip" or not present, suitable for Content-Encoding header.                          // 508
//   content:   The asset content (byte buffer).                                                             // 509
//   refcount:  Number of places where this asset's ID appears in the database. Since Mongo doesn't          // 510
//       have transactions, this needs to bias towards over-counting; a backup GC could be used              // 511
//       to catch leaked assets, although it's probably not a big deal in practice.                          // 512
                                                                                                             // 513
AssetUploadTokens = new Mongo.Collection("assetUploadTokens");                                               // 514
// Collection of tokens representing a single-use permission to upload an asset, such as a new               // 515
// profile picture.                                                                                          // 516
//                                                                                                           // 517
// Each contains:                                                                                            // 518
//   _id:       Random ID.                                                                                   // 519
//   purpose:   Contains one of the following, indicating how the asset is to be used:                       // 520
//       profilePicture: Indicates that the upload is a new profile picture. Contains fields:                // 521
//           userId: User whose picture shall be replaced.                                                   // 522
//           identityId: Which of the user's identities shall be updated.                                    // 523
//   expires:   Time when this token will go away if unused.                                                 // 524
                                                                                                             // 525
Plans = new Mongo.Collection("plans");                                                                       // 526
// Subscription plans, which determine quota.                                                                // 527
//                                                                                                           // 528
// Each contains:                                                                                            // 529
//   _id: Plan ID, usually a short string like "free", "standard", "large", "mega", ...                      // 530
//   storage: Number of bytes this user is allowed to store.                                                 // 531
//   compute: Number of kilobyte-RAM-seconds this user is allowed to consume.                                // 532
//   computeLabel: Label to display to the user describing this plan's compute units.                        // 533
//   grains: Total number of grains this user can create (often `Infinity`).                                 // 534
//   price: Price per month in US cents.                                                                     // 535
//   hidden: If true, a user cannot switch to this plan, but some users may be on it and are                 // 536
//       allowed to switch away.                                                                             // 537
//   title: Title from display purposes. If missing, default to capitalizing _id.                            // 538
                                                                                                             // 539
AppIndex = new Mongo.Collection("appIndex");                                                                 // 540
// A mirror of the data from the App Market index                                                            // 541
//                                                                                                           // 542
// Each contains:                                                                                            // 543
//   _id: the appId of the app                                                                               // 544
//  The rest of the fields are defined in src/sandstorm/app-index/app-index.capnp:AppIndexForMarket          // 545
                                                                                                             // 546
KeybaseProfiles = new Mongo.Collection("keybaseProfiles");                                                   // 547
// Cache of Keybase profile information. The profile for a user is re-fetched every time a package           // 548
// by that user is installed, as well as if the keybase profile is requested and not already                 // 549
// present for some reason.                                                                                  // 550
//                                                                                                           // 551
// Each contains:                                                                                            // 552
//   _id: PGP key fingerprint (SHA-1, hex, all-caps)                                                         // 553
//   displayName: Display name from Keybase. (NOT VERIFIED AT ALL.)                                          // 554
//   handle: Keybase handle.                                                                                 // 555
//   proofs: The "proofs_summary.all" array from the Keybase lookup. See the non-existent Keybase            // 556
//     docs for details. We also add a boolean "status" field to each proof indicating whether               // 557
//     we have directly verified the proof ourselves. Its values may be "unverified" (Keybase                // 558
//     returned this but we haven't checked it directly), "verified" (we verified the proof and it           // 559
//     is valid), "invalid" (we checked the proof and it was definitely bogus), or "checking" (the           // 560
//     server is currently actively checking this proof). Note that if a check fails due to network          // 561
//     errors, the status goes back to "unverified".                                                         // 562
//                                                                                                           // 563
//     WARNING: Currently verification is NOT IMPLEMENTED, so all proofs will be "unverified"                // 564
//       for now and we just trust Keybase.                                                                  // 565
                                                                                                             // 566
if (Meteor.isServer) {                                                                                       // 567
  Meteor.publish("credentials", function () {                                                                // 568
    // Data needed for isSignedUp() and isAdmin() to work.                                                   // 569
                                                                                                             // 570
    if (this.userId) {                                                                                       // 571
      return [                                                                                               // 572
        Meteor.users.find({_id: this.userId},                                                                // 573
            {fields: {signupKey: 1, isAdmin: 1, expires: 1, storageUsage: 1,                                 // 574
                      plan: 1, planBonus: 1, hasCompletedSignup: 1, experiments: 1,                          // 575
                      referredIdentityIds: 1}}),                                                             // 576
        Plans.find()                                                                                         // 577
      ];                                                                                                     // 578
    } else {                                                                                                 // 579
      return [];                                                                                             // 580
    }                                                                                                        // 581
  });                                                                                                        // 582
}                                                                                                            // 583
                                                                                                             // 584
isDemoUser = function() {                                                                                    // 585
  // Returns true if this is a demo user.                                                                    // 586
                                                                                                             // 587
  var user = Meteor.user();                                                                                  // 588
  if (user && user.expires) {                                                                                // 589
    return true;                                                                                             // 590
  } else {                                                                                                   // 591
    return false;                                                                                            // 592
  }                                                                                                          // 593
}                                                                                                            // 594
                                                                                                             // 595
isSignedUp = function() {                                                                                    // 596
  // Returns true if the user has presented an invite key.                                                   // 597
                                                                                                             // 598
  var user = Meteor.user();                                                                                  // 599
                                                                                                             // 600
  if (!user) return false;  // not signed in                                                                 // 601
                                                                                                             // 602
  if (!user.loginIdentities) return false;  // not an account                                                // 603
                                                                                                             // 604
  if (user.expires) return false;  // demo user.                                                             // 605
                                                                                                             // 606
  if (Meteor.settings.public.allowUninvited) return true;  // all accounts qualify                           // 607
                                                                                                             // 608
  if (user.signupKey) return true;  // user is invited                                                       // 609
                                                                                                             // 610
  return false;                                                                                              // 611
}                                                                                                            // 612
                                                                                                             // 613
isSignedUpOrDemo = function () {                                                                             // 614
  var user = Meteor.user();                                                                                  // 615
                                                                                                             // 616
  if (!user) return false;  // not signed in                                                                 // 617
                                                                                                             // 618
  if (!user.loginIdentities) return false;  // not an account                                                // 619
                                                                                                             // 620
  if (user.expires) return true;  // demo user.                                                              // 621
                                                                                                             // 622
  if (Meteor.settings.public.allowUninvited) return true;  // all accounts qualify                           // 623
                                                                                                             // 624
  if (user.signupKey) return true;  // user is invited                                                       // 625
                                                                                                             // 626
  return false;                                                                                              // 627
}                                                                                                            // 628
                                                                                                             // 629
var calculateReferralBonus = function(user) {                                                                // 630
  // This function returns an object of the form:                                                            // 631
  //                                                                                                         // 632
  // - {grains: 0, storage: 0}                                                                               // 633
  //                                                                                                         // 634
  // which are extra resources this account gets as part of participating in the referral                    // 635
  // program. (Storage is measured in bytes, as usual for plans.)                                            // 636
                                                                                                             // 637
  // TODO(cleanup): Consider moving referral bonus logic into Oasis payments module (since it's              // 638
  //   payments-specific) and aggregating into `planBonus`.                                                  // 639
                                                                                                             // 640
  // Authorization note: Only call this if accountId is the current user!                                    // 641
  var isPaid = (user.plan && user.plan !== "free");                                                          // 642
                                                                                                             // 643
  successfulReferralsCount = countReferrals(user);                                                           // 644
  if (isPaid) {                                                                                              // 645
    var maxPaidStorageBonus = 30 * 1e9;                                                                      // 646
    return {grains: 0,                                                                                       // 647
            storage: Math.min(                                                                               // 648
              successfulReferralsCount * 2 * 1e9,                                                            // 649
              maxPaidStorageBonus)};                                                                         // 650
  } else {                                                                                                   // 651
    var maxFreeStorageBonus = 2 * 1e9;                                                                       // 652
    var bonus = {                                                                                            // 653
      storage: Math.min(                                                                                     // 654
        successfulReferralsCount * 50 * 1e6,                                                                 // 655
        maxFreeStorageBonus)                                                                                 // 656
    };                                                                                                       // 657
    if (successfulReferralsCount > 0) {                                                                      // 658
      bonus.grains = Infinity;                                                                               // 659
    } else {                                                                                                 // 660
      bonus.grains = 0;                                                                                      // 661
    }                                                                                                        // 662
    return bonus;                                                                                            // 663
  }                                                                                                          // 664
}                                                                                                            // 665
                                                                                                             // 666
var countReferrals = function (user) {                                                                       // 667
  var referredIdentityIds = user.referredIdentityIds;                                                        // 668
  return (referredIdentityIds && referredIdentityIds.length || 0);                                           // 669
}                                                                                                            // 670
                                                                                                             // 671
getUserQuota = function (user) {                                                                             // 672
  var plan = Plans.findOne(user.plan || "free");                                                             // 673
  var referralBonus = calculateReferralBonus(user);                                                          // 674
  var bonus = user.planBonus || {};                                                                          // 675
  var userQuota = {                                                                                          // 676
    storage: plan.storage + referralBonus.storage + (bonus.storage || 0),                                    // 677
    grains: plan.grains + referralBonus.grains + (bonus.grains || 0),                                        // 678
    compute: plan.compute + (bonus.compute || 0)                                                             // 679
  };                                                                                                         // 680
  return userQuota;                                                                                          // 681
}                                                                                                            // 682
                                                                                                             // 683
isUserOverQuota = function (user) {                                                                          // 684
  // Return false if user has quota space remaining, true if it is full. When this returns true,             // 685
  // we will not allow the user to create new grains, though they may be able to open existing ones          // 686
  // which may still increase their storage usage.                                                           // 687
  //                                                                                                         // 688
  // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                        // 689
                                                                                                             // 690
  if (!Meteor.settings.public.quotaEnabled || user.isAdmin) return false;                                    // 691
                                                                                                             // 692
  var plan = getUserQuota(user);                                                                             // 693
  if (plan.grains < Infinity) {                                                                              // 694
    var count = Grains.find({userId: user._id}, {fields: {}, limit: plan.grains}).count();                   // 695
    if (count >= plan.grains) return "outOfGrains";                                                          // 696
  }                                                                                                          // 697
                                                                                                             // 698
  return plan && user.storageUsage && user.storageUsage >= plan.storage && "outOfStorage";                   // 699
}                                                                                                            // 700
                                                                                                             // 701
isUserExcessivelyOverQuota = function (user) {                                                               // 702
  // Return true if user is so far over quota that we should prevent their existing grains from              // 703
  // running at all.                                                                                         // 704
  //                                                                                                         // 705
  // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                        // 706
                                                                                                             // 707
  if (!Meteor.settings.public.quotaEnabled || user.isAdmin) return false;                                    // 708
                                                                                                             // 709
  var quota = getUserQuota(user);                                                                            // 710
                                                                                                             // 711
  // quota.grains = Infinity means unlimited grains. IEEE754 defines Infinity == Infinity.                   // 712
  if (quota.grains < Infinity) {                                                                             // 713
    var count = Grains.find({userId: user._id}, {fields: {}, limit: quota.grains * 2}).count();              // 714
    if (count >= quota.grains * 2) return "outOfGrains";                                                     // 715
  }                                                                                                          // 716
                                                                                                             // 717
  return quota && user.storageUsage && user.storageUsage >= quota.storage * 1.2 && "outOfStorage";           // 718
}                                                                                                            // 719
                                                                                                             // 720
isAdmin = function() {                                                                                       // 721
  // Returns true if the user is the administrator.                                                          // 722
                                                                                                             // 723
  var user = Meteor.user();                                                                                  // 724
  if (user && user.isAdmin) {                                                                                // 725
    return true;                                                                                             // 726
  } else {                                                                                                   // 727
    return false;                                                                                            // 728
  }                                                                                                          // 729
}                                                                                                            // 730
                                                                                                             // 731
isAdminById = function(id) {                                                                                 // 732
  // Returns true if the user's id is the administrator.                                                     // 733
                                                                                                             // 734
  var user = Meteor.users.findOne({_id: id}, {fields: {isAdmin: 1}})                                         // 735
  if (user && user.isAdmin) {                                                                                // 736
    return true;                                                                                             // 737
  } else {                                                                                                   // 738
    return false;                                                                                            // 739
  }                                                                                                          // 740
}                                                                                                            // 741
                                                                                                             // 742
findAdminUserForToken = function (token) {                                                                   // 743
  if (!token.requirements) {                                                                                 // 744
    return;                                                                                                  // 745
  }                                                                                                          // 746
  var requirements = token.requirements.filter(function (requirement) {                                      // 747
    return "userIsAdmin" in requirement;                                                                     // 748
  });                                                                                                        // 749
                                                                                                             // 750
  if (requirements.length > 1) {                                                                             // 751
    return;                                                                                                  // 752
  }                                                                                                          // 753
  if (requirements.length === 0) {                                                                           // 754
    return;                                                                                                  // 755
  }                                                                                                          // 756
  return requirements[0].userIsAdmin;                                                                        // 757
};                                                                                                           // 758
                                                                                                             // 759
var wildcardHost = Meteor.settings.public.wildcardHost.toLowerCase().split("*");                             // 760
                                                                                                             // 761
if (wildcardHost.length != 2) {                                                                              // 762
  throw new Error("Wildcard host must contain exactly one asterisk.");                                       // 763
}                                                                                                            // 764
                                                                                                             // 765
matchWildcardHost = function(host) {                                                                         // 766
  // See if the hostname is a member of our wildcard. If so, extract the ID.                                 // 767
                                                                                                             // 768
  var prefix = wildcardHost[0];                                                                              // 769
  var suffix = wildcardHost[1];                                                                              // 770
                                                                                                             // 771
  // We remove everything after the first ":" character so that our                                          // 772
  // comparison logic ignores port numbers.                                                                  // 773
  suffix = suffix.split(":")[0];                                                                             // 774
  host = host.split(":")[0];                                                                                 // 775
                                                                                                             // 776
  if (host.lastIndexOf(prefix, 0) >= 0 &&                                                                    // 777
      host.indexOf(suffix, -suffix.length) >= 0 &&                                                           // 778
      host.length >= prefix.length + suffix.length) {                                                        // 779
    var id = host.slice(prefix.length, -suffix.length);                                                      // 780
    if (id.match(/^[-a-z0-9]*$/)) {                                                                          // 781
      return id;                                                                                             // 782
    }                                                                                                        // 783
  }                                                                                                          // 784
                                                                                                             // 785
  return null;                                                                                               // 786
}                                                                                                            // 787
                                                                                                             // 788
makeWildcardHost = function (id) {                                                                           // 789
  return wildcardHost[0] + id + wildcardHost[1];                                                             // 790
}                                                                                                            // 791
                                                                                                             // 792
if (Meteor.isServer) {                                                                                       // 793
  var Url = Npm.require("url");                                                                              // 794
  getWildcardOrigin = function () {                                                                          // 795
    // The wildcard URL can be something like "foo-*-bar.example.com", but sometimes when we're              // 796
    // trying to specify a pattern matching hostnames (say, a Content-Security-Policy directive),            // 797
    // an astrisk is only allowed as the first character and must be followed by a period. So we need        // 798
    // "*.example.com" instead -- which matches more than we actually want, but is the best we can           // 799
    // really do. We also add the protocol to the front (again, that's what CSP wants).                      // 800
                                                                                                             // 801
    // TODO(cleanup): `protocol` is computed in other files, like proxy.js. Put it somewhere common.         // 802
    var protocol = Url.parse(process.env.ROOT_URL).protocol;                                                 // 803
                                                                                                             // 804
    var dotPos = wildcardHost[1].indexOf(".");                                                               // 805
    if (dotPos < 0) {                                                                                        // 806
      return protocol + "//*";                                                                               // 807
    } else {                                                                                                 // 808
      return protocol + "//*" + wildcardHost[1].slice(dotPos);                                               // 809
    }                                                                                                        // 810
  }                                                                                                          // 811
}                                                                                                            // 812
                                                                                                             // 813
allowDevAccounts = function () {                                                                             // 814
  var setting = Settings.findOne({_id: "devAccounts"});                                                      // 815
  if (setting) {                                                                                             // 816
    return setting.value;                                                                                    // 817
  } else {                                                                                                   // 818
    return Meteor.settings && Meteor.settings.public &&                                                      // 819
           Meteor.settings.public.allowDevAccounts;                                                          // 820
  }                                                                                                          // 821
};                                                                                                           // 822
                                                                                                             // 823
sendReferralProgramNotification = function (userId) {                                                        // 824
  Notifications.upsert({                                                                                     // 825
    userId: userId,                                                                                          // 826
    referral: true,                                                                                          // 827
  }, {                                                                                                       // 828
    userId: userId,                                                                                          // 829
    referral: true,                                                                                          // 830
    timestamp: new Date(),                                                                                   // 831
    isUnread: true,                                                                                          // 832
  });                                                                                                        // 833
};                                                                                                           // 834
                                                                                                             // 835
roleAssignmentPattern = {                                                                                    // 836
  none : Match.Optional(null),                                                                               // 837
  allAccess: Match.Optional(null),                                                                           // 838
  roleId: Match.Optional(Match.Integer),                                                                     // 839
  addPermissions: Match.Optional([Boolean]),                                                                 // 840
  removePermissions: Match.Optional([Boolean]),                                                              // 841
};                                                                                                           // 842
                                                                                                             // 843
SandstormDb = function () {                                                                                  // 844
  this.collections = {                                                                                       // 845
    // Direct access to underlying collections. DEPRECATED.                                                  // 846
    //                                                                                                       // 847
    // TODO(cleanup): Over time, we will provide methods covering each supported query and remove            // 848
    //   direct access to the collections.                                                                   // 849
                                                                                                             // 850
    packages: Packages,                                                                                      // 851
    devPackages: DevPackages,                                                                                // 852
    userActions: UserActions,                                                                                // 853
    grains: Grains,                                                                                          // 854
    contacts: Contacts,                                                                                      // 855
    sessions: Sessions,                                                                                      // 856
    signupKeys: SignupKeys,                                                                                  // 857
    activityStats: ActivityStats,                                                                            // 858
    deleteStats: DeleteStats,                                                                                // 859
    fileTokens: FileTokens,                                                                                  // 860
    apiTokens: ApiTokens,                                                                                    // 861
    notifications: Notifications,                                                                            // 862
    statsTokens: StatsTokens,                                                                                // 863
    misc: Misc,                                                                                              // 864
    settings: Settings,                                                                                      // 865
    appIndex: AppIndex,                                                                                      // 866
    keybaseProfiles: KeybaseProfiles,                                                                        // 867
                                                                                                             // 868
    // Intentionally omitted:                                                                                // 869
    // - Migrations, since it's used only within this package.                                               // 870
    // - RoleAssignments, since it is deprecated and only used by the migration that eliminated it.          // 871
  };                                                                                                         // 872
};                                                                                                           // 873
                                                                                                             // 874
// TODO(cleanup): These methods should not be defined freestanding and should use collection                 // 875
//   objects created in SandstormDb's constructor rather than globals.                                       // 876
                                                                                                             // 877
_.extend(SandstormDb.prototype, {                                                                            // 878
  isDemoUser: isDemoUser,                                                                                    // 879
  isSignedUp: isSignedUp,                                                                                    // 880
  isSignedUpOrDemo: isSignedUpOrDemo,                                                                        // 881
  getUserQuota: getUserQuota,                                                                                // 882
  isUserOverQuota: isUserOverQuota,                                                                          // 883
  isUserExcessivelyOverQuota: isUserExcessivelyOverQuota,                                                    // 884
  isAdmin: isAdmin,                                                                                          // 885
  isAdminById: isAdminById,                                                                                  // 886
  findAdminUserForToken: findAdminUserForToken,                                                              // 887
  matchWildcardHost: matchWildcardHost,                                                                      // 888
  makeWildcardHost: makeWildcardHost,                                                                        // 889
  allowDevAccounts: allowDevAccounts,                                                                        // 890
  roleAssignmentPattern: roleAssignmentPattern,                                                              // 891
});                                                                                                          // 892
                                                                                                             // 893
if (Meteor.isServer) {                                                                                       // 894
  SandstormDb.prototype.getWildcardOrigin = getWildcardOrigin;                                               // 895
}                                                                                                            // 896
                                                                                                             // 897
// =======================================================================================                   // 898
// Below this point are newly-written or refactored functions.                                               // 899
                                                                                                             // 900
_.extend(SandstormDb.prototype, {                                                                            // 901
  getUser: function getUser (userId) {                                                                       // 902
    check(userId, Match.OneOf(String, undefined, null));                                                     // 903
    if (userId) {                                                                                            // 904
      return Meteor.users.findOne(userId);                                                                   // 905
    }                                                                                                        // 906
  },                                                                                                         // 907
                                                                                                             // 908
  getIdentity: function getIdentity (identityId) {                                                           // 909
    check(identityId, String);                                                                               // 910
    var identity = Meteor.users.findOne({_id: identityId});                                                  // 911
    if (identity) {                                                                                          // 912
      SandstormDb.fillInProfileDefaults(identity);                                                           // 913
      SandstormDb.fillInIntrinsicName(identity);                                                             // 914
      SandstormDb.fillInPictureUrl(identity);                                                                // 915
      return identity;                                                                                       // 916
    }                                                                                                        // 917
  },                                                                                                         // 918
                                                                                                             // 919
  userHasIdentity: function (userId, identityId) {                                                           // 920
    check(userId, String);                                                                                   // 921
    check(identityId, String);                                                                               // 922
                                                                                                             // 923
    if (userId === identityId) return true;                                                                  // 924
                                                                                                             // 925
    var user = Meteor.users.findOne(userId);                                                                 // 926
    return SandstormDb.getUserIdentityIds(user).indexOf(identityId) != -1;                                   // 927
  },                                                                                                         // 928
                                                                                                             // 929
  userGrains: function userGrains (userId) {                                                                 // 930
    check(userId, Match.OneOf(String, undefined, null));                                                     // 931
    return this.collections.grains.find({userId: userId});                                                   // 932
  },                                                                                                         // 933
                                                                                                             // 934
  currentUserGrains: function currentUserGrains () {                                                         // 935
    return this.userGrains(Meteor.userId());                                                                 // 936
  },                                                                                                         // 937
                                                                                                             // 938
  getGrain: function getGrain (grainId) {                                                                    // 939
    check(grainId, String);                                                                                  // 940
    return this.collections.grains.findOne(grainId);                                                         // 941
  },                                                                                                         // 942
                                                                                                             // 943
  userApiTokens: function userApiTokens (userId) {                                                           // 944
    check(userId, Match.OneOf(String, undefined, null));                                                     // 945
    var identityIds = SandstormDb.getUserIdentityIds(this.getUser(userId));                                  // 946
    return this.collections.apiTokens.find({'owner.user.identityId': {$in: identityIds}});                   // 947
  },                                                                                                         // 948
                                                                                                             // 949
  currentUserApiTokens: function currentUserApiTokens () {                                                   // 950
    return this.userApiTokens(Meteor.userId());                                                              // 951
  },                                                                                                         // 952
                                                                                                             // 953
  userActions: function userActions (user) {                                                                 // 954
    return this.collections.userActions.find({userId: user});                                                // 955
  },                                                                                                         // 956
                                                                                                             // 957
  currentUserActions: function currentUserActions () {                                                       // 958
    return this.userActions(Meteor.userId());                                                                // 959
  },                                                                                                         // 960
                                                                                                             // 961
  iconSrcForPackage: function iconSrcForPackage (pkg, usage) {                                               // 962
    return Identicon.iconSrcForPackage(pkg, usage, this.makeWildcardHost("static"));                         // 963
  },                                                                                                         // 964
                                                                                                             // 965
  getDenormalizedGrainInfo: function getDenormalizedGrainInfo(grainId) {                                     // 966
    var grain = this.getGrain(grainId);                                                                      // 967
    var pkg = this.collections.packages.findOne(grain.packageId);                                            // 968
    var appTitle = (pkg && pkg.manifest && pkg.manifest.appTitle) || { defaultText: ""};                     // 969
    var grainInfo = { appTitle: appTitle };                                                                  // 970
                                                                                                             // 971
    if (pkg && pkg.manifest && pkg.manifest.metadata && pkg.manifest.metadata.icons) {                       // 972
      var icons = pkg.manifest.metadata.icons;                                                               // 973
      grainInfo.icon = icons.grain || icons.appGrid;                                                         // 974
    }                                                                                                        // 975
                                                                                                             // 976
    // Only provide an app ID if we have no icon asset to provide and need to offer an identicon.            // 977
    if (!grainInfo.icon && pkg) {                                                                            // 978
      grainInfo.appId = pkg.appId;                                                                           // 979
    }                                                                                                        // 980
                                                                                                             // 981
    return grainInfo;                                                                                        // 982
  },                                                                                                         // 983
                                                                                                             // 984
  getPlan: function (id) {                                                                                   // 985
    check(id, String);                                                                                       // 986
    var plan = Plans.findOne(id);                                                                            // 987
    if (!plan) {                                                                                             // 988
      throw new Error("no such plan: " + id);                                                                // 989
    }                                                                                                        // 990
    return plan;                                                                                             // 991
  },                                                                                                         // 992
                                                                                                             // 993
  listPlans: function () {                                                                                   // 994
    return Plans.find({}, {sort: {price: 1}});                                                               // 995
  },                                                                                                         // 996
                                                                                                             // 997
  getMyPlan: function () {                                                                                   // 998
    var user = Meteor.user();                                                                                // 999
    return user && Plans.findOne(user.plan || "free");                                                       // 1000
  },                                                                                                         // 1001
                                                                                                             // 1002
  getMyReferralBonus: function(user) {                                                                       // 1003
    // This function is called from the server and from the client, similar to getMyPlan().                  // 1004
    //                                                                                                       // 1005
    // The parameter may be omitted in which case the current user is assumed.                               // 1006
                                                                                                             // 1007
    return calculateReferralBonus(user || Meteor.user());                                                    // 1008
  },                                                                                                         // 1009
                                                                                                             // 1010
  getMyUsage: function (user) {                                                                              // 1011
    user = user || Meteor.user();                                                                            // 1012
    if (user && (Meteor.isServer || user.pseudoUsage)) {                                                     // 1013
      if (Meteor.isClient) {                                                                                 // 1014
        // Filled by pseudo-subscription to "getMyUsage". WARNING: The subscription is currenly              // 1015
        // not reactive.                                                                                     // 1016
        return user.pseudoUsage;                                                                             // 1017
      } else {                                                                                               // 1018
        return {                                                                                             // 1019
          grains: Grains.find({userId: user._id}).count(),                                                   // 1020
          storage: user.storageUsage || 0,                                                                   // 1021
          compute: 0   // not tracked yet                                                                    // 1022
        };                                                                                                   // 1023
      }                                                                                                      // 1024
    } else {                                                                                                 // 1025
      return {grains: 0, storage: 0, compute: 0};                                                            // 1026
    }                                                                                                        // 1027
  },                                                                                                         // 1028
                                                                                                             // 1029
  isUninvitedFreeUser: function () {                                                                         // 1030
    if (!Meteor.settings.public.allowUninvited) return false;                                                // 1031
                                                                                                             // 1032
    var user = Meteor.user();                                                                                // 1033
    return user && !user.expires && (!user.plan || user.plan === "free");                                    // 1034
  },                                                                                                         // 1035
                                                                                                             // 1036
  getSetting: function (name) {                                                                              // 1037
    var setting = Settings.findOne(name);                                                                    // 1038
    return setting && setting.value;                                                                         // 1039
  },                                                                                                         // 1040
                                                                                                             // 1041
  addUserActions: function(packageId) {                                                                      // 1042
    //TODO(cleanup): implement this with meteor methods rather than client-side inserts/removes.             // 1043
    var pack = Packages.findOne(packageId);                                                                  // 1044
    if (pack) {                                                                                              // 1045
      // Remove old versions.                                                                                // 1046
      UserActions.find({userId: Meteor.userId(), appId: pack.appId})                                         // 1047
          .forEach(function (action) {                                                                       // 1048
        UserActions.remove(action._id);                                                                      // 1049
      });                                                                                                    // 1050
                                                                                                             // 1051
      // Install new.                                                                                        // 1052
      var actions = pack.manifest.actions;                                                                   // 1053
      for (i in actions) {                                                                                   // 1054
        var action = actions[i];                                                                             // 1055
        if ("none" in action.input) {                                                                        // 1056
          var userAction = {                                                                                 // 1057
            userId: Meteor.userId(),                                                                         // 1058
            packageId: pack._id,                                                                             // 1059
            appId: pack.appId,                                                                               // 1060
            appTitle: pack.manifest.appTitle,                                                                // 1061
            appMarketingVersion: pack.manifest.appMarketingVersion,                                          // 1062
            appVersion: pack.manifest.appVersion,                                                            // 1063
            title: action.title,                                                                             // 1064
            nounPhrase: action.nounPhrase,                                                                   // 1065
            command: action.command                                                                          // 1066
          };                                                                                                 // 1067
          UserActions.insert(userAction);                                                                    // 1068
        } else {                                                                                             // 1069
          // TODO(someday):  Implement actions with capability inputs.                                       // 1070
        }                                                                                                    // 1071
      }                                                                                                      // 1072
                                                                                                             // 1073
      Meteor.call("deleteUnusedPackages", pack.appId);                                                       // 1074
    }                                                                                                        // 1075
  },                                                                                                         // 1076
                                                                                                             // 1077
  sendAdminNotification: function (message, link) {                                                          // 1078
    Meteor.users.find({isAdmin: true}, {fields: {_id: 1}}).forEach(function (user) {                         // 1079
      Notifications.insert({                                                                                 // 1080
        admin: {                                                                                             // 1081
          action: link,                                                                                      // 1082
          type: "reportStats",                                                                               // 1083
        },                                                                                                   // 1084
        userId: user._id,                                                                                    // 1085
        text: {defaultText: message},                                                                        // 1086
        timestamp: new Date(),                                                                               // 1087
        isUnread: true,                                                                                      // 1088
      });                                                                                                    // 1089
    });                                                                                                      // 1090
  },                                                                                                         // 1091
                                                                                                             // 1092
  getKeybaseProfile: function (keyFingerprint) {                                                             // 1093
    return this.collections.keybaseProfiles.findOne(keyFingerprint) || {};                                   // 1094
  },                                                                                                         // 1095
                                                                                                             // 1096
  getServerTitle: function () {                                                                              // 1097
    var setting = Settings.findOne({_id: "serverTitle"});                                                    // 1098
    return setting ? setting.value : "";  // empty if subscription is not ready.                             // 1099
  },                                                                                                         // 1100
                                                                                                             // 1101
  getReturnAddress: function () {                                                                            // 1102
    var setting = Settings.findOne({_id: "returnAddress"});                                                  // 1103
    return setting ? setting.value : "";  // empty if subscription is not ready.                             // 1104
  }                                                                                                          // 1105
});                                                                                                          // 1106
                                                                                                             // 1107
var appNameFromPackage = function(packageObj) {                                                              // 1108
  // This function takes a Package object from Mongo and returns an                                          // 1109
  // app title.                                                                                              // 1110
  var manifest = packageObj.manifest;                                                                        // 1111
  if (!manifest) return packageObj.appId || packageObj._id || "unknown";                                     // 1112
  var action = manifest.actions[0];                                                                          // 1113
  appName = (manifest.appTitle && manifest.appTitle.defaultText) ||                                          // 1114
    appNameFromActionName(action.title.defaultText);                                                         // 1115
  return appName;                                                                                            // 1116
};                                                                                                           // 1117
                                                                                                             // 1118
var appNameFromActionName = function(name) {                                                                 // 1119
  // Hack: Historically we only had action titles, like "New Etherpad Document", not app                     // 1120
  //   titles. But for this UI we want app titles. As a transitionary measure, try to                        // 1121
  //   derive the app title from the action title.                                                           // 1122
  // TODO(cleanup): Get rid of this once apps have real titles.                                              // 1123
  if (!name) {                                                                                               // 1124
    return "(unnamed)";                                                                                      // 1125
  }                                                                                                          // 1126
  if (name.lastIndexOf("New ", 0) === 0) {                                                                   // 1127
    name = name.slice(4);                                                                                    // 1128
  }                                                                                                          // 1129
  if (name.lastIndexOf("Hacker CMS", 0) === 0) {                                                             // 1130
    name = "Hacker CMS";                                                                                     // 1131
  } else {                                                                                                   // 1132
    var space = name.indexOf(" ");                                                                           // 1133
    if (space > 0) {                                                                                         // 1134
      name = name.slice(0, space);                                                                           // 1135
    }                                                                                                        // 1136
  }                                                                                                          // 1137
  return name;                                                                                               // 1138
};                                                                                                           // 1139
                                                                                                             // 1140
var appShortDescriptionFromPackage = function (pkg) {                                                        // 1141
  return pkg && pkg.manifest && pkg.manifest.metadata &&                                                     // 1142
         pkg.manifest.metadata.shortDescription &&                                                           // 1143
         pkg.manifest.metadata.shortDescription.defaultText;                                                 // 1144
};                                                                                                           // 1145
                                                                                                             // 1146
var nounPhraseForActionAndAppTitle = function(action, appTitle) {                                            // 1147
  // A hack to deal with legacy apps not including fields in their manifests.                                // 1148
  // I look forward to the day I can remove most of this code.                                               // 1149
  // Attempt to figure out the appropriate noun that this action will create.                                // 1150
  // Use an explicit noun phrase is one is available.  Apps should add these in the future.                  // 1151
  if (action.nounPhrase) return action.nounPhrase.defaultText;                                               // 1152
  // Otherwise, try to guess one from the structure of the action title field                                // 1153
  if (action.title && action.title.defaultText) {                                                            // 1154
    var text = action.title.defaultText;                                                                     // 1155
    // Strip a leading "New "                                                                                // 1156
    if (text.lastIndexOf("New ", 0) === 0) {                                                                 // 1157
      var candidate = text.slice(4);                                                                         // 1158
      // Strip a leading appname too, if provided                                                            // 1159
      if (candidate.lastIndexOf(appTitle, 0) === 0) {                                                        // 1160
        var newCandidate = candidate.slice(appTitle.length);                                                 // 1161
        // Unless that leaves you with no noun, in which case, use "grain"                                   // 1162
        if (newCandidate.length > 0) {                                                                       // 1163
          return newCandidate.toLowerCase();                                                                 // 1164
        } else {                                                                                             // 1165
          return "grain";                                                                                    // 1166
        }                                                                                                    // 1167
      }                                                                                                      // 1168
      return candidate.toLowerCase();                                                                        // 1169
    }                                                                                                        // 1170
    // Some other verb phrase was given.  Just use it verbatim, and hope the app author updates              // 1171
    // the package soon.                                                                                     // 1172
    return text;                                                                                             // 1173
  } else {                                                                                                   // 1174
    return "grain";                                                                                          // 1175
  }                                                                                                          // 1176
};                                                                                                           // 1177
                                                                                                             // 1178
// Static methods on SandstormDb that don't need an instance.                                                // 1179
// Largely things that deal with backwards-compatibility.                                                    // 1180
_.extend(SandstormDb, {                                                                                      // 1181
  appNameFromActionName: appNameFromActionName,                                                              // 1182
  appNameFromPackage: appNameFromPackage,                                                                    // 1183
  appShortDescriptionFromPackage: appShortDescriptionFromPackage,                                            // 1184
  nounPhraseForActionAndAppTitle: nounPhraseForActionAndAppTitle,                                            // 1185
});                                                                                                          // 1186
                                                                                                             // 1187
if (Meteor.isServer) {                                                                                       // 1188
  var Crypto = Npm.require("crypto");                                                                        // 1189
  var ContentType = Npm.require("content-type");                                                             // 1190
  var Zlib = Npm.require("zlib");                                                                            // 1191
                                                                                                             // 1192
  var replicaNumber = Meteor.settings.replicaNumber || 0;                                                    // 1193
                                                                                                             // 1194
  var computeStagger = function (n) {                                                                        // 1195
    // Compute a fraction in the range [0, 1) such that, for any natural number k, the values                // 1196
    // of computeStagger(n) for all n in [1, 2^k) are uniformly distributed between 0 and 1.                 // 1197
    // The sequence looks like:                                                                              // 1198
    //   0, 1/2, 1/4, 3/4, 1/8, 3/8, 5/8, 7/8, 1/16, ...                                                     // 1199
    //                                                                                                       // 1200
    // We use this to determine how we'll stagger periodic events performed by this replica.                 // 1201
    // Notice that this allows us to compute a stagger which is independent of the number of                 // 1202
    // front-end replicas present; we can add more replicas to the end without affecting how the             // 1203
    // earlier ones schedule their events.                                                                   // 1204
    var denom = 1;                                                                                           // 1205
    while (denom <= n) denom <<= 1;                                                                          // 1206
    var num = n * 2 - denom + 1;                                                                             // 1207
    return num / denom;                                                                                      // 1208
  }                                                                                                          // 1209
                                                                                                             // 1210
  var stagger = computeStagger(replicaNumber);                                                               // 1211
                                                                                                             // 1212
  SandstormDb.periodicCleanup = function (intervalMs, callback) {                                            // 1213
    // Register a database cleanup function than should run periodically, roughly once every                 // 1214
    // interval of the given length.                                                                         // 1215
    //                                                                                                       // 1216
    // In a blackrock deployment with multiple front-ends, the frequency of the cleanup will be              // 1217
    // scaled appropriately on the assumption that more data is being generated demanding more               // 1218
    // frequent cleanups.                                                                                    // 1219
                                                                                                             // 1220
    check(intervalMs, Number);                                                                               // 1221
    check(callback, Function);                                                                               // 1222
                                                                                                             // 1223
    if (intervalMs < 120000) {                                                                               // 1224
      throw new Error("less than 2-minute cleanup interval seems too fast; " +                               // 1225
                      "are you using the right units?");                                                     // 1226
    }                                                                                                        // 1227
                                                                                                             // 1228
    // Schedule first cleanup to happen at the next intervalMs interval from the epoch, so that              // 1229
    // the schedule is independent of the exact startup time.                                                // 1230
    var first = intervalMs - Date.now() % intervalMs;                                                        // 1231
                                                                                                             // 1232
    // Stagger cleanups across replicas so that we don't have all replicas trying to clean the               // 1233
    // same data at the same time.                                                                           // 1234
    first += Math.floor(intervalMs * computeStagger(replicaNumber));                                         // 1235
                                                                                                             // 1236
    // If the stagger put us more than an interval away from now, back up.                                   // 1237
    if (first > intervalMs) first -= intervalMs;                                                             // 1238
                                                                                                             // 1239
    Meteor.setTimeout(function () {                                                                          // 1240
      callback();                                                                                            // 1241
      Meteor.setInterval(callback, intervalMs);                                                              // 1242
    }, first);                                                                                               // 1243
  }                                                                                                          // 1244
                                                                                                             // 1245
  // TODO(cleanup): Node 0.12 has a `gzipSync` but 0.10 (which Meteor still uses) does not.                  // 1246
  var gzipSync = Meteor.wrapAsync(Zlib.gzip, Zlib);                                                          // 1247
                                                                                                             // 1248
  var BufferSmallerThan = function (limit) {                                                                 // 1249
    return Match.Where(function (buf) {                                                                      // 1250
      check(buf, Buffer);                                                                                    // 1251
      return buf.length < limit;                                                                             // 1252
    });                                                                                                      // 1253
  }                                                                                                          // 1254
                                                                                                             // 1255
  var DatabaseId = Match.Where(function (s) {                                                                // 1256
    check(s, String);                                                                                        // 1257
    return !!s.match(/^[a-zA-Z0-9_]+$/);                                                                     // 1258
  });                                                                                                        // 1259
                                                                                                             // 1260
  addStaticAsset = function (metadata, content) {                                                            // 1261
    // Add a new static asset to the database. If `content` is a string rather than a buffer, it             // 1262
    // will be automatically gzipped before storage; do not specify metadata.encoding in this case.          // 1263
                                                                                                             // 1264
    if (typeof content === "string" && !metadata.encoding) {                                                 // 1265
      content = gzipSync(new Buffer(content, "utf8"));                                                       // 1266
      metadata.encoding = "gzip";                                                                            // 1267
    }                                                                                                        // 1268
                                                                                                             // 1269
    check(metadata, {                                                                                        // 1270
      mimeType: String,                                                                                      // 1271
      encoding: Match.Optional("gzip")                                                                       // 1272
    });                                                                                                      // 1273
    check(content, BufferSmallerThan(1 << 20));                                                              // 1274
                                                                                                             // 1275
    // Validate content type.                                                                                // 1276
    metadata.mimeType = ContentType.format(ContentType.parse(metadata.mimeType));                            // 1277
                                                                                                             // 1278
    var hasher = Crypto.createHash("sha256");                                                                // 1279
    hasher.update(metadata.mimeType + "\n" + metadata.encoding + "\n", "utf8");                              // 1280
    hasher.update(content);                                                                                  // 1281
    var hash = hasher.digest("base64");                                                                      // 1282
                                                                                                             // 1283
    var existing = StaticAssets.findAndModify({                                                              // 1284
      query: {hash: hash, refcount: {$gte: 1}},                                                              // 1285
      update: {$inc: {refcount: 1}},                                                                         // 1286
      fields: {_id: 1, refcount: 1},                                                                         // 1287
    });                                                                                                      // 1288
    if (existing) {                                                                                          // 1289
      return existing._id;                                                                                   // 1290
    }                                                                                                        // 1291
                                                                                                             // 1292
    return StaticAssets.insert(_.extend({                                                                    // 1293
      hash: hash,                                                                                            // 1294
      content: content,                                                                                      // 1295
      refcount: 1                                                                                            // 1296
    }, metadata));                                                                                           // 1297
  }                                                                                                          // 1298
                                                                                                             // 1299
  SandstormDb.prototype.addStaticAsset = addStaticAsset;                                                     // 1300
                                                                                                             // 1301
  SandstormDb.prototype.refStaticAsset = function (id) {                                                     // 1302
    // Increment the refcount on an existing static asset.                                                   // 1303
    //                                                                                                       // 1304
    // You must call this BEFORE adding the new reference to the DB, in case of failure between              // 1305
    // the two calls. (This way, the failure case is a storage leak, which is probably not a big             // 1306
    // deal and can be fixed by GC, rather than a mysteriously missing asset.)                               // 1307
                                                                                                             // 1308
    check(id, String);                                                                                       // 1309
                                                                                                             // 1310
    var existing = StaticAssets.findAndModify({                                                              // 1311
      query: {hash: hash},                                                                                   // 1312
      update: {$inc: {refcount: 1}},                                                                         // 1313
      fields: {_id: 1, refcount: 1},                                                                         // 1314
    });                                                                                                      // 1315
    if (!existing) {                                                                                         // 1316
      throw new Error("refStaticAsset() called on asset that doesn't exist");                                // 1317
    }                                                                                                        // 1318
  }                                                                                                          // 1319
                                                                                                             // 1320
  SandstormDb.prototype.unrefStaticAsset = function (id) {                                                   // 1321
    // Decrement refcount on a static asset and delete if it has reached zero.                               // 1322
    //                                                                                                       // 1323
    // You must call this AFTER removing the reference from the DB, in case of failure between               // 1324
    // the two calls. (This way, the failure case is a storage leak, which is probably not a big             // 1325
    // deal and can be fixed by GC, rather than a mysteriously missing asset.)                               // 1326
                                                                                                             // 1327
    check(id, String);                                                                                       // 1328
                                                                                                             // 1329
    var existing = StaticAssets.findAndModify({                                                              // 1330
      query: {_id: id},                                                                                      // 1331
      update: {$inc: {refcount: -1}},                                                                        // 1332
      fields: {_id: 1, refcount: 1},                                                                         // 1333
      new: true,                                                                                             // 1334
    });                                                                                                      // 1335
    if (!existing) {                                                                                         // 1336
      console.error(new Error("unrefStaticAsset() called on asset that doesn't exist").stack);               // 1337
    } else if (existing.refcount <= 0) {                                                                     // 1338
      StaticAssets.remove({_id: existing._id});                                                              // 1339
    }                                                                                                        // 1340
  }                                                                                                          // 1341
                                                                                                             // 1342
  SandstormDb.prototype.getStaticAsset = function (id) {                                                     // 1343
    // Get a static asset's mimeType, encoding, and raw content.                                             // 1344
                                                                                                             // 1345
    check(id, String);                                                                                       // 1346
                                                                                                             // 1347
    var asset = StaticAssets.findOne(id, {fields: {_id: 0, mimeType: 1, encoding: 1, content: 1}});          // 1348
    if (asset) {                                                                                             // 1349
      // TODO(perf): Mongo converts buffers to something else. Figure out a way to avoid a copy              // 1350
      //   here.                                                                                             // 1351
      asset.content = new Buffer(asset.content);                                                             // 1352
    }                                                                                                        // 1353
    return asset;                                                                                            // 1354
  }                                                                                                          // 1355
                                                                                                             // 1356
  SandstormDb.prototype.newAssetUpload = function (purpose) {                                                // 1357
    check(purpose, {profilePicture: {userId: DatabaseId, identityId: DatabaseId}});                          // 1358
                                                                                                             // 1359
    return AssetUploadTokens.insert({                                                                        // 1360
      purpose: purpose,                                                                                      // 1361
      expires: new Date(Date.now() + 300000),  // in 5 minutes                                               // 1362
    });                                                                                                      // 1363
  }                                                                                                          // 1364
                                                                                                             // 1365
  SandstormDb.prototype.fulfillAssetUpload = function (id) {                                                 // 1366
    // Indicates that the given asset upload has completed. It will be removed and its purpose               // 1367
    // returned. If no matching upload exists, returns undefined.                                            // 1368
                                                                                                             // 1369
    check(id, String);                                                                                       // 1370
                                                                                                             // 1371
    var upload = AssetUploadTokens.findAndModify({                                                           // 1372
      query: {_id: id},                                                                                      // 1373
      remove: true                                                                                           // 1374
    });                                                                                                      // 1375
                                                                                                             // 1376
    if (upload.expires.valueOf() < Date.now()) {                                                             // 1377
      return undefined;  // already expired                                                                  // 1378
    } else {                                                                                                 // 1379
      return upload.purpose;                                                                                 // 1380
    }                                                                                                        // 1381
  }                                                                                                          // 1382
                                                                                                             // 1383
  function cleanupExpiredAssetUploads() {                                                                    // 1384
    AssetUploadTokens.remove({expires: {$lt: Date.now()}});                                                  // 1385
  }                                                                                                          // 1386
                                                                                                             // 1387
  // Cleanup tokens every hour.                                                                              // 1388
  SandstormDb.periodicCleanup(3600000, cleanupExpiredAssetUploads);                                          // 1389
                                                                                                             // 1390
  var packageCache = {};                                                                                     // 1391
  // Package info is immutable. Let's cache to save on mongo queries.                                        // 1392
                                                                                                             // 1393
  SandstormDb.prototype.getPackage = function (packageId) {                                                  // 1394
    // Get the given package record. Since package info is immutable, cache the data in the server           // 1395
    // to reduce mongo query overhead, since it turns out we have to fetch specific packages a               // 1396
    // lot.                                                                                                  // 1397
                                                                                                             // 1398
    if (packageId in packageCache) {                                                                         // 1399
      return packageCache[packageId];                                                                        // 1400
    }                                                                                                        // 1401
                                                                                                             // 1402
    var pkg = Packages.findOne(packageId);                                                                   // 1403
    if (pkg && pkg.status === "ready") {                                                                     // 1404
      packageCache[packageId] = pkg;                                                                         // 1405
    }                                                                                                        // 1406
    return pkg;                                                                                              // 1407
  }                                                                                                          // 1408
                                                                                                             // 1409
  SandstormDb.prototype.sendAppUpdateNotifications = function (appId, packageId, name,                       // 1410
                                                               versionNumber, marketingVersion) {            // 1411
    var db = this;                                                                                           // 1412
    var actions = db.collections.userActions.find({appId: appId, appVersion: {$lt: versionNumber}},          // 1413
      {fields: {userId: 1}});                                                                                // 1414
    actions.forEach(function (action) {                                                                      // 1415
      var userId = action.userId;                                                                            // 1416
      var updater = {                                                                                        // 1417
        userId: userId,                                                                                      // 1418
        timestamp: new Date(),                                                                               // 1419
        isUnread: true,                                                                                      // 1420
      };                                                                                                     // 1421
                                                                                                             // 1422
      // Set only the appId that we care about. Use mongo's dot notation to specify only a single            // 1423
      // field inside of an object to update                                                                 // 1424
      updater["appUpdates." + appId] = {                                                                     // 1425
        marketingVersion: marketingVersion,                                                                  // 1426
        packageId: packageId,                                                                                // 1427
        name: name,                                                                                          // 1428
        version: versionNumber,                                                                              // 1429
      };                                                                                                     // 1430
      db.collections.notifications.upsert({userId: userId}, { $set: updater});                               // 1431
    });                                                                                                      // 1432
                                                                                                             // 1433
    db.collections.appIndex.update({_id: appId}, {$set: {hasSentNotifications: true}});                      // 1434
                                                                                                             // 1435
    // In the case where we replaced a previous notification and that was the only reference to the          // 1436
    // package, we need to clean it up                                                                       // 1437
    Meteor.call("deleteUnusedPackages", appId);                                                              // 1438
  };                                                                                                         // 1439
                                                                                                             // 1440
  SandstormDb.prototype.sendReferralProgramNotification = sendReferralProgramNotification;                   // 1441
                                                                                                             // 1442
  SandstormDb.prototype.upgradeGrains =  function (appId, version, packageId, backend) {                     // 1443
    check(appId, String);                                                                                    // 1444
    check(version, Match.Integer);                                                                           // 1445
    check(packageId, String);                                                                                // 1446
                                                                                                             // 1447
    var selector = {                                                                                         // 1448
      userId: Meteor.userId(),                                                                               // 1449
      appId: appId,                                                                                          // 1450
      appVersion: { $lte: version },                                                                         // 1451
      packageId: { $ne: packageId }                                                                          // 1452
    };                                                                                                       // 1453
                                                                                                             // 1454
    if (!this.isSimulation) {                                                                                // 1455
      Grains.find(selector).forEach(function (grain) {                                                       // 1456
        backend.shutdownGrain(grain._id, grain.userId);                                                      // 1457
      });                                                                                                    // 1458
    }                                                                                                        // 1459
                                                                                                             // 1460
    Grains.update(selector, { $set: { appVersion: version, packageId: packageId }}, {multi: true});          // 1461
  };                                                                                                         // 1462
                                                                                                             // 1463
  SandstormDb.prototype.startInstall = function (packageId, url, retryFailed, isAutoUpdated) {               // 1464
    // Mark package for possible installation.                                                               // 1465
                                                                                                             // 1466
    var fields = {status: "download", progress: 0, url: url, isAutoUpdated: !!isAutoUpdated};                // 1467
                                                                                                             // 1468
    if (retryFailed) {                                                                                       // 1469
      Packages.update({_id: packageId, status: "failed"}, {$set: fields});                                   // 1470
    } else {                                                                                                 // 1471
      try {                                                                                                  // 1472
        fields._id = packageId;                                                                              // 1473
        Packages.insert(fields);                                                                             // 1474
      } catch (err) {                                                                                        // 1475
        console.error("Simultaneous startInstall()s?", err.stack);                                           // 1476
      }                                                                                                      // 1477
    }                                                                                                        // 1478
  };                                                                                                         // 1479
                                                                                                             // 1480
  var ValidKeyFingerprint = Match.Where(function (keyFingerprint) {                                          // 1481
    check(keyFingerprint, String);                                                                           // 1482
    return !!keyFingerprint.match(/^[0-9A-F]{40}$/);                                                         // 1483
  });                                                                                                        // 1484
                                                                                                             // 1485
  SandstormDb.prototype.updateKeybaseProfileAsync = function (keyFingerprint) {                              // 1486
    // Asynchronously fetch the given Keybase profile and populate the KeybaseProfiles collection.           // 1487
                                                                                                             // 1488
    check(keyFingerprint, ValidKeyFingerprint);                                                              // 1489
                                                                                                             // 1490
    console.log("fetching keybase", keyFingerprint);                                                         // 1491
                                                                                                             // 1492
    HTTP.get(                                                                                                // 1493
        "https://keybase.io/_/api/1.0/user/lookup.json?key_fingerprint=" + keyFingerprint +                  // 1494
        "&fields=basics,profile,proofs_summary", {                                                           // 1495
      timeout: 5000                                                                                          // 1496
    }, function (err, keybaseResponse) {                                                                     // 1497
      if (err) {                                                                                             // 1498
        console.log("keybase lookup error:", err.stack);                                                     // 1499
        return;                                                                                              // 1500
      }                                                                                                      // 1501
                                                                                                             // 1502
      if (!keybaseResponse.data) {                                                                           // 1503
        console.log("keybase didn't return JSON? Headers:", keybaseResponse.headers);                        // 1504
        return;                                                                                              // 1505
      }                                                                                                      // 1506
                                                                                                             // 1507
      var profile = (keybaseResponse.data.them || [])[0];                                                    // 1508
                                                                                                             // 1509
      if (profile) {                                                                                         // 1510
        var record = {                                                                                       // 1511
          displayName: (profile.profile || {}).full_name,                                                    // 1512
          handle: (profile.basics || {}).username,                                                           // 1513
          proofs: (profile.proofs_summary || {}).all || []                                                   // 1514
        };                                                                                                   // 1515
                                                                                                             // 1516
        record.proofs.forEach(function (proof) {                                                             // 1517
          // Remove potentially Mongo-incompatible stuff. (Currently Keybase returns nothing that            // 1518
          // this would filter.)                                                                             // 1519
          for (var field in proof) {                                                                         // 1520
            // Don't allow field names containing '.' or '$'. Also don't allow sub-objects mainly            // 1521
            // because I'm too lazy to check the field names recursively (and Keybase doesn't                // 1522
            // return any objects anyway).                                                                   // 1523
            if (field.match(/[.$]/) || typeof(proof[field]) === "object") {                                  // 1524
              delete proof[field];                                                                           // 1525
            }                                                                                                // 1526
          }                                                                                                  // 1527
                                                                                                             // 1528
          // Indicate not verified.                                                                          // 1529
          // TODO(security): Asynchronously verify proofs. Presumably we can borrow code from the            // 1530
          //   Keybase node-based CLI.                                                                       // 1531
          proof.status = "unverified";                                                                       // 1532
        });                                                                                                  // 1533
                                                                                                             // 1534
        KeybaseProfiles.update(keyFingerprint, {$set: record}, {upsert: true});                              // 1535
      } else {                                                                                               // 1536
        // Keybase reports no match, so remove what we know of this user. We don't want to remove            // 1537
        // the item entirely from the cache as this will cause us to repeatedly re-fetch the data            // 1538
        // from Keybase.                                                                                     // 1539
        //                                                                                                   // 1540
        // TODO(someday): We could perhaps keep the proofs if we can still verify them directly,             // 1541
        //   but at present we don't have the ability to verify proofs.                                      // 1542
        KeybaseProfiles.update(keyFingerprint,                                                               // 1543
            {$unset: {displayName: "", handle: "", proofs: ""}}, {upsert: true});                            // 1544
      }                                                                                                      // 1545
    });                                                                                                      // 1546
  };                                                                                                         // 1547
                                                                                                             // 1548
  SandstormDb.prototype.deleteUnusedAccount = function(backend, identityId) {                                // 1549
    // If there is an *unused* account that has `identityId` as a login identity, deletes it.                // 1550
                                                                                                             // 1551
    check(identityId, String);                                                                               // 1552
    var account = Meteor.users.findOne({"loginIdentities.id": identityId});                                  // 1553
    if (account &&                                                                                           // 1554
        account.loginIdentities.length == 1 &&                                                               // 1555
        account.nonloginIdentities.length == 0 &&                                                            // 1556
        !Grains.findOne({userId: account._id}) &&                                                            // 1557
        !ApiTokens.findOne({accountId: account._id}) &&                                                      // 1558
        (!account.plan || account.plan === "free") &&                                                        // 1559
        !account.payments &&                                                                                 // 1560
        !Contacts.findOne({ownerId: account._id})) {                                                         // 1561
      Meteor.users.remove({_id: account._id});                                                               // 1562
      backend.deleteUser(account._id);                                                                       // 1563
    }                                                                                                        // 1564
  }                                                                                                          // 1565
                                                                                                             // 1566
  Meteor.publish("keybaseProfile", function (keyFingerprint) {                                               // 1567
    check(keyFingerprint, ValidKeyFingerprint);                                                              // 1568
    var db = this.connection.sandstormDb;                                                                    // 1569
                                                                                                             // 1570
    var cursor = db.collections.keybaseProfiles.find(keyFingerprint);                                        // 1571
    if (cursor.count() === 0) {                                                                              // 1572
      // Fire off async update.                                                                              // 1573
      db.updateKeybaseProfileAsync(keyFingerprint);                                                          // 1574
    }                                                                                                        // 1575
                                                                                                             // 1576
    return cursor;                                                                                           // 1577
  });                                                                                                        // 1578
                                                                                                             // 1579
  Meteor.publish("appIndex", function (appId) {                                                              // 1580
    check(appId, String);                                                                                    // 1581
    var db = this.connection.sandstormDb;                                                                    // 1582
    var cursor = db.collections.appIndex.find({_id: appId});                                                 // 1583
    return cursor;                                                                                           // 1584
  });                                                                                                        // 1585
                                                                                                             // 1586
  Meteor.publish("userPackages", function() {                                                                // 1587
    // Users should be able to see packages that are either:                                                 // 1588
    // 1. referenced by one of their userActions                                                             // 1589
    // 2. referenced by one of their grains                                                                  // 1590
    var self = this;                                                                                         // 1591
    var db = this.connection.sandstormDb;                                                                    // 1592
                                                                                                             // 1593
    // Note that package information, once it is in the database, is static. There's no need to              // 1594
    // reactively subscribe to changes to a package since they don't change. It's also unecessary            // 1595
    // to reactively remove a package from the client side when it is removed on the server, or              // 1596
    // when the client stops using it, because the worst case is the client has a small amount               // 1597
    // of extra info on a no-longer-used package held in memory until they refresh Sandstorm.                // 1598
    // So, we implement this as a cache: the first time each package ID shows up among the user's            // 1599
    // stuff, we push the package info to the client, and then we never update it.                           // 1600
    //                                                                                                       // 1601
    // Alternatively, we could subscribe to each individual package query, but this would waste              // 1602
    // lots of server-side resources watching for events that will never happen or don't matter.             // 1603
    var hasPackage = {};                                                                                     // 1604
    var refPackage = function (packageId) {                                                                  // 1605
      // Ignore dev apps.                                                                                    // 1606
      if (packageId.lastIndexOf("dev-", 0) === 0) return;                                                    // 1607
                                                                                                             // 1608
      if (!hasPackage[packageId]) {                                                                          // 1609
        hasPackage[packageId] = true;                                                                        // 1610
        var pkg = db.getPackage(packageId);                                                                  // 1611
        if (pkg) {                                                                                           // 1612
          self.added("packages", packageId, pkg);                                                            // 1613
        } else {                                                                                             // 1614
          console.error(                                                                                     // 1615
              "shouldn't happen: missing package referenced by user's stuff:", packageId);                   // 1616
        }                                                                                                    // 1617
      }                                                                                                      // 1618
    };                                                                                                       // 1619
                                                                                                             // 1620
    // package source 1: packages referred to by actions                                                     // 1621
    var actions = db.userActions(this.userId);                                                               // 1622
    var actionsHandle = actions.observe({                                                                    // 1623
      added: function(newAction) {                                                                           // 1624
        refPackage(newAction.packageId);                                                                     // 1625
      },                                                                                                     // 1626
      changed: function(oldAction, newAction) {                                                              // 1627
        refPackage(newAction.packageId);                                                                     // 1628
      }                                                                                                      // 1629
    });                                                                                                      // 1630
                                                                                                             // 1631
    // package source 2: packages referred to by grains directly                                             // 1632
    var grains = db.userGrains(this.userId);                                                                 // 1633
    var grainsHandle = grains.observe({                                                                      // 1634
      added: function(newGrain) {                                                                            // 1635
        refPackage(newGrain.packageId);                                                                      // 1636
      },                                                                                                     // 1637
      changed: function(oldGrain, newGrain) {                                                                // 1638
        refPackage(newGrain.packageId);                                                                      // 1639
      }                                                                                                      // 1640
    });                                                                                                      // 1641
                                                                                                             // 1642
    this.onStop(function () {                                                                                // 1643
      actionsHandle.stop();                                                                                  // 1644
      grainsHandle.stop();                                                                                   // 1645
    });                                                                                                      // 1646
    this.ready();                                                                                            // 1647
  });                                                                                                        // 1648
}                                                                                                            // 1649
                                                                                                             // 1650
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/profile.js                                                                          //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Sandstorm - Personal Cloud Sandbox                                                                        // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                     // 2
// All rights reserved.                                                                                      // 3
//                                                                                                           // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                           // 5
// you may not use this file except in compliance with the License.                                          // 6
// You may obtain a copy of the License at                                                                   // 7
//                                                                                                           // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                              // 9
//                                                                                                           // 10
// Unless required by applicable law or agreed to in writing, software                                       // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                         // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                  // 13
// See the License for the specific language governing permissions and                                       // 14
// limitations under the License.                                                                            // 15
                                                                                                             // 16
var makeIdenticon;                                                                                           // 17
var httpProtocol;                                                                                            // 18
                                                                                                             // 19
if (Meteor.isServer) {                                                                                       // 20
  SandstormDb.ensureSubscriberHasIdentity = function(publishHandler, identityId) {                           // 21
    // Helper for publish functions that need to restrict access based on whether the subscriber             // 22
    // has a given identity linked. Automatically stops the subscription if the user loses the               // 23
    // identity. Returns a boolean indicating whether the user initially has the identity.                   // 24
                                                                                                             // 25
    var userId = publishHandler.userId;                                                                      // 26
    if (userId === identityId) {                                                                             // 27
      return true;                                                                                           // 28
    } else {                                                                                                 // 29
      var hasIdentityCursor =                                                                                // 30
          Meteor.users.find({$or: [{_id: userId, "loginIdentities.id": identityId},                          // 31
                                   {_id: userId, "nonloginIdentities.id": identityId}]});                    // 32
      if (hasIdentityCursor.count() == 0) {                                                                  // 33
        publishHandler.stop();                                                                               // 34
        return false;                                                                                        // 35
      }                                                                                                      // 36
      var handle = hasIdentityCursor.observe({removed: function () { publishHandler.stop(); }});             // 37
      publishHandler.onStop(function () { handle.stop(); });                                                 // 38
      return true;                                                                                           // 39
    }                                                                                                        // 40
  }                                                                                                          // 41
                                                                                                             // 42
  Meteor.publish("identityProfile", function (identityId) {                                                  // 43
    check(identityId, String);                                                                               // 44
    if (!SandstormDb.ensureSubscriberHasIdentity(this, identityId)) return;                                  // 45
                                                                                                             // 46
    return Meteor.users.find({_id: identityId},                                                              // 47
      {fields: {                                                                                             // 48
        "profile":1,                                                                                         // 49
        "unverifiedEmail":1,                                                                                 // 50
        "expires": 1,                                                                                        // 51
        "createdAt": 1,                                                                                      // 52
                                                                                                             // 53
        "services.dev.name":1,                                                                               // 54
                                                                                                             // 55
        "services.google.id":1,                                                                              // 56
        "services.google.email":1,                                                                           // 57
        "services.google.verified_email":1,                                                                  // 58
        "services.google.name":1,                                                                            // 59
        "services.google.picture":1,                                                                         // 60
        "services.google.gender":1,                                                                          // 61
                                                                                                             // 62
        "services.github.id":1,                                                                              // 63
        "services.github.email":1,                                                                           // 64
        "services.github.emails":1,                                                                          // 65
        "services.github.username":1,                                                                        // 66
                                                                                                             // 67
        "services.email.email":1,                                                                            // 68
      }});                                                                                                   // 69
  }),                                                                                                        // 70
                                                                                                             // 71
  Meteor.publish("accountIdentities", function () {                                                          // 72
    // Maybe this should be folded into the "credentials" subscription?                                      // 73
                                                                                                             // 74
    if (!this.userId) return [];                                                                             // 75
                                                                                                             // 76
    return Meteor.users.find(                                                                                // 77
      {_id: this.userId},                                                                                    // 78
      {fields: {                                                                                             // 79
        "profile": 1,                                                                                        // 80
        "verifiedEmail": 1,                                                                                  // 81
        "loginIdentities": 1,                                                                                // 82
        "nonloginIdentities": 1,                                                                             // 83
        "expires": 1,                                                                                        // 84
        "primaryEmail": 1,                                                                                   // 85
      }});                                                                                                   // 86
  });                                                                                                        // 87
                                                                                                             // 88
  makeIdenticon = function (id) {                                                                            // 89
    // We only make identicons client-side.                                                                  // 90
    return undefined;                                                                                        // 91
  }                                                                                                          // 92
                                                                                                             // 93
  var Url = Npm.require("url");                                                                              // 94
  httpProtocol = Url.parse(process.env.ROOT_URL).protocol;                                                   // 95
} else {                                                                                                     // 96
  var identiconCache = {};                                                                                   // 97
                                                                                                             // 98
  makeIdenticon = function (id) {                                                                            // 99
    // Given a cryptographic hash as input, generate an identicon. We always use the user's                  // 100
    // identity ID as the input hash (the same thing passed to apps as the user ID).                         // 101
                                                                                                             // 102
    // Pass to identicon.js exactly the hash that we'd pass to apps as X-Sandstorm-User-Id, so               // 103
    // that apps can themselves use identicon.js to produce consistent identicons. As it turns out           // 104
    // identicon.js doesn't use the second half of the hash even if we provide it, but slice it              // 105
    // anyway to be safe.                                                                                    // 106
    id = id.slice(0, 32);                                                                                    // 107
                                                                                                             // 108
    if (id in identiconCache) {                                                                              // 109
      return identiconCache[id];                                                                             // 110
    }                                                                                                        // 111
                                                                                                             // 112
    // Unfortunately, Github's algorithm uses MD5. Whatever, we don't expect these to be secure.             // 113
    var data = new Identicon(id, 512).toString();                                                            // 114
    var result = "data:image/png;base64," + data;                                                            // 115
    identiconCache[id] = result;                                                                             // 116
    return result;                                                                                           // 117
  }                                                                                                          // 118
                                                                                                             // 119
  httpProtocol = window.location.protocol;                                                                   // 120
}                                                                                                            // 121
                                                                                                             // 122
var GENDERS = {male: "male", female: "female", neutral: "neutral", robot: "robot"};                          // 123
                                                                                                             // 124
function staticAssetUrl(id, staticHost) {                                                                    // 125
  if (id) {                                                                                                  // 126
    return staticHost + "/" + id;                                                                            // 127
  } else {                                                                                                   // 128
    return undefined;                                                                                        // 129
  }                                                                                                          // 130
}                                                                                                            // 131
                                                                                                             // 132
function filterHandle(handle) {                                                                              // 133
  // Convert disallowed letters into underscores, but no more than one underscore in a row.                  // 134
  return handle && handle.toLowerCase().split(/[^a-z0-9_]+/g).join("_");                                     // 135
}                                                                                                            // 136
                                                                                                             // 137
function emailToHandle(email) {                                                                              // 138
  // Turn an email address into a handle.                                                                    // 139
                                                                                                             // 140
  if (!email) return undefined;                                                                              // 141
                                                                                                             // 142
  // Use the stuff before the @. Ignore stuff after '+' because it's commonly used for filters               // 143
  // on the same account.                                                                                    // 144
  var parts = email.split("@");                                                                              // 145
  var base = filterHandle(parts[0].split("+")[0]);                                                           // 146
                                                                                                             // 147
  var domain = (parts[1]||"").split(".");                                                                    // 148
  if (domain[domain.length - 1] === "name") {                                                                // 149
    // Oh, a .name domain. Let's use                                                                         // 150
    base = filterHandle(domain.slice(0, domain.length - 1).join("."));                                       // 151
  } else if (_.contains(["me", "self", "contact", "admin", "administrator", "root", "info",                  // 152
                         "sandstorm", "sandstormio", "inbox", "indiegogo", "mail", "email"],                 // 153
                         base)) {                                                                            // 154
    // This is probably an address at a vanity domain. Use the domain itself as the handle.                  // 155
    base = filterHandle(domain[0]);                                                                          // 156
  }                                                                                                          // 157
                                                                                                             // 158
  return filterHandle(base);                                                                                 // 159
}                                                                                                            // 160
                                                                                                             // 161
SandstormDb.fillInProfileDefaults = function(user) {                                                         // 162
  var profile = user.profile;                                                                                // 163
  if (profile.service === "github") {                                                                        // 164
    profile.name = profile.name || user.services.github.username || "Name Unknown";                          // 165
    profile.handle = profile.handle || filterHandle(user.services.github.username) ||                        // 166
        filterHandle(profile.name);                                                                          // 167
  } else if (profile.service === "google") {                                                                 // 168
    profile.name = profile.name || user.services.google.name || "Name Unknown";                              // 169
    profile.handle = profile.handle || emailToHandle(user.services.google.email) ||                          // 170
        filterHandle(profile.name);                                                                          // 171
    profile.pronoun = profile.pronoun || GENDERS[user.services.google.gender] || "neutral";                  // 172
  } else if (profile.service === "email") {                                                                  // 173
    var email = user.services.email.email                                                                    // 174
    profile.name = profile.name || email.split("@")[0];                                                      // 175
    profile.handle = profile.handle || emailToHandle(email);                                                 // 176
  } else if (profile.service === "dev") {                                                                    // 177
    var lowerCaseName = user.services.dev.name.split(" ")[0].toLowerCase();                                  // 178
    profile.name = profile.name || user.services.dev.name;                                                   // 179
    profile.handle = profile.handle || filterHandle(lowerCaseName);                                          // 180
    profile.pronoun = profile.pronoun ||                                                                     // 181
        (_.contains(["alice", "carol", "eve"], lowerCaseName) ? "female" :                                   // 182
         _.contains(["bob", "dave"], lowerCaseName) ? "male" : "neutral");                                   // 183
  } else if (profile.service === "demo") {                                                                   // 184
    profile.name = profile.name || "Demo User";                                                              // 185
    profile.handle = profile.handle || "demo";                                                               // 186
  } else {                                                                                                   // 187
    throw new Error("unrecognized identity service: ", profile.service);                                     // 188
  }                                                                                                          // 189
                                                                                                             // 190
  profile.pronoun = profile.pronoun || "neutral";                                                            // 191
}                                                                                                            // 192
                                                                                                             // 193
SandstormDb.fillInIntrinsicName = function(user) {                                                           // 194
  var profile = user.profile;                                                                                // 195
  if (profile.service === "github") {                                                                        // 196
    profile.intrinsicName = user.services.github.username;                                                   // 197
  } else if (profile.service === "google") {                                                                 // 198
    profile.intrinsicName = user.services.google.name;                                                       // 199
    user.privateIntrinsicName = user.services.google.email;                                                  // 200
  } else if (profile.service === "email") {                                                                  // 201
    profile.intrinsicName = user.services.email.email                                                        // 202
  } else if (profile.service === "dev") {                                                                    // 203
    profile.intrinsicName = user.services.dev.name;                                                          // 204
  } else if (profile.service === "demo") {                                                                   // 205
    profile.intrinsicName = "demo on " + user.createdAt.toISOString().substring(0,10);                       // 206
  } else {                                                                                                   // 207
    throw new Error("unrecognized identity service: ", profile.service);                                     // 208
  }                                                                                                          // 209
}                                                                                                            // 210
                                                                                                             // 211
SandstormDb.getVerifiedEmails = function(identity) {                                                         // 212
  if (identity.services.google && identity.services.google.email &&                                          // 213
      identity.services.google.verified_email) {                                                             // 214
    return [identity.services.google.email];                                                                 // 215
  } else if (identity.services.email) {                                                                      // 216
    return [identity.services.email.email];                                                                  // 217
  } else if (identity.services.github && identity.services.github.emails) {                                  // 218
    return _.pluck(_.filter(identity.services.github.emails,                                                 // 219
                            function(email) { return email.verified; }),                                     // 220
                   "email");                                                                                 // 221
  }                                                                                                          // 222
  return [];                                                                                                 // 223
}                                                                                                            // 224
                                                                                                             // 225
SandstormDb.prototype.findIdentitiesByEmail = function(email) {                                              // 226
  // Returns an array of identities which have the given email address as one of their verified              // 227
  // addresses.                                                                                              // 228
                                                                                                             // 229
  check(email, String);                                                                                      // 230
                                                                                                             // 231
  return Meteor.users.find({$or: [                                                                           // 232
    {"services.google.email": email},                                                                        // 233
    {"services.email.email": email},                                                                         // 234
    {"services.github.emails.email": email},                                                                 // 235
  ]}).fetch().filter(function (identity) {                                                                   // 236
    // Verify that the email is verified, since our query doesn't technically do that.                       // 237
    return SandstormDb.getVerifiedEmails(identity).indexOf(email) >= 0;                                      // 238
  });                                                                                                        // 239
}                                                                                                            // 240
                                                                                                             // 241
SandstormDb.prototype.findAccountsByEmail = function(email) {                                                // 242
  var identityIds = _.pluck(this.findIdentitiesByEmail(email), "_id");                                       // 243
  return Meteor.users.find({$or: [                                                                           // 244
    {"loginIdentities.id": {$in: identityIds}},                                                              // 245
    {"nonloginIdentities.id": {$in: identityIds}},                                                           // 246
  ]}).fetch();                                                                                               // 247
};                                                                                                           // 248
                                                                                                             // 249
SandstormDb.fillInPictureUrl = function(user) {                                                              // 250
  var staticHost = httpProtocol + "//" + makeWildcardHost("static");                                         // 251
  user.profile.pictureUrl = staticAssetUrl(user.profile.picture, staticHost) ||                              // 252
    makeIdenticon(user._id);                                                                                 // 253
}                                                                                                            // 254
                                                                                                             // 255
SandstormDb.getUserIdentityIds = function (user) {                                                           // 256
  // Given an account user object, returns an array containing the ID of each identity linked to the         // 257
  // account. Always returns the most recently added login identity first.                                   // 258
  if (user && user.loginIdentities) {                                                                        // 259
    return _.pluck(user.nonloginIdentities.concat(user.loginIdentities), "id").reverse();                    // 260
  } else {                                                                                                   // 261
    return [];                                                                                               // 262
  }                                                                                                          // 263
};                                                                                                           // 264
                                                                                                             // 265
SandstormDb.getUserEmails = function (user) {                                                                // 266
  // Given a user object, returns an array containing all email addresses associated with that user.         // 267
  // Each entry in the array is an object of the form:                                                       // 268
  //     `{email: String, verified: Bool, primary: Optional(Bool)}`                                          // 269
  //                                                                                                         // 270
  // At most one entry in the result has `primary = true`.                                                   // 271
  //                                                                                                         // 272
  // TODO(cleanup): This actually does need to query the database to fetch profile information               // 273
  //   for linked identities, so it probably makes more sense for it to be a non-static method               // 274
  //   on SandstormDb.                                                                                       // 275
                                                                                                             // 276
  var identityIds = SandstormDb.getUserIdentityIds(user);                                                    // 277
  verifiedEmails = {};                                                                                       // 278
  unverifiedEmails = {};                                                                                     // 279
                                                                                                             // 280
  identityIds.forEach(function (id) {                                                                        // 281
    var identity = Meteor.users.findOne({_id: id});                                                          // 282
    if (identity && identity.services) {                                                                     // 283
      SandstormDb.getVerifiedEmails(identity).forEach(function(verifiedEmail) {                              // 284
        if (verifiedEmail) {                                                                                 // 285
          verifiedEmails[verifiedEmail] = true;                                                              // 286
        }                                                                                                    // 287
      });                                                                                                    // 288
    }                                                                                                        // 289
    if (identity && identity.unverifiedEmail) {                                                              // 290
      unverifiedEmails[identity.unverifiedEmail] = true;                                                     // 291
    }                                                                                                        // 292
  });                                                                                                        // 293
                                                                                                             // 294
  var result = [];                                                                                           // 295
  _.keys(verifiedEmails).map(function (email) {                                                              // 296
    result.push({email: email,                                                                               // 297
                 verified: true,                                                                             // 298
                 primary: email === user.primaryEmail});                                                     // 299
  });                                                                                                        // 300
  _.keys(unverifiedEmails).map(function (email) {                                                            // 301
    if (!(email in verifiedEmails)) { result.push({email: email, verified: false}) }                         // 302
  });                                                                                                        // 303
                                                                                                             // 304
  // If `user.primaryEmail` is not among the verified emails, mark the first as primary.                     // 305
  if (!(user.primaryEmail in verifiedEmails) && result.length > 0) {                                         // 306
    result[0].primary = true;                                                                                // 307
  }                                                                                                          // 308
  return result;                                                                                             // 309
}                                                                                                            // 310
                                                                                                             // 311
SandstormDb.prototype.addContact = function addContact (ownerId, identityId) {                               // 312
  check(ownerId, String);                                                                                    // 313
  check(identityId, String);                                                                                 // 314
  var db = this;                                                                                             // 315
  var profile = db.getIdentity(identityId).profile;                                                          // 316
  db.collections.contacts.upsert({ownerId: ownerId, identityId: identityId}, {                               // 317
    ownerId: ownerId,                                                                                        // 318
    petname: profile && profile.name,                                                                        // 319
    created: new Date(),                                                                                     // 320
    identityId: identityId,                                                                                  // 321
  });                                                                                                        // 322
}                                                                                                            // 323
                                                                                                             // 324
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/user.js                                                                             //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Sandstorm - Personal Cloud Sandbox                                                                        // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                     // 2
// All rights reserved.                                                                                      // 3
//                                                                                                           // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                           // 5
// you may not use this file except in compliance with the License.                                          // 6
// You may obtain a copy of the License at                                                                   // 7
//                                                                                                           // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                              // 9
//                                                                                                           // 10
// Unless required by applicable law or agreed to in writing, software                                       // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                         // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                  // 13
// See the License for the specific language governing permissions and                                       // 14
// limitations under the License.                                                                            // 15
                                                                                                             // 16
var Crypto = Npm.require("crypto");                                                                          // 17
var Future = Npm.require("fibers/future");                                                                   // 18
                                                                                                             // 19
userPictureUrl = function (user) {                                                                           // 20
  if (user.services && !(user.profile && user.profile.picture)) {                                            // 21
    // Try to determine user's avatar URL from login service.                                                // 22
                                                                                                             // 23
    var google = user.services.google;                                                                       // 24
    if (google && google.picture) {                                                                          // 25
      return google.picture;                                                                                 // 26
    }                                                                                                        // 27
                                                                                                             // 28
    var github = user.services.github;                                                                       // 29
    if (github && github.id) {                                                                               // 30
      return "https://avatars.githubusercontent.com/u/" + github.id;                                         // 31
    }                                                                                                        // 32
                                                                                                             // 33
    // Note that we do NOT support Gravatar for email addresses because pinging Gravatar would be            // 34
    // a data leak, revealing that the user has logged into this Sandstorm server. Google and                // 35
    // Github are different because they are actually the identity providers, so they already know           // 36
    // the user logged in.                                                                                   // 37
  }                                                                                                          // 38
}                                                                                                            // 39
                                                                                                             // 40
fetchPicture = function (url) {                                                                              // 41
  try {                                                                                                      // 42
    var result = HTTP.get(url, {                                                                             // 43
      npmRequestOptions: { encoding: null },                                                                 // 44
      timeout: 5000                                                                                          // 45
    });                                                                                                      // 46
                                                                                                             // 47
    var metadata = {};                                                                                       // 48
                                                                                                             // 49
    metadata.mimeType = result.headers["content-type"];                                                      // 50
    if (metadata.mimeType.lastIndexOf("image/png", 0) === -1 &&                                              // 51
        metadata.mimeType.lastIndexOf("image/jpeg", 0) === -1) {                                             // 52
      throw new Error("unexpected Content-Type:", metadata.mimeType);                                        // 53
    }                                                                                                        // 54
    var enc = result.headers["content-encoding"];                                                            // 55
    if (enc && enc !== "identity") {                                                                         // 56
      metadata.encoding = enc;                                                                               // 57
    }                                                                                                        // 58
                                                                                                             // 59
    return addStaticAsset(metadata, result.content);                                                         // 60
  } catch (err) {                                                                                            // 61
    console.error("failed to fetch user profile picture:", url, err.stack);                                  // 62
  }                                                                                                          // 63
}                                                                                                            // 64
                                                                                                             // 65
var ValidHandle = Match.Where(function (handle) {                                                            // 66
  check(handle, String);                                                                                     // 67
  return !!handle.match(/^[a-z_][a-z0-9_]*$/);                                                               // 68
});                                                                                                          // 69
                                                                                                             // 70
Accounts.onCreateUser(function (options, user) {                                                             // 71
  if (user.loginIdentities) {                                                                                // 72
    // it's an account                                                                                       // 73
    check(user, {_id: String,                                                                                // 74
                 createdAt: Date,                                                                            // 75
                 isAdmin: Match.Optional(Boolean),                                                           // 76
                 hasCompletedSignup: Match.Optional(Boolean),                                                // 77
                 signupKey: Match.Optional(String),                                                          // 78
                 signupNote: Match.Optional(String),                                                         // 79
                 signupEmail: Match.Optional(String),                                                        // 80
                 expires: Match.Optional(Date),                                                              // 81
                 appDemoId: Match.Optional(String),                                                          // 82
                 loginIdentities: [{id: String}],                                                            // 83
                 nonloginIdentities: [{id: String}]});                                                       // 84
                                                                                                             // 85
    if (Meteor.settings.public.quotaEnabled) {                                                               // 86
      user.experiments = user.experiments || {};                                                             // 87
      user.experiments = {                                                                                   // 88
        firstTimeBillingPrompt: Math.random() < 0.5 ? "control" : "test"                                     // 89
      };                                                                                                     // 90
      if (!("expires" in user)) {                                                                            // 91
        sendReferralProgramNotification(user._id);                                                           // 92
      }                                                                                                      // 93
    }                                                                                                        // 94
                                                                                                             // 95
    return user;                                                                                             // 96
  }                                                                                                          // 97
                                                                                                             // 98
  // Check profile.                                                                                          // 99
  if (options.profile) {                                                                                     // 100
    // TODO(cleanup): This check also appears in accounts-ui-methods.js.                                     // 101
    check(options.profile, Match.ObjectIncluding({                                                           // 102
      name: Match.OneOf(null, Match.Optional(String)),                                                       // 103
      handle: Match.Optional(ValidHandle),                                                                   // 104
      pronoun: Match.Optional(Match.OneOf("male", "female", "neutral", "robot")),                            // 105
    }));                                                                                                     // 106
  }                                                                                                          // 107
                                                                                                             // 108
  check(options.unverifiedEmail, Match.Optional(String));                                                    // 109
                                                                                                             // 110
  if (options.unverifiedEmail) {                                                                             // 111
    user.unverifiedEmail = options.unverifiedEmail;                                                          // 112
  }                                                                                                          // 113
  user.profile = _.pick(options.profile || {}, "name", "handle", "pronouns");                                // 114
                                                                                                             // 115
  // Try downloading avatar.                                                                                 // 116
  var url = userPictureUrl(user);                                                                            // 117
  if (url) {                                                                                                 // 118
    var assetId = fetchPicture(url);                                                                         // 119
    if (assetId) {                                                                                           // 120
      user.profile.picture = assetId;                                                                        // 121
    }                                                                                                        // 122
  }                                                                                                          // 123
                                                                                                             // 124
  var serviceUserId;                                                                                         // 125
  if (user.services && user.services.dev) {                                                                  // 126
    check(user.services.dev, {name: String, isAdmin: Boolean, hasCompletedSignup: Boolean});                 // 127
    serviceUserId = user.services.dev.name;                                                                  // 128
    user.profile.service = "dev";                                                                            // 129
  } else if ("expires" in user) {                                                                            // 130
    serviceUserId = user._id;                                                                                // 131
    user.profile.service = "demo";                                                                           // 132
  } else if (user.services && user.services.email) {                                                         // 133
    check(user.services.email,                                                                               // 134
          {email: String,                                                                                    // 135
           tokens: [{digest: String, algorithm: String, createdAt: Date}]});                                 // 136
    serviceUserId = user.services.email.email;                                                               // 137
    user.profile.service = "email";                                                                          // 138
  } else if (user.services && "google" in user.services) {                                                   // 139
    serviceUserId = user.services.google.id;                                                                 // 140
    user.profile.service = "google";                                                                         // 141
  } else if (user.services && "github" in user.services) {                                                   // 142
    serviceUserId = user.services.github.id;                                                                 // 143
    user.profile.service = "github";                                                                         // 144
  } else {                                                                                                   // 145
    throw new Meteor.Error(400, "user does not have a recognized identity provider: " +                      // 146
                           JSON.stringify(user));                                                            // 147
  }                                                                                                          // 148
  user._id = Crypto.createHash("sha256")                                                                     // 149
    .update(user.profile.service + ":" + serviceUserId).digest("hex");                                       // 150
                                                                                                             // 151
  return user;                                                                                               // 152
});                                                                                                          // 153
                                                                                                             // 154
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/migrations.js                                                                       //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Minimal tooling for doing run-at-least-once, ordered migrations.                                          // 1
//                                                                                                           // 2
// Because migrations can experience partial failure and likely have                                         // 3
// side-effects, we should be careful to make sure all migrations are                                        // 4
// idempotent and safe to accidentally run multiple times.                                                   // 5
                                                                                                             // 6
var Future = Npm.require("fibers/future");                                                                   // 7
var Url = Npm.require("url");                                                                                // 8
                                                                                                             // 9
var updateLoginStyleToRedirect = function() {                                                                // 10
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;                 // 11
  ["google", "github"].forEach(function(serviceName) {                                                       // 12
    var config = configurations.findOne({service: serviceName});                                             // 13
    if (config && config.loginStyle !== "redirect") {                                                        // 14
      configurations.update({service: serviceName}, {$set: {loginStyle: "redirect"}});                       // 15
    }                                                                                                        // 16
  });                                                                                                        // 17
};                                                                                                           // 18
                                                                                                             // 19
var enableLegacyOAuthProvidersIfNotInSettings = function() {                                                 // 20
  // In the before time, Google and Github login were enabled by default.                                    // 21
  //                                                                                                         // 22
  // This actually didn't make much sense, required the first user to configure                              // 23
  // OAuth, and had some trust-the-first-user properties that weren't totally                                // 24
  // secure.                                                                                                 // 25
  //                                                                                                         // 26
  // Now, we have admin-token, so we wish to disable all logins by default                                   // 27
  // (since they need to be configured anyway) but since users may have never                                // 28
  // explicitly told Sandstorm that Google or Github login should be enabled,                                // 29
  // we can't just use the value in the Settings collection, since it might                                  // 30
  // never have been set.                                                                                    // 31
  //                                                                                                         // 32
  // Thus, if the service is configured but absent from Settings, we should                                  // 33
  // explicitly enable it in Settings, and then the rest of the logic can just                               // 34
  // depend on what value is in Settings and default to false without breaking                               // 35
  // user installations.                                                                                     // 36
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;                 // 37
  ["google", "github"].forEach(function(serviceName) {                                                       // 38
    var config = configurations.findOne({service: serviceName});                                             // 39
    var serviceConfig = Settings.findOne({_id: serviceName});                                                // 40
    if (config && !serviceConfig) {                                                                          // 41
      // Only explicitly enable the login service if:                                                        // 42
      // 1) the service is already configured                                                                // 43
      // 2) there is no sandstorm configuration already present (the user was                                // 44
      //    using the previous default behavior).                                                            // 45
      Settings.insert({_id: serviceName, value: true});                                                      // 46
    }                                                                                                        // 47
  });                                                                                                        // 48
};                                                                                                           // 49
                                                                                                             // 50
var denormalizeInviteInfo = function() {                                                                     // 51
  // When a user is invited via a signup token, the `signupKey` field of their user table entry              // 52
  // has always been populated to indicate the key they used. This points into the SignupKeys table          // 53
  // which has more information about the key, namely a freeform note entered by the admin when              // 54
  // they created the key. In the case that the email invite form was used, the note has the form            // 55
  // "E-mail invite to <address>".                                                                           // 56
  //                                                                                                         // 57
  // Later, we decided it was useful to indicate in the users table visible to the admin                     // 58
  // information about the invite terms. Namely, for email invites we want to show the address               // 59
  // and for others we want to show the note. To make this efficient, fields `signupNote` and                // 60
  // `signupEmail` were added to the users table. We can backfill these values by denormalizing              // 61
  // from the SignupKeys table.                                                                              // 62
                                                                                                             // 63
  Meteor.users.find().forEach(function (user) {                                                              // 64
    if (user.signupKey && (typeof user.signupKey) === "string" && user.signupKey !== "admin") {              // 65
      var signupInfo = SignupKeys.findOne(user.signupKey);                                                   // 66
      if (signupInfo && signupInfo.note) {                                                                   // 67
        var newFields = { signupNote: signupInfo.note };                                                     // 68
                                                                                                             // 69
        var prefix = "E-mail invite to ";                                                                    // 70
        if (signupInfo.note.lastIndexOf(prefix) === 0) {                                                     // 71
          newFields.signupEmail = signupInfo.note.slice(prefix.length);                                      // 72
        }                                                                                                    // 73
                                                                                                             // 74
        Meteor.users.update(user._id, {$set: newFields});                                                    // 75
      }                                                                                                      // 76
    }                                                                                                        // 77
  });                                                                                                        // 78
}                                                                                                            // 79
                                                                                                             // 80
function mergeRoleAssignmentsIntoApiTokens() {                                                               // 81
  RoleAssignments.find().forEach(function(roleAssignment) {                                                  // 82
    ApiTokens.insert({                                                                                       // 83
      grainId: roleAssignment.grainId,                                                                       // 84
      userId: roleAssignment.sharer,                                                                         // 85
      roleAssignment: roleAssignment.roleAssignment,                                                         // 86
      petname: roleAssignment.petname,                                                                       // 87
      created: roleAssignment.created,                                                                       // 88
      owner: {user: {userId: roleAssignment.recipient,                                                       // 89
                     title: roleAssignment.title}},                                                          // 90
    });                                                                                                      // 91
  });                                                                                                        // 92
}                                                                                                            // 93
                                                                                                             // 94
function fixOasisStorageUsageStats() {}                                                                      // 95
// This migration only pertained to Oasis and it was successfully applied there. Since it referred           // 96
// to some global variables that we later wanted to remove and/or rename, we've since replaced it            // 97
// with a no-op.                                                                                             // 98
                                                                                                             // 99
function fetchProfilePictures() {                                                                            // 100
  Meteor.users.find({}).forEach(function (user) {                                                            // 101
    var url = userPictureUrl(user);                                                                          // 102
    if (url) {                                                                                               // 103
      console.log("Fetching user picture:", url);                                                            // 104
      var assetId = fetchPicture(url);                                                                       // 105
      if (assetId) {                                                                                         // 106
        Meteor.users.update(user._id, {$set: {"profile.picture": assetId}});                                 // 107
      }                                                                                                      // 108
    }                                                                                                        // 109
  });                                                                                                        // 110
}                                                                                                            // 111
                                                                                                             // 112
function assignPlans() {                                                                                     // 113
  if (Meteor.settings.public.quotaEnabled && SandstormDb.paymentsMigrationHook) {                            // 114
    SandstormDb.paymentsMigrationHook(SignupKeys, Plans.find().fetch());                                     // 115
  }                                                                                                          // 116
}                                                                                                            // 117
                                                                                                             // 118
function removeKeyrings() {                                                                                  // 119
  // These blobs full of public keys were not intended to find their way into mongo and while                // 120
  // harmless they slow things down because they're huge. Remove them.                                       // 121
  Packages.update({"manifest.metadata.pgpKeyring": {$exists: true}},                                         // 122
      {$unset: {"manifest.metadata.pgpKeyring": ""}},                                                        // 123
      {multi: true});                                                                                        // 124
}                                                                                                            // 125
                                                                                                             // 126
function useLocalizedTextInUserActions() {                                                                   // 127
  function toLocalizedText(newObj, oldObj, field) {                                                          // 128
    if (field in oldObj) {                                                                                   // 129
      if (typeof oldObj[field] === "string") {                                                               // 130
        newObj[field] = {defaultText: oldObj[field]};                                                        // 131
      } else {                                                                                               // 132
        newObj[field] = oldObj[field];                                                                       // 133
      }                                                                                                      // 134
    }                                                                                                        // 135
  }                                                                                                          // 136
  UserActions.find({}).forEach(function (userAction) {                                                       // 137
    var fields = {};                                                                                         // 138
    toLocalizedText(fields, userAction, "appTitle");                                                         // 139
    toLocalizedText(fields, userAction, "title");                                                            // 140
    toLocalizedText(fields, userAction, "nounPhrase");                                                       // 141
    UserActions.update(userAction._id, {$set: fields});                                                      // 142
  });                                                                                                        // 143
}                                                                                                            // 144
                                                                                                             // 145
function verifyAllPgpSignatures() {                                                                          // 146
  Packages.find({}).forEach(function (pkg) {                                                                 // 147
    try {                                                                                                    // 148
      console.log("checking PGP signature for package:", pkg._id);                                           // 149
      var info = waitPromise(globalBackend.cap().tryGetPackage(pkg._id));                                    // 150
      if (info.authorPgpKeyFingerprint) {                                                                    // 151
        console.log("  " + info.authorPgpKeyFingerprint);                                                    // 152
        Packages.update(pkg._id,                                                                             // 153
            {$set: {authorPgpKeyFingerprint: info.authorPgpKeyFingerprint}});                                // 154
      } else {                                                                                               // 155
        console.log("  no signature");                                                                       // 156
      }                                                                                                      // 157
    } catch (err) {                                                                                          // 158
      console.error(err.stack);                                                                              // 159
    }                                                                                                        // 160
  });                                                                                                        // 161
}                                                                                                            // 162
                                                                                                             // 163
function splitUserIdsIntoAccountIdsAndIdentityIds() {                                                        // 164
  var Crypto = Npm.require("crypto");                                                                        // 165
  Meteor.users.find().forEach(function (user) {                                                              // 166
    var identity = {};                                                                                       // 167
    var serviceUserId;                                                                                       // 168
    if ("devName" in user) {                                                                                 // 169
      identity.service = "dev";                                                                              // 170
      serviceUserId = user.devName;                                                                          // 171
    } else if ("expires" in user) {                                                                          // 172
      identity.service = "demo";                                                                             // 173
      serviceUserId = user._id;                                                                              // 174
    } else if (user.services && "google" in user.services) {                                                 // 175
      identity.service = "google";                                                                           // 176
      if (user.services.google.email && user.services.google.verified_email) {                               // 177
        identity.verifiedEmail = user.services.google.email;                                                 // 178
      }                                                                                                      // 179
      serviceUserId = user.services.google.id;                                                               // 180
    } else if (user.services && "github" in user.services) {                                                 // 181
      identity.service = "github";                                                                           // 182
      identity.unverifiedEmail = user.services.github.email;                                                 // 183
      serviceUserId = user.services.github.id;                                                               // 184
    } else if (user.services && "emailToken" in user.services) {                                             // 185
      identity.service = "emailToken";                                                                       // 186
      identity.verifiedEmail = user.services.emailToken.email;                                               // 187
      serviceUserId = user.services.emailToken.email;                                                        // 188
    }                                                                                                        // 189
                                                                                                             // 190
    identity.id = Crypto.createHash("sha256")                                                                // 191
        .update(identity.service + ":" + serviceUserId).digest("hex");                                       // 192
                                                                                                             // 193
    if (user.profile) {                                                                                      // 194
      if (user.profile.name) {                                                                               // 195
        identity.name = user.profile.name;                                                                   // 196
      }                                                                                                      // 197
      if (user.profile.handle) {                                                                             // 198
        identity.handle = user.profile.handle;                                                               // 199
      }                                                                                                      // 200
      if (user.profile.picture) {                                                                            // 201
        identity.picture = user.profile.picture;                                                             // 202
      }                                                                                                      // 203
      if (user.profile.pronoun) {                                                                            // 204
        identity.pronoun = user.profile.pronoun;                                                             // 205
      }                                                                                                      // 206
      if (user.profile.email) {                                                                              // 207
        identity.unverifiedEmail = user.profile.email;                                                       // 208
      }                                                                                                      // 209
    }                                                                                                        // 210
    identity.main = true;                                                                                    // 211
                                                                                                             // 212
    Meteor.users.update(user._id, {$set: {identities: [identity]}});                                         // 213
                                                                                                             // 214
    Grains.update({userId: user._id}, {$set: {identityId: identity.id}}, {multi: true});                     // 215
    Sessions.update({userId: user._id}, {$set: {identityId: identity.id}}, {multi: true});                   // 216
    ApiTokens.update({userId: user._id},                                                                     // 217
                     {$set: {identityId: identity.id}},                                                      // 218
                     {multi: true});                                                                         // 219
    ApiTokens.update({"owner.user.userId": user._id},                                                        // 220
                     {$set: {"owner.user.identityId": identity.id}},                                         // 221
                     {multi: true});                                                                         // 222
    ApiTokens.update({"owner.grain.introducerUser": user._id},                                               // 223
                     {$set: {"owner.grain.introducerIdentity": identity.id}},                                // 224
                     {multi: true});                                                                         // 225
                                                                                                             // 226
    while (ApiTokens.update({"requirements.permissionsHeld.userId": user._id},                               // 227
                            {$set: {"requirements.$.permissionsHeld.identityId": identity.id},               // 228
                             $unset: {"requirements.$.permissionsHeld.userId": 1}},                          // 229
                            {multi: true}) > 0);                                                             // 230
    // The `$` operatorer modifies the first element in the array that matches the query. Since              // 231
    // there may be many matches, we need to repeat until no documents are modified.                         // 232
                                                                                                             // 233
  });                                                                                                        // 234
                                                                                                             // 235
  ApiTokens.remove({userInfo: {$exists: true}});                                                             // 236
  // We've renamed `Grain.UserInfo.userId` to `Grain.userInfo.identityId`. The only place                    // 237
  // that this field could show up in the database was in this deprecated, no-longer-functional              // 238
  // form of API token.                                                                                      // 239
}                                                                                                            // 240
                                                                                                             // 241
function appUpdateSettings() {                                                                               // 242
  Settings.insert({_id: "appMarketUrl", value: "https://apps.sandstorm.io"});                                // 243
  Settings.insert({_id: "appIndexUrl", value: "https://app-index.sandstorm.io"});                            // 244
  Settings.insert({_id: "appUpdatesEnabled", value: true});                                                  // 245
}                                                                                                            // 246
                                                                                                             // 247
function moveDevAndEmailLoginDataIntoIdentities() {                                                          // 248
  Meteor.users.find().forEach(function (user) {                                                              // 249
    if (user.identities.length != 1) {                                                                       // 250
      throw new Error("User does not have exactly one identity: ", user);                                    // 251
    }                                                                                                        // 252
    var identity = user.identities[0];                                                                       // 253
    if (Match.test(identity.service, Object)) { return; } // Already migrated.                               // 254
                                                                                                             // 255
    var newIdentity = _.pick(identity, "id", "main", "noLogin", "verifiedEmail", "unverifiedEmail");         // 256
    newIdentity.profile = _.pick(identity, "name", "handle", "picture", "pronoun");                          // 257
                                                                                                             // 258
    var serviceObject = {};                                                                                  // 259
    var fieldsToUnset = {}                                                                                   // 260
                                                                                                             // 261
    if (identity.service === "dev") {                                                                        // 262
      serviceObject.name = user.devName;                                                                     // 263
      fieldsToUnset.devName = 1;                                                                             // 264
    } else if (identity.service === "emailToken") {                                                          // 265
      serviceObject.tokens = user.services.emailToken.tokens;                                                // 266
      serviceObject.email = user.services.emailToken.email;                                                  // 267
      fieldsToUnset["services.emailToken"] = 1;                                                              // 268
    }                                                                                                        // 269
                                                                                                             // 270
    newIdentity.service = {};                                                                                // 271
    newIdentity.service[identity.service] = serviceObject;                                                   // 272
                                                                                                             // 273
    var modifier = {$set: {identities: [newIdentity]}};                                                      // 274
    if (Object.keys(fieldsToUnset).length > 0) {                                                             // 275
      modifier["$unset"] = fieldsToUnset;                                                                    // 276
    }                                                                                                        // 277
                                                                                                             // 278
    Meteor.users.update({_id: user._id}, modifier);                                                          // 279
  });                                                                                                        // 280
}                                                                                                            // 281
                                                                                                             // 282
function repairEmailIdentityIds() {                                                                          // 283
  var Crypto = Npm.require("crypto");                                                                        // 284
  Meteor.users.find({"identities.service.emailToken": {$exists: 1}}).forEach(function (user) {               // 285
    if (user.identities.length != 1) {                                                                       // 286
      throw new Error("User does not have exactly one identity: ", user);                                    // 287
    }                                                                                                        // 288
    var identity = user.identities[0];                                                                       // 289
    var newIdentity = _.pick(identity, "main", "noLogin", "verifiedEmail", "unverifiedMail",                 // 290
                             "profile");                                                                     // 291
    newIdentity.service = {email: identity.service.emailToken};                                              // 292
    newIdentity.id = Crypto.createHash("sha256")                                                             // 293
      .update("email:" + identity.service.emailToken.email).digest("hex");                                   // 294
                                                                                                             // 295
    Grains.update({identityId: identity.id}, {$set: {identityId: newIdentity.id}}, {multi: true});           // 296
    Sessions.update({identityId: identity.id}, {$set: {identityId: newIdentity.id}}, {multi: true});         // 297
    ApiTokens.update({identityId: identity.id},                                                              // 298
                     {$set: {identityId: newIdentity.id}},                                                   // 299
                     {multi: true});                                                                         // 300
    ApiTokens.update({"owner.user.identityId": identity.id},                                                 // 301
                     {$set: {"owner.user.identityId": newIdentity.id}},                                      // 302
                     {multi: true});                                                                         // 303
    ApiTokens.update({"owner.grain.introducerIdentity": identity.id},                                        // 304
                     {$set: {"owner.grain.introducerIdentity": newIdentity.id}},                             // 305
                     {multi: true});                                                                         // 306
                                                                                                             // 307
    while (ApiTokens.update({"requirements.permissionsHeld.identityId": identity.id},                        // 308
                            {$set: {"requirements.$.permissionsHeld.identityId": newIdentity.id}},           // 309
                            {multi: true}) > 0);                                                             // 310
                                                                                                             // 311
    Meteor.users.update({_id: user._id}, {$set: {identities: [newIdentity]}});                               // 312
  });                                                                                                        // 313
}                                                                                                            // 314
                                                                                                             // 315
function splitAccountUsersAndIdentityUsers() {                                                               // 316
  Meteor.users.find({"identities": {$exists: true}}).forEach(function (user) {                               // 317
    if (user.identities.length != 1) {                                                                       // 318
      throw new Error("User does not have exactly one identity: ", user);                                    // 319
    }                                                                                                        // 320
    var identity = user.identities[0];                                                                       // 321
    var identityUser = _.pick(user, "createdAt", "lastActive", "expires")                                    // 322
    identityUser._id = identity.id;                                                                          // 323
    identityUser.profile = identity.profile;                                                                 // 324
    _.extend(identityUser, _.pick(identity, "unverifiedEmail"));                                             // 325
    identityUser.profile.service = Object.keys(identity.service)[0];                                         // 326
                                                                                                             // 327
    // Updating this user needs to be a two step process because the `services` field typically              // 328
    // contains subfields that are constrained to be unique by Mongo indices.                                // 329
    identityUser.stagedServices = _.omit(user.services, "resume");                                           // 330
    if (identity.service.dev) {                                                                              // 331
      identityUser.stagedServices.dev = identity.service.dev;                                                // 332
    } else if (identity.service.email) {                                                                     // 333
      identityUser.stagedServices.email = identity.service.email;                                            // 334
    }                                                                                                        // 335
                                                                                                             // 336
    var accountUser = _.pick(user, "_id", "createdAt", "lastActive", "expires",                              // 337
                             "isAdmin", "signupKey", "signupNote", "signupEmail",                            // 338
                             "plan", "storageUsage", "isAppDemoUser", "appDemoId",                           // 339
                             "payments", "dailySentMailCount", "hasCompletedSignup");                        // 340
    accountUser.loginIdentities = [_.pick(identity, "id")];                                                  // 341
    accountUser.nonloginIdentities = [];                                                                     // 342
    if (user.services && user.services.resume) {                                                             // 343
      accountUser.services = {resume: user.services.resume};                                                 // 344
    }                                                                                                        // 345
                                                                                                             // 346
    accountUser.stashedOldUser = user;                                                                       // 347
                                                                                                             // 348
    ApiTokens.update({identityId: identityUser._id}, {$set: {accountId: user._id}},                          // 349
                     {multi: true});                                                                         // 350
                                                                                                             // 351
    Meteor.users.upsert({_id: identityUser._id}, identityUser);                                              // 352
    Meteor.users.update({_id: user._id}, accountUser);                                                       // 353
  });                                                                                                        // 354
                                                                                                             // 355
  Meteor.users.find({"stagedServices": {$exists: true}}).forEach(function (identity) {                       // 356
    Meteor.users.update({_id: identity._id}, {$unset: {stagedServices: 1},                                   // 357
                                              $set: {services: identity.stagedServices}});                   // 358
  });                                                                                                        // 359
}                                                                                                            // 360
                                                                                                             // 361
function populateContactsFromApiTokens() {                                                                   // 362
  ApiTokens.find({"owner.user.identityId": {$exists: 1},                                                     // 363
                  accountId: {$exists: 1}}).forEach(function(token) {                                        // 364
    var identityId = token.owner.user.identityId;                                                            // 365
    var identity = SandstormDb.prototype.getIdentity(identityId);                                            // 366
    if (identity) {                                                                                          // 367
      var profile = identity.profile;                                                                        // 368
      Contacts.upsert({ownerId: token.accountId, identityId: identityId}, {                                  // 369
        ownerId: token.accountId,                                                                            // 370
        petname: profile && profile.name,                                                                    // 371
        created: new Date(),                                                                                 // 372
        identityId: identityId,                                                                              // 373
        profile: profile,                                                                                    // 374
      });                                                                                                    // 375
    }                                                                                                        // 376
  });                                                                                                        // 377
}                                                                                                            // 378
                                                                                                             // 379
function cleanUpApiTokens() {                                                                                // 380
  // The `splitUserIdsIntoAccountIdsAndIdentityIds()` migration only added `identityId` in cases             // 381
  // where the user still existed in the database.                                                           // 382
  ApiTokens.remove({userId: {$exists: true}, identityId: {$exists: false}});                                 // 383
  ApiTokens.remove({"owner.user.userId": {$exists: true},                                                    // 384
                    "owner.user.identityId": {$exists: false}});                                             // 385
                                                                                                             // 386
  // For a while we were accidentally setting `appIcon` instead of `icon`.                                   // 387
  ApiTokens.find({"owner.user.denormalizedGrainMetadata.appIcon": {$exists: true}}).forEach(                 // 388
      function (apiToken) {                                                                                  // 389
    var icon = apiToken.owner.user.denormalizedGrainMetadata.appIcon;                                        // 390
    ApiTokens.update({_id: apiToken._id},                                                                    // 391
                     {$set: {"owner.user.denormalizedGrainMetadata.icon": icon},                             // 392
                      $unset: {"owner.user.denormalizedGrainMetadata.appIcon": true}});                      // 393
  });                                                                                                        // 394
                                                                                                             // 395
  // For a while the `identityId` field of child UiView tokens was not getting set.                          // 396
  function repairChain(parentToken) {                                                                        // 397
    ApiTokens.find({parentToken: parentToken._id, grainId: {$exists: true},                                  // 398
                    identityId: {$exists: false}}).forEach(function (childToken) {                           // 399
      ApiTokens.update({_id: childToken._id}, {$set: {identityId: parentToken.identityId}});                 // 400
      repairChain(childToken);                                                                               // 401
    });                                                                                                      // 402
  }                                                                                                          // 403
  ApiTokens.find({grainId: {$exists: true}, identityId: {$exists: true},                                     // 404
                  parentToken: {$exists: false}}).forEach(repairChain);                                      // 405
}                                                                                                            // 406
                                                                                                             // 407
function initServerTitleAndReturnAddress() {                                                                 // 408
  var hostname = Url.parse(process.env.ROOT_URL).hostname;                                                   // 409
  Settings.insert({_id: "serverTitle", value: hostname});                                                    // 410
  Settings.insert({_id: "returnAddress", value: "no-reply@" + hostname});                                    // 411
}                                                                                                            // 412
                                                                                                             // 413
function sendReferralNotifications() {                                                                       // 414
  if (Meteor.settings.public.quotaEnabled) {                                                                 // 415
    Meteor.users.find({                                                                                      // 416
      loginIdentities: {$exists: true},                                                                      // 417
      expires: {$exists: false},                                                                             // 418
    }, {fields: {_id: 1}}).forEach(function (user) {                                                         // 419
      sendReferralProgramNotification(user._id);                                                             // 420
    });                                                                                                      // 421
  }                                                                                                          // 422
}                                                                                                            // 423
                                                                                                             // 424
function assignBonuses() {                                                                                   // 425
  if (Meteor.settings.public.quotaEnabled && SandstormDb.bonusesMigrationHook) {                             // 426
    SandstormDb.bonusesMigrationHook();                                                                      // 427
  }                                                                                                          // 428
}                                                                                                            // 429
                                                                                                             // 430
// This must come after all the functions named within are defined.                                          // 431
// Only append to this list!  Do not modify or remove list entries;                                          // 432
// doing so is likely change the meaning and semantics of user databases.                                    // 433
var MIGRATIONS = [                                                                                           // 434
  updateLoginStyleToRedirect,                                                                                // 435
  enableLegacyOAuthProvidersIfNotInSettings,                                                                 // 436
  denormalizeInviteInfo,                                                                                     // 437
  mergeRoleAssignmentsIntoApiTokens,                                                                         // 438
  fixOasisStorageUsageStats,                                                                                 // 439
  fetchProfilePictures,                                                                                      // 440
  assignPlans,                                                                                               // 441
  removeKeyrings,                                                                                            // 442
  useLocalizedTextInUserActions,                                                                             // 443
  verifyAllPgpSignatures,                                                                                    // 444
  splitUserIdsIntoAccountIdsAndIdentityIds,                                                                  // 445
  appUpdateSettings,                                                                                         // 446
  moveDevAndEmailLoginDataIntoIdentities,                                                                    // 447
  repairEmailIdentityIds,                                                                                    // 448
  splitAccountUsersAndIdentityUsers,                                                                         // 449
  populateContactsFromApiTokens,                                                                             // 450
  cleanUpApiTokens,                                                                                          // 451
  initServerTitleAndReturnAddress,                                                                           // 452
  sendReferralNotifications,                                                                                 // 453
  assignBonuses,                                                                                             // 454
];                                                                                                           // 455
                                                                                                             // 456
function migrateToLatest() {                                                                                 // 457
  if (Meteor.settings.replicaNumber) {                                                                       // 458
    // This is a replica. Wait for the first replica to perform migrations.                                  // 459
                                                                                                             // 460
    console.log("Waiting for migrations on replica zero...");                                                // 461
                                                                                                             // 462
    var done = new Future();                                                                                 // 463
    var change = function (doc) {                                                                            // 464
      console.log("Migrations applied elsewhere: " + doc.value + "/" + MIGRATIONS.length);                   // 465
      if (doc.value >= MIGRATIONS.length) done.return();                                                     // 466
    }                                                                                                        // 467
    var observer = Migrations.find({_id: "migrations_applied"}).observe({                                    // 468
      added: change,                                                                                         // 469
      changed: change                                                                                        // 470
    });                                                                                                      // 471
                                                                                                             // 472
    done.wait();                                                                                             // 473
    observer.stop();                                                                                         // 474
    console.log("Migrations have completed on replica zero.");                                               // 475
                                                                                                             // 476
  } else {                                                                                                   // 477
    var applied = Migrations.findOne({_id: "migrations_applied"});                                           // 478
    var start;                                                                                               // 479
    if (!applied) {                                                                                          // 480
      // Migrations table is not yet seeded with a value.  This means it has                                 // 481
      // applied 0 migrations.  Persist this.                                                                // 482
      Migrations.insert({_id: "migrations_applied", value: 0});                                              // 483
      start = 0;                                                                                             // 484
    } else {                                                                                                 // 485
      start = applied.value;                                                                                 // 486
    }                                                                                                        // 487
    console.log("Migrations already applied: " + start + "/" + MIGRATIONS.length);                           // 488
                                                                                                             // 489
    for (var i = start ; i < MIGRATIONS.length ; i++) {                                                      // 490
      // Apply migration i, then record that migration i was successfully run.                               // 491
      console.log("Applying migration " + (i+1));                                                            // 492
      MIGRATIONS[i]();                                                                                       // 493
      Migrations.update({_id: "migrations_applied"}, {$set: {value: i+1}});                                  // 494
      console.log("Applied migration " + (i+1));                                                             // 495
    }                                                                                                        // 496
  }                                                                                                          // 497
}                                                                                                            // 498
                                                                                                             // 499
SandstormDb.prototype.migrateToLatest = migrateToLatest;                                                     // 500
                                                                                                             // 501
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['sandstorm-db'] = {
  SandstormDb: SandstormDb
};

})();

//# sourceMappingURL=sandstorm-db.js.map
